#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>

#include <SDL/SDL.h>

#include <inc/log.h>
#include <inc/asm.h>
#include <inc/sig.h>
#include <inc/tid.h>
#include <inc/dmm.h>
#include <inc/mem.h>
#include <inc/int.h>
#include <inc/xms.h>
#include <inc/ems.h>
#include <inc/ldt.h>
#include <inc/dos.h>
#include <inc/sel.h>
#include <inc/vga.h>
#include <inc/vid.h>
#include <inc/kbd.h>
#include <inc/mou.h>
#include <inc/pic.h>
#include <inc/ldr.h>
#include <inc/eio.h>
#include <inc/pit.h>
#include <inc/rtc.h>
#include <inc/dma.h>
#include <inc/mup.h>
#include <inc/evt.h>
#include <inc/dpmi.h>
#include <inc/aud.h>
#include <inc/cli.h>
#include <inc/vbe.h>
#include <inc/gus.h>
#include <inc/sb16.h>
#include <inc/sou.h>
#include <inc/pie.h>
#include <inc/x86.h>
#include <inc/cmos.h>

#define LOG_PREFIX "ADB"

static struct imginfo info;

static char *safe_cwd (char *suffix, int suffix_len)
{
  int retlen;
  char *cwd;

  if (suffix_len) {
    /*                               /   z */
    retlen = PATH_MAX + suffix_len + 1 + 1;
  }
  else {
    /*               z */
    retlen = PATH_MAX + 1;
  }

  cwd = malloc (retlen);

  if (NULL == cwd) {
    ERRFail ("malloc (%d) for exe directory failed",
             PATH_MAX + suffix_len + 1);
  }

  if (NULL == getcwd (cwd, PATH_MAX)) {
    ERRFail ("failed to obtain current working directory");
  }

  if (suffix) {
    strcat (cwd, "/");
    strncat (cwd, suffix, suffix_len);
  }

  return cwd;
}

static void split (char *full, char **a_dir, char **a_name)
{
  char *name;
  char *last_slash;
  char *dir;

  last_slash = strrchr (full, '/');

  if (NULL == last_slash) {
    dir = safe_cwd (NULL, 0);
    name = full;
  }
  else {
    int dir_len;

    dir_len = last_slash - full;

    if (full[0] == '/') {
      dir = strndup (full, dir_len);
      if (NULL == dir) {
        ERRFail ("strndup(`%s', %d) failed", full, dir_len);
      }
    }
    else {
      dir = safe_cwd (full, dir_len);
    }
    name = last_slash + 1;
  }
  *a_name = name;
  *a_dir = dir;
}

static void alrm (int nsig)
{
  DEREF (nsig);
  VID_run ();
}

void finished (const char *reason)
{
  const char *msg = "\r\nPress any key to exit";

  SIG_block ();
  VID_set_title ("completed");

  signal (SIGALRM, alrm);
  SIG_unblock (0);

  VGA_print_string_TTY (reason, strlen (reason));
  VGA_print_string_TTY (msg, strlen (msg));
  EVT_wait_for_key ();
  exit (EXIT_SUCCESS);
}

static void run ()
{
  regs86 r;
  us psp;

  psp = DOS_get_PSP () >> 4;

  memset (&r, 0, sizeof (r));

  SIGECX (&r) = 0xff;
  SIGEDX (&r) = psp;
  SIGEDI (&r) = 0xfffe;
  SIGESI (&r) = 0x100;
  SIGEBP (&r) = 0x91c;
  SIGEIP (&r) = info.ip;
  SIGESP (&r) = GETLOWORD (info.sp);

  SIGCS (&r) = info.cs;
  SIGDS (&r) = psp;
  SIGES (&r) = psp;
  SIGSS (&r) = info.ss;
  SIG_start (&r);
}

int main (int argc, char **argv)
{
  char *exe;
  char *dir, *name;
  char *args;

  SDL_Init (SDL_INIT_NOPARACHUTE);

  AUD_pre_init ();
  DMM_pre_init ();
  DOS_pre_init ();
  EMS_pre_init ();
  EVT_pre_init ();
  GUS_pre_init ();
  KBD_pre_init ();
  LOG_pre_init ();
  MEM_pre_init ();
  MOU_pre_init ();
  PIE_pre_init ();
  PIT_pre_init ();
  SB16_pre_init ();
  SIG_pre_init ();
  VBE_pre_init ();
  VGA_pre_init ();
  VID_pre_init ();
  X86_pre_init ();
  XMS_pre_init ();
  CLI_pre_init ();
  DPMI_pre_init ();

  MUP_init (argc, argv);
  exe = MUP_get_string ("exe");

  if (NULL == exe) {
    _log ("Usage: minde program.exe [arguments]\n");
    exit (EXIT_FAILURE);
  }

  TID_init ();
  LOG_init ();
  ASM_init ();
  CLI_init ();
  MEM_init ();
  DMM_init ();
  XMS_init ();
  SIG_init ();
  PIE_init ();
  LDT_init ();
  SEL_init ();
  INT_init ();
  EMS_init ();
  DOS_init ();
  DPMI_init ();
  EIO_init ();
  X86_init ();
  VID_init ();
  EVT_init ();
  PIC_init ();
  PIT_init ();
  KBD_init ();
  MOU_init ();
  CMOS_init ();
  RTC_init ();
  DMA_init ();
  SOU_init ();
  AUD_init ();

  split (exe, &dir, &name);
  LDR_load (exe, &info);

  DOS_map_default_drive (dir);
  DOS_put_name (name);
  args = MUP_get_string ("args");
  if (NULL != args)
    DOS_put_cmdline (args);

  dir = MUP_get_string ("dir");
  if (NULL != dir) {
    if (-1 == DOS_set_cur_dir (dir)) {
      int drive;
      char *map_path;
      char *wd;

      _log ("dir `%s' (from minde.mup) does not denote valid DOS directory\n",
           dir);

      drive = DOS_get_drive ();
      if (-1 == DOS_get_fs_map_info (drive, &map_path, &wd)) {
        _log ("failed to obtain default drive mapping");
        exit (EXIT_FAILURE);
      }
      _log ("drive %c=`%s'\n"
            "%c's current directory `%s'\n",
            drive + 'A', map_path, drive + 'A', wd);
    }
  }

  VID_set_title (name);

  run ();
  return 0;
}
