#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

#include <inc/mem.h>
#include <inc/tid.h>
#include <inc/aud.h>
#include <inc/bit.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "AUD"
static int log_level;

#define IOCTL(args) do {                        \
  int ret = ioctl args;                         \
  if (-1 == ret) {                              \
    ERRFail (#args);                            \
  }                                             \
  ldebug ("ioctl " #args " = %d\n", ret);       \
} while (0)

static int _fragsize;
static int audio_fd = -1;
static int freq;
static int conf_nfrags;
static int conf_fragsize;
static int nfrags;
static int fragsize;
static int bufsize;
static int nchannels;
static int fmt;
static int rpos;
static int wpos;
static int atom;
static int live;
static int leftover;
static int bytes_per_second;
static void *buf;
static enum {DONT, DSP, TID} estimate = DONT;
static audemu_e emu = AUD_EMU_NONE;

static void (*copy_fn)(void *, void *, int);

static void copy_no_conversion (void *dst, void *src, int size)
{
  memcpy (dst, src, size);
}

static void copy_u16_to_s16 (void *dst, void *src, int size)
{
  int i;
  us *out, *in;

  out = dst;
  in = src;

  for (i = 0; i < size / 2; i++) {
    out[i] = in[i] + 0x8000;
  }
}

static void pab (struct audio_buf_info *abinfo)
{
  DEREF (abinfo);

  ldebug ("fragments %d, fragstotal %d, fragsize %d, bytes %d\n"
          "rpos %d, wpos %d, live %d\n",
          abinfo->fragments,
          abinfo->fragstotal,
          abinfo->fragsize,
          abinfo->bytes,
          rpos, wpos, live);
}

void AUD_reset (int rfreq, int rnchannels, audfmt_e rfmt)
{
  int fmt_;
  int bits16;

  if (-1 == audio_fd) {
    AUD_open (rfreq, rnchannels, rfmt);
    return;
  }

  switch (rfmt) {
    case AUD_FMT_U8:
      bits16 = 0;
      fmt_ = AFMT_U8;
      copy_fn = copy_no_conversion;
      atom = 1;
      break;

    case AUD_FMT_S8:
      Fail ("can not play 8bit signed");

    case AUD_FMT_S16:
      bits16 = 1;
      fmt_ = AFMT_S16_LE;
      copy_fn = copy_no_conversion;
      atom = 2;
      break;

    case AUD_FMT_U16:
      bits16 = 1;
      fmt_ = AFMT_S16_LE;
      copy_fn = copy_u16_to_s16;
      atom = 2;
      break;

    default:
      Aldri ();
  }

  if ((fmt_ == fmt) && (bits16 + 1 == nchannels) && (rfreq == freq))
    return;
  else {
    AUD_open (rfreq, rnchannels, rfmt);
  }
    /* Fail ("yes i'm lazy"); */
}

void AUD_open (int rfreq, int rnchannels, audfmt_e rfmt)
{
  int fmt_;
  int mmmmssss;
  struct audio_buf_info abinfo;
  int _fmt;
  int _freq;
  int _nchannels;
  int bits16;

  bits16 = 0;

  switch (rfmt) {
    case AUD_FMT_U8:
      bits16 = 0;
      fmt_ = AFMT_U8;
      copy_fn = copy_no_conversion;
      atom = 1;
      break;

    case AUD_FMT_S8:
      Fail ("can not play 8bit signed");

    case AUD_FMT_S16:
      bits16 = 1;
      fmt_ = AFMT_S16_LE;
      copy_fn = copy_no_conversion;
      atom = 2;
      break;

    case AUD_FMT_U16:
      bits16 = 1;
      fmt_ = AFMT_S16_LE;
      copy_fn = copy_u16_to_s16;
      atom = 2;
      break;

    default:
      Aldri ();
  }

  if (buf) {
    if (-1 == munmap (buf, bufsize)) {
      ERRFail ("failed to unmap audio buffer %p %d",
               buf, bufsize);
    }
    buf = 0;
  }

  if (-1 != audio_fd)
    close (audio_fd);

  audio_fd = open ("/dev/dsp", O_WRONLY | O_NONBLOCK);
  if (-1 == audio_fd) {
    ERRFail ("can not open /dev/dsp");
  }

  _fmt = fmt_;
  _freq = rfreq;
  _nchannels = rnchannels;

  IOCTL ((audio_fd, SNDCTL_DSP_RESET, 1));
  IOCTL ((audio_fd, SNDCTL_DSP_SAMPLESIZE, &_fmt));
  IOCTL ((audio_fd, SNDCTL_DSP_CHANNELS, &_nchannels));
  IOCTL ((audio_fd, SNDCTL_DSP_SPEED, &_freq));
  IOCTL ((audio_fd, SNDCTL_DSP_NONBLOCK));

  /* from oss.pdf:

     The argument to this call is an integer encoded as 0xMMMMSSSS (in
     hex). The 16 least significant bits determine the fragment
     size. The size is 2^SSSS. For example SSSS=0008 gives fragment
     size of 256 bytes (2^8). The minimum is 16 bytes (SSSS=4) and the
     maximum is total_buffer_size/2. Some devices or processor
     architectures may require larger fragments - in this case the
     requested fragment size is automatically increased.

     So ahem... 4096 = 2^12, and grand total 0x0004000c
  */

  mmmmssss = (conf_nfrags << 16) | conf_fragsize;
  IOCTL ((audio_fd, SNDCTL_DSP_SETFRAGMENT, &mmmmssss));

  linfo ("_fmt = %d, fmt = %d\n"
         "_channels = %d, rnchannels = %d\n"
         "_freq = %d, freq = %d\n",
         _fmt, fmt_,
         _nchannels, rnchannels,
         _freq, rfreq);

  if (_fmt != fmt_) {
    Fail ("format %d != %d", _fmt, fmt_);
  }

  if (_nchannels != rnchannels) {
    Fail ("channels %d != %d", _nchannels, rnchannels);
  }

  if (_freq != rfreq) {
    Fail ("freq %d != %d", _freq, rfreq);
  }

  IOCTL ((audio_fd, SNDCTL_DSP_GETOSPACE, &abinfo));

  nfrags = abinfo.fragstotal;
  fragsize = abinfo.fragsize;
  freq = _freq;
  fmt = _fmt;
  nchannels = rnchannels;
  atom <<= nchannels >>  1;
  bufsize = nfrags * fragsize;

  bytes_per_second = (freq << (nchannels >> 1)) << bits16;

  linfo ("bytes per second %d\n", bytes_per_second);

  linfo ("fragments %d, fragstotal %d, fragsize %d, bytes %d, bufsize %d\n",
         abinfo.fragments,
         abinfo.fragstotal,
         abinfo.fragsize,
         abinfo.bytes,
         bufsize);

  if (NULL == buf) {
    /* buf = MEM_map (bufsize, MEM_MAP_WRONLY); */
    buf = MEM_map (bufsize, MEM_MAP_RDWR);
    if (MEM_MAP_FAILED == buf) {
      ERRFail ("MEM_map audio buf");
    }
  }

  rpos = 0;
  wpos = 0;
  live = 0;
}

int AUD_write (void *in_buf, int size)
{
  int to_copy, temp;
  ub *in, *out;

  to_copy = MIN (bufsize - live, size);

  temp = to_copy;

  in = in_buf;
  out = buf;

  while (temp) {
    int copy;

    copy = MIN (temp, bufsize - wpos);
    copy_fn (out + wpos, in, copy);

    wpos += copy;
    if (wpos == bufsize) {
      wpos = 0;
    }

    temp -= copy;
    in += copy;
    live += copy;
  }

  return to_copy;
}

void AUD_run ()
{
  int res;
  int bytes;
  struct audio_buf_info abinfo;

  if (0 == live)
    return;

  res = ioctl (audio_fd, SNDCTL_DSP_GETOSPACE, &abinfo);

  if (-1 == res) {
    int err;

    err = errno;
    lwarn ("SNDCTL_DSP_GETOSPACE failed with %s\n", strerror (err));
  }

  bytes = abinfo.bytes;
  bytes = MIN (live, bytes);
#if 0
  bytes = (bytes / fragsize) * fragsize;
#endif

  while (bytes) {
    int left, play, written;

    left = bufsize - rpos;
    play = MIN (left, bytes);
    written = write (audio_fd, (void *) ((ui) buf + rpos), play);

    if (-1 == written) {
      if (EAGAIN == errno || EINTR == errno) {
        return;
      }
      else {
        ERRFail ("write audio");
      }
    }

    play = written;
    live -= play;
    rpos += play;
    bytes -= play;

    if (rpos == bufsize) {
      rpos = 0;
    }
  }
}

static int get_dsp_bytes ()
{
  int res;
  struct count_info info;

  res = ioctl (audio_fd, SNDCTL_DSP_GETOPTR, &info);
  if (-1 == res) {
    int err;

    err = errno;
    lwarn ("SNDCTL_DSP_GETOPTR failed with %s\n", strerror (err));
    return -1;
  }
  else {
    ldebug ("bytes %d\n", info.bytes);
    return info.bytes;
  }
}

void AUD_adjust_estimate (int _leftover)
{
  leftover = _leftover;
}

int AUD_get_free ()
{
  int free, elapsed;

  free = bufsize - live;

  if (0 == free)
    return 0;

  elapsed = free;
  switch (estimate) {
    case DONT:
      break;

    case DSP:
    {
      static int old_bytes;
      int bytes;

      bytes = get_dsp_bytes ();
      if (bytes <= 0)
        return free;

      elapsed = bytes - old_bytes;
      old_bytes = bytes;
      ldebug ("dsp elapsed %d bytes\n", elapsed);
      break;
    }

    case TID:
    {
      static ul old_nsec;
      ul nsec, delta;
      ul ua_elapsed;
      ul al_elapsed;

      nsec = TID_get_abs_nsec ();
      delta = nsec - old_nsec;
      old_nsec = nsec;

      ua_elapsed = (delta * bytes_per_second) / (ul) TID_NSEC;
      al_elapsed = ua_elapsed & ~3ULL;

      ldebug ("tid elapsed %llu bytes\n", ua_elapsed);

      if (al_elapsed > (ul) INT_MAX)
        elapsed = INT_MAX;
      else
        elapsed = al_elapsed;

      elapsed += leftover;
    }
  }

  if (elapsed > free) {
    lwarn ("audio can not keep up\n");
    return free;
  }
  else {
    return elapsed;
  }
}

int AUD_get_live ()
{
  return live;
}

int AUD_get_buffer_size ()
{
  return bufsize;
}

static int audio_est_setter (char *desc)
{
  if (0 == strcmp (desc, "none")) {
    estimate = DONT;
  }
  else if (0 == strcmp (desc, "dsp")) {
    estimate = DSP;
  }
  else if (0 == strcmp (desc, "tid")) {
    estimate = TID;
  }
  else {
    log ("bogus audio estimation type `%s'\n"
         "must be one of: `none', `dsp', `tid'\n", desc);
    return -1;
  }
  return 0;
}

static int audio_emu_setter (char *desc)
{
  if (0 == strcmp (desc, "none")) {
    emu = AUD_EMU_NONE;
  }
  else if (0 == strcmp (desc, "gus")) {
    emu = AUD_EMU_GUS;
  }
  else if (0 == strcmp (desc, "sb")) {
    emu = AUD_EMU_SB16;
  }
  else {
    Fail ("bogus sound emulation `%s'\n"
          "must be one of: `none', `gus', `sb'", desc);
    return -1;
  }
  return 0;
}

void AUD_pre_init ()
{
  static struct MUP_key keys[] = {
    {"emu"               , TAG_SETTER , {.setter = &audio_emu_setter}},
    {"estimate"          , TAG_SETTER , {.setter = audio_est_setter}},
    {"no.env"            , TAG_IN_INT , {.ii = 0}},
    {"no.irq"            , TAG_IN_INT , {.ii = 0}},
    {"frag.size"         , TAG_INT    , {.ip = &_fragsize}},
    {"nfrags"            , TAG_INT    , {.ip = &conf_nfrags}},
  };

  static struct MUP_group group =
    {"audio.", LENOFA (keys), keys, NULL};

  MUP_add_group (&group);
}

audemu_e AUD_get_emu_type (void)
{
  return emu;
}

void AUD_init ()
{
  int fsp;

  DEREF (pab);

  log_level = MUP_get_log_level ("aud");

  fsp = _fragsize;
  if (0 != (fsp & (fsp - 1))) {
    Fail ("fragment size %d is not power of 2", fsp);
  }

  conf_fragsize = lsbindex (fsp);
}
