#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <inc/x86.h>
#include <inc/asm.h>
#include <inc/sig.h>
#include <inc/sel.h>
#include <inc/cli.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "CLI"
static int log_level;

#define MFL (                                   \
  X86_EFLAGS_2O                                 \
| X86_EFLAGS_3Z                                 \
| X86_EFLAGS_5Z                                 \
| X86_EFLAGS_TF                                 \
| X86_EFLAGS_IF                                 \
| X86_EFLAGS_IOPL                               \
| X86_EFLAGS_15Z)

#define VFL (X86_EFLAGS_2O | X86_EFLAGS_IF)

#define MEFL (MFL)
#define VEFL (VFL)

enum loc {
  None,
  Stack,
  Reg
};

struct cli {
  ui addr;
  enum loc loc;
  int big;
  int disp;
};

static int search_dist = 5;
static cli_type_e cli_type = CLI_PUSHF;
static struct cli *clis;
static int nclis_allocated;

static int find_addr (ui addr)
{
  int i;

  for (i = 0; i < nclis_allocated; i++) {
    if (clis[i].addr == addr)
      return i;
  }
  return -1;
}

static int new_cli (ui addr)
{
  int i;

  for (i = 0; i < nclis_allocated; i++) {
    if (0 == clis[i].addr) {
      clis[i].addr = addr;
      clis[i].loc = None;
      return i;
    }
  }

  clis = xrealloc (clis, nclis_allocated * 2 * sizeof (struct cli));
  memset (clis + nclis_allocated, 0, nclis_allocated * sizeof (struct cli));
  i = nclis_allocated;
  nclis_allocated *= 2;

  Fail ("out of cli space");
}

static int process (int enabled, regs86 *r, int ncli)
{
  switch (clis[ncli].loc) {
    case None:
      Aldri ();

    case Stack:
    {
      ui pos;
      ui *fp;

      pos = SIGESP (r) + clis[ncli].disp;

      lverbose ("enabled %d flags on stack at %#x disp %d\n",
                enabled, pos, clis[ncli].disp);

      fp = (ui *) SELLIN (SIGSS (r), pos);
      if (enabled)
        fp[0] = fp[0] | X86_EFLAGS_TF;
      else
        fp[0] = fp[0] & ~X86_EFLAGS_IF;
      break;
    }

    case Reg:
    {
      reg86 *regp;

      lverbose ("enabled %d flags in %s\n",
                enabled, X86_reg_name (clis[ncli].disp));
      regp = SIG_gregp (r, clis[ncli].disp);

      if (enabled)
        regp[0] = regp[0] | X86_EFLAGS_TF;
      else
        regp[0] = regp[0] & ~X86_EFLAGS_IF;
      break;
    }

    default:
      Fail ("clis[%d].loc = %d, what??", ncli, clis[ncli].loc);
  }

  return ncli;
}

int CLI_find_pushf (int enabled, regs86 *r)
{
  int ncli, dist, res, pos, opref, flags_big, c32, s32;
  ui addr;
  ub *pc, *pushfpc;
  regs86 replay;

  c32 = ASM_is_segment_big (SIGCS (r));
  s32 = ASM_is_segment_big (SIGSS (r));
  Fail_if ((-1 == c32) || (-1 == s32));
  addr = SELLIN (SIGCS (r), SIGEIP (r));
  pc = (ub *) addr;

  if (0 == s32)
    SIGESP (r) &= 0xffff;

  ncli = find_addr (addr);

  if (-1 != ncli) {
    if (None == clis[ncli].loc)
      return -1;
    else
      return process (enabled, r, ncli);
  }

  ncli = new_cli (addr);

  pushfpc = memchr (pc - search_dist, 0x9c, search_dist);
  if (NULL == pushfpc) {
    ldebug ("no pushf in sight %#06x %#010x %p\n",
            SIGCS (r), (int) SIGEIP (r), pc);
    return -1;
  }
  else {
    linfo ("pushf found at %p\n", pushfpc);
  }

  opref = 0x66 == pushfpc[-1];
  flags_big = opref ^ c32;

  linfo ("flags_big = %d c32 %d opref %d\n", flags_big, c32, opref);
  dist = pc - pushfpc;
  if (dist == 1) {
    pos = 0;
    goto analyze;
  }

  memcpy (&replay, r, sizeof (replay));
  SIGEIP (&replay) -= dist - 1;

  res = X86_replay_stack (&replay, SIGEIP (r));

  if (-1 == res) {
    lwarn ("X86 stack replaying failed\n");
    return -1;
  }

  if (res > 7) {
    pos = SIGESP (r)  - SIGESP (&replay);

    if (pos < 0) {
      lwarn ("stack underflow after replay %#x vs %#x\n",
             (int) SIGESP (r), (int) SIGESP (&replay));
      return -1;
    }

  analyze:

    if (flags_big) {
      ui flags, mask, estimate;

      flags = *(ui *) SELLIN (SIGSS (r), SIGESP (r) + pos);
      mask = MEFL | (SIGEFL (r) & (0xffff0000 ^ X86_EFLAGS_RF));
      estimate = VEFL | (SIGEFL (r) & (0xffff0000 ^ X86_EFLAGS_RF));

      if ((flags & mask) != estimate) {
        lwarn ("at %#06x %#010x\n"
               "stack flags=%#x\n"
               "real flags=%#x\n"
               "estimated flags = %#x\n"
               "pos = %d\n",
               (int) SIGCS (r), (int) SIGEIP (r),
               flags, (int) SIGEFL (r), estimate, pos);
        return -1;
      }

      linfo ("found cli pushfd\n");
      clis[ncli].loc = Stack;
      clis[ncli].big = 1;
      clis[ncli].disp = pos;
    }
    else {
      us flags, mask, estimate;

      flags = *(us *) SELLIN (SIGSS (r), SIGESP (r) + pos);
      mask = MFL;
      estimate = VFL;
      if ((flags & mask) != estimate) {
        lwarn ("at %#06x %#010x\n"
               "stack flags=%#x\n"
               "real flags=%#x\n"
               "estimated flags = %#x\n"
               "pos = %d\n",
               (int) SIGCS (r), (int) SIGEIP (r),
               flags, (int) SIGEFL (r), estimate, pos);
        return -1;
      }

      linfo ("found cli pushfw\n");
      clis[ncli].loc = Stack;
      clis[ncli].big = 0;
      clis[ncli].disp = pos;
    }
  }
  else {
    linfo ("found cli pushf pop %s\n", X86_reg_name (res));

    clis[ncli].loc = Reg;
    clis[ncli].big = flags_big;
    clis[ncli].disp = res;
  }

  return process (enabled, r, ncli);;
}

void CLI_false_positive (int ncli)
{
  clis[ncli].loc = None;
}

static int cli_type_setter (char *desc)
{
  if (0 == strcmp (desc, "pushf")) {
    cli_type = CLI_PUSHF;
  }
  else if (0 == strcmp (desc, "obey")) {
    cli_type = CLI_OBEY;
  }
  else if (0 == strcmp (desc, "ignore")) {
    cli_type = CLI_IGNORE;
  }
  else {
    log ("wrong cli type `%s'\n"
         "moust be one of: `pushf', `obey', `ignore'\n", desc);
    return -1;
  }
  return 0;
}

void CLI_pre_init ()
{
  static struct MUP_key keys[] = {
    {"cli", TAG_SETTER, {.setter = &cli_type_setter}},
    {"pushf.dist", TAG_INT, {.ip = &search_dist}},
  };

  MUP_add_keys (keys, LENOFA (keys));
}

cli_type_e CLI_get_type ()
{
  return cli_type;
}

void CLI_init ()
{
  int size;

  log_level = MUP_get_log_level ("cli");

  size = 100 * sizeof (struct cli);
  clis = calloc (size, 1);
  if (NULL == clis) {
    ERRFail ("failed to alloc %d bytes for cli handling", size);
  }
  nclis_allocated = 100;
}
