#include <string.h>
#include <stdlib.h>

#include <inc/mem.h>
#include <inc/int.h>
#include <inc/ldt.h>
#include <inc/x86.h>
#include <inc/sel.h>
#include <inc/sig.h>
#include <inc/dos.h>
#include <inc/dpmi.h>
#include <inc/sig.h>
#include <inc/red.h>
#include <inc/dmm.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "DPMI"
static int log_level;

struct dpmi_block {
  void *addr;
  ui size;
};

static int vif_always_on;
static int dont_free;
static int ignore_rm_stack = 1;

#define BAD_HANDLE 0x10000U
#define MAX_SPECIFIC_SELECTORS 0x10
#define NCALLBACKS 16
#define INITIAL_HANDLES 100

static struct DPMI_callback callbacks[NCALLBACKS];
static ui rm_stack;
static int bios_selector;
static int first_specific_selector;
static int specific_selectors_allocated;

static struct dpmi_block *blocks;
static ui nblocks;

static inline int check_handle (ui handle)
{
  ui h;

  h = handle ^ 0x80800000;
  if ((h & 0xffff0000U) || (h >= nblocks)) {
    Fail ("illegal handle %#x %u max %u",
          handle, h, nblocks);
  }
  if (0 == blocks[h].addr) {
    lwarn ("attempt to operate on unallocated handle %#x %u\n",
          handle, h);
    return BAD_HANDLE;
  }
  return h;
}

static ui get_free_handle ()
{
  ui i;
  int was_blocked;

  for (i = 0; i < nblocks; i++) {
    if (0 == blocks[i].addr)
      return i;
  }

  was_blocked = SIG_block ();
  blocks = xrealloc (blocks, nblocks * 2 * sizeof (struct dpmi_block));
  memset (blocks + nblocks, 0, nblocks * sizeof (struct dpmi_block));
  i = nblocks;
  nblocks *= 2;
  SIG_unblock (was_blocked);

  return i;
}

int DPMI_alloc_dos (int nparas, us *psel, us *pseg)
{
  int res;
  int sel;
  ui base;

  if (0 == nparas) {
    lwarn ("zero allocation attempt\n");
    return -1;
  }

  res = DOS_alloc (nparas, &base);

  if (-1 == res) {
    lwarn ("DOS_alloc (%d) failed\n", nparas);
    return 1;
  }

  sel = LDT_alloc_selectors (1);
  if (-1 == sel) {
    lwarn ("LDT_alloc_selectors (1) failed\n");
    DOS_free (base);
    return 2;
  }

  ldebug ("(nparas=%d) base=%#x, sel=%#x\n", nparas, base, sel);
  LDT_set_selector_base (sel, base);
  LDT_set_selector_limit (sel, nparas << 4);
  /* LDT_set_selector_access_rights (sel, 0xf3); */

  *psel = sel;
  *pseg = MEM_USER (base) >> 4;
  return 0;
}

void DPMI_free_dos (int sel)
{
  ui base;

  base = LDT_get_selector_base (sel);

  if (base & 0xf) {
    lwarn ("segment operation but selector %#x does not map segment\n"
           "base = %#x\n", sel, base);
  }

  lverbose ("free sel %#x, base %#x\n", sel, base);
  LDT_free_selector (sel);
  DOS_free (base);
}

int DPMI_resize_dos (int nparas, int sel)
{
  ui base;

  if (0 == nparas) {
    lwarn ("zero re-allocation attempt\n");
    return -1;
  }

  base = SEL_get_selector_base (sel);

  if (nparas >= 0x10000 - 0x10) {
    Fail ("resize of DOS selector(%#x) to %d >= 64K not implemented",
          sel, nparas << 4);
  }

  if (base & 0xf) {
    lwarn ("segment operation but selector %#x does not map segment\n"
           "base = %#x\n", sel, base);
    return -1;
  }

  if (-1 == DOS_resize (base, nparas)) {
    return -1;
  }

  LDT_set_selector_limit (sel, nparas << 4);
  return 0;
}

int DPMI_alloc (ui size, ui *rethandle, ui *retaddr)
{
  void *addr;
  int handle;
  struct dpmi_block *block;

  if (0 == size) {
    lwarn ("zero allocation attempt\n");
    return -1;
  }

  addr = NULL;
  *retaddr = 0;
  *rethandle = 0;

  handle = get_free_handle ();
  block = blocks + handle;

  addr = DMM_alloc (size);
  ldebug ("(size=%u) = %p..%p handle [%#x]%p\n",
          size, addr,
          (char *) addr + size,
          handle, (void *) block);

  if (NULL == addr) {
    return -1;
  }

  block->addr = addr;
  block->size = size;

  *retaddr = MEM_USER ((ui) addr);
  *rethandle = handle | 0x80800000U;
  return 0;
}

int DPMI_realloc (ui old_handle_mangled, ui new_size,
                  ui *rethandle, ui *retaddr)
{
  ui old_handle, new_handle;
  struct dpmi_block *block;
  void *addr;

  ldebug ("(old_handle=%#x, new_size=%u)\n",
          old_handle_mangled, new_size);

  Fail_if (0 == new_size);

  old_handle = check_handle (old_handle_mangled);
  if (BAD_HANDLE == old_handle) {
    return -1;
  }

  new_handle = get_free_handle ();
  block = blocks + old_handle;

  addr = DMM_realloc (block->addr, block->size, new_size);

  if (NULL == addr) {
    ERRFail ("handle %#x realloc(old_addr=%p, old_size=%u, new_size=%u)",
             old_handle_mangled, block->addr, block->size, new_size);
  }

  block->addr = 0;
  block->size = 0;

  block = blocks + new_handle;

  *retaddr = MEM_USER ((ui) addr);
  *rethandle = new_handle | 0x80800000U;
  return 0;
}

int DPMI_free (ui handle_mangled)
{
  ui handle;
  struct dpmi_block *block;

  if (0 == dont_free) {
    ldebug ("(handle_mangled=%#x)\n", handle_mangled);

    handle = check_handle (handle_mangled);
    if (BAD_HANDLE == handle) {
      return -1;
    }

    block = blocks + handle;

    if (0 == block->addr) {
      lwarn ("freeing null - unlikely\n");
      return -1;
    }

    DMM_free (block->addr, block->size);

    block->addr = 0;
    block->size = 0;
  }
  return 0;
}

int DPMI_create_seg_alias (us seg)
{
  us alias;
  ui limit, base;

  if (MAX_SPECIFIC_SELECTORS == specific_selectors_allocated) {
    return -1;
  }

  alias = first_specific_selector + specific_selectors_allocated * 8;
  specific_selectors_allocated++;

  limit = 0x10000;
  base = MEM_REAL (seg << 4);

  LDT_set_selector_base (alias, base);
  LDT_set_selector_limit (alias, limit);
  /* LDT_set_selector_access_rights (alias, 0x40f3); */
  return alias;
}

int DPMI_create_alias (us sel)
{
  return LDT_create_alias (sel);
}

int DPMI_get_bios_selector ()
{
  return bios_selector;
}

void DPMI_get_rm_stack (us *seg, us *off)
{
  *seg = rm_stack >> 4;
  *off = DPMI_RM_ASYNC_STACK_PARAS << 4;
  ldebug ("%04x:%04x\n", *seg, *off);
}

int DPMI_alloc_callback (us proc_sel, ui proc_off,
                         us data_sel, ui data_off)
{
  int i;

  for (i = 0; i < NCALLBACKS; i++)
    if (0 == callbacks[i].used)
      break;

  if (callbacks[i].used)
    return -1;

  callbacks[i].proc_sel = proc_sel;
  callbacks[i].proc_off = proc_off;
  callbacks[i].data_sel = data_sel;
  callbacks[i].data_off = data_off;
  callbacks[i].used = 1;
  return i;
}

int DPMI_free_callback (ui i)
{
  if ((i > NCALLBACKS) || (0 == callbacks[i].used))
    return -1;

  callbacks[i].used = 0;
  return 0;
}

struct DPMI_callback *DPMI_get_callback (ui i)
{
  if ((i > NCALLBACKS) || (0 == callbacks[i].used))
    return NULL;

  return callbacks + i;
}

int DPMI_vif (int op)
{
  int e;

  e = INT_is_enabled ();
  if (0 == vif_always_on) {
    switch (op) {
      case 0: INT_disable (); break;
      case 1: INT_enable (); break;
      case 2: break;
      default:
        Aldri ();
    }
  }
  return e;
}

void DPMI_call_rm (int nint, int is_retf_frame, int words, char *d)
{
  int has_stack;
  struct vm86_regs r;

  if (0 != words)
    Aldri ();

  memset (&r, 0, sizeof (r));
  r.eax = DPMI_EAX (d);
  r.ebx = DPMI_EBX (d);
  r.ecx = DPMI_ECX (d);
  r.edx = DPMI_EDX (d);
  r.esi = DPMI_ESI (d);
  r.edi = DPMI_EDI (d);
  r.ebp = DPMI_EBP (d);
  r.eflags = DPMI_CFL (d);
  r.ds = DPMI_DS (d);
  r.es = DPMI_ES (d);
  r.fs = DPMI_FS (d);
  r.gs = DPMI_GS (d);

  if (-1 == nint) {
    r.cs = DPMI_CS (d);
    r.eip = DPMI_IP (d);
  }
  else {
    us seg, off;

    INT_get_rm (nint, &seg, &off);
    r.cs = seg;
    r.eip = off;
  }

  has_stack = ignore_rm_stack ? 0 : (0 != (DPMI_SS (d) | DPMI_SP (d)));

  if (has_stack) {
    us *sp;

    r.ss = DPMI_SS (d);
    r.esp = DPMI_SP (d);

    r.esp -= is_retf_frame ? 4 : 6;
    sp = (us *) ((r.ss << 4) + r.esp);
    sp[0] = RED_SWITCH_TO_PM;
    sp[1] = RED_SEGMENT;
    if (0 == is_retf_frame) sp[2] = r.eflags;
  }

  if (-1 != nint) r.eflags &= ~(X86_EFLAGS_TF | X86_EFLAGS_IF);

  SIG_vm86 (&r, 0 == has_stack);

  DPMI_EAX (d) = r.eax;
  DPMI_EBX (d) = r.ebx;
  DPMI_ECX (d) = r.ecx;
  DPMI_EDX (d) = r.edx;
  DPMI_ESI (d) = r.esi;
  DPMI_EDI (d) = r.edi;
  DPMI_EBP (d) = r.ebp;
  /* DPMI_SP (d) = r.esp; */
  DPMI_CFL (d) = r.eflags;
  /* DPMI_CS (d) = r.cs; */
  DPMI_DS (d) = r.ds;
  DPMI_ES (d) = r.es;
  /* DPMI_SS (d) = r.ss; */
  DPMI_FS (d) = r.fs;
  DPMI_GS (d) = r.gs;
  /* DPMI_IP (d) = r.eip; */
}

void DPMI_pre_init ()
{
  static struct MUP_key keys[] = {
    {"do.not.free"     , TAG_INT , {.ip = &dont_free}},
    {"vif.always.on"   , TAG_INT , {.ip = &vif_always_on}},
    {"ignore.rm.stack" , TAG_INT , {.ip = &ignore_rm_stack}},
  };

  static struct MUP_group group =
    {"dpmi.", LENOFA (keys), keys, NULL};

  MUP_add_group (&group);
}

void DPMI_init ()
{
  int i;
  ub *redir;

  log_level = MUP_get_log_level ("dpmi");

  blocks = malloc (INITIAL_HANDLES * sizeof (struct dpmi_block));
  if (NULL == blocks) {
    ERRFail ("failed to allocate %d bytes for DPMI handles",
             INITIAL_HANDLES * sizeof (struct dpmi_block));
  }

  nblocks = INITIAL_HANDLES;
  memset (blocks, 0, INITIAL_HANDLES * sizeof (struct dpmi_block));

  first_specific_selector = LDT_alloc_selectors (MAX_SPECIFIC_SELECTORS);

  if (-1 == first_specific_selector) {
    Fail ("DPMI_init failed to allocate specific selectors");
  }

  bios_selector = DPMI_create_seg_alias (0x40);

  rm_stack =
    RED_FLAT ((RED_CALLBACK_0 + NCALLBACKS +
               15 + DPMI_RM_ASYNC_STACK_PARAS * 16));

  redir = (ub *) RED_FLAT (0);
  redir[RED_SAVE_RESTORE_RM] = 0xcb;
  redir[RED_SAVE_RESTORE_PM] = 0xcb;
  redir[RED_SWITCH_TO_PM] = RED_ICE;
  redir[RED_RAW_SWITCH_TO_PM] = RED_ICE;
  redir[RED_RAW_SWITCH_TO_RM] = RED_ICE;

  for (i = 0; i < NCALLBACKS; i++) {
    redir[RED_CALLBACK_0 + i] = RED_ICE;
  }
}
