#include <SDL/SDL.h>
#include <string.h>
#include <stdlib.h>

#include <inc/evt.h>
#include <inc/tid.h>
#include <inc/kbd.h>
#include <inc/vid.h>
#include <inc/mou.h>
#include <inc/log.h>
#include <inc/sig.h>
#include <inc/mup.h>

#define LOG_PREFIX "EVT"
static int log_level = -1;

static ul period;
static ul last_update;
static int f12_pressed;

static void pause (void);
static int process_event (SDL_Event *event, int in_pause)
{
  switch (event->type) {
    case SDL_KEYUP:
    case SDL_KEYDOWN:
      ldebug ("key %s\n", SDL_GetKeyName (event->key.keysym.sym));

      if (SDLK_F12 == event->key.keysym.sym) {
        if (SDL_KEYUP == event->type) {
          f12_pressed = 0;
        }
        else if (SDL_KEYDOWN == event->type) {
          f12_pressed = 1;
        }
        return 0;
      }

      if (f12_pressed) {
        switch (event->key.keysym.sym) {
          case SDLK_f:
            if (SDL_KEYUP == event->type)
              VID_flip_fullscreen ();
            break;

          case SDLK_p:
            if (SDL_KEYUP == event->type) {
              return 1;
            }
            break;

          case SDLK_q:
            exit (EXIT_SUCCESS);
            break;

          case SDLK_d:
            if (SDL_KEYUP == event->type) {
              SIG_break (0);
            }
            break;

          case SDLK_g:
            if (SDL_KEYUP == event->type) {
              SDL_GrabMode mode;

              mode = SDL_WM_GrabInput (SDL_GRAB_QUERY);

              if (SDL_GRAB_OFF == mode)
                SDL_WM_GrabInput (SDL_GRAB_ON);
              else
                SDL_WM_GrabInput (SDL_GRAB_OFF);
            }

          default:
            break;
        }
      }
      else {
        if (0 == in_pause)
          KBD_event (&event->key);
      }
      break;

    case SDL_MOUSEMOTION:
      ldebug ("mousemotion\n");
      if (0 == in_pause)
        MOU_motion_event (&event->motion);
      break;

    case SDL_MOUSEBUTTONUP:
    case SDL_MOUSEBUTTONDOWN:
      ldebug ("mousebutton\n");
      if (0 == in_pause)
        MOU_button_event (&event->button);
      break;

#ifdef USE_HERMES
    case SDL_VIDEORESIZE:
      ldebug ("resize\n");
      VID_resize (event->resize.w, event->resize.h);
      break;
#endif

    case SDL_QUIT:
      exit (EXIT_SUCCESS);
      break;

    default:
      break;
  }

  return 0;
}

static void pause ()
{
  /* HERE BE DRAGONS (wrt sig locking/blocking) */
  SDL_Event event;
  char *saved_title;
  char *new_title;
  char *old_title;
  char *old_icon;
  int saved_title_valid;
  char tprefix[] = "Paused: ";

  TID_pause ();
  SDL_WM_GetCaption (&old_title, &old_icon);

  saved_title_valid = 0;
  saved_title = NULL;
  new_title = NULL;

  if (old_title) {
    int old_title_len;

    old_title_len = strlen (old_title);
    saved_title = malloc (old_title_len + 1);
    if (NULL == saved_title)
      goto next;

    strcpy (saved_title, old_title);
    new_title = malloc (old_title_len + sizeof (tprefix) + 1);
    if (NULL == new_title) {
      free (saved_title);
      goto next;
    }

    strcpy (new_title, tprefix);
    strcat (new_title, old_title);

    SDL_WM_SetCaption (new_title, old_icon);
    saved_title_valid = 1;
  }

 next:
  for (;;) {
    SDL_WaitEvent (&event);
    if (process_event (&event, 1))
      break;
  }

  if (saved_title_valid) {
    SDL_WM_SetCaption (saved_title, old_icon);
    free (new_title);
  }

  TID_resume ();
}

static void poll ()
{
  SDL_Event event;

  while (SDL_PollEvent (&event)) {
    if (process_event (&event, 0)) {
      pause ();
    }
  }
}

void EVT_run ()
{
  ul now, delta;

  now = TID_get_abs_nsec ();
  delta = now - last_update;

  if (delta >= period) {
    int was_locked;

    /* fugly */
    was_locked = VID_lock ();
    if (was_locked)
      return;

    poll ();

    last_update = now;
    VID_unlock ();
  }
}

void EVT_wait_for_key ()
{
  SDL_Event event;

  SDL_EnableKeyRepeat (0, 0);
  SDL_Delay (100);
  while (SDL_PollEvent (&event)) {

    switch (event.type) {
      case SDL_QUIT:
        return;

      case SDL_KEYUP:
      case SDL_KEYDOWN:
        break;

      default:
        break;
    }
  }

  while (SDL_WaitEvent (&event)) {
    switch (event.type) {
      case SDL_QUIT:
      case SDL_KEYUP:
      case SDL_KEYDOWN:
        return;
    }
  }
}

void EVT_pre_init ()
{
  static struct MUP_key keys[] =
    {{"event.poll.freq", TAG_IN_INT, {.ii = 14}}};

  MUP_add_keys (keys, LENOFA (keys));
}

void EVT_init ()
{
  int rfreq;

  log_level = MUP_get_log_level ("evt");
  DEREF (log_level);

  rfreq = MUP_get_int ("event.poll.freq");
  rfreq = rfreq ? rfreq : 14;

  period = TID_NSEC / rfreq;
}
