open Format

let use_bgr                        = ref false
let hw_sup                         = 1
let tty_sup                        = 4
let color_mode                     = 8
let graphics_mode                  = 16
let vga_compatible                 = 32
let vga_compatible_memory_mode_sup = 64
let lfb_sup                        = 128

let text_model                     = 0
let planar_model                   = 3
let packed_pixel_model             = 4
let direct_color_model             = 6

type ifmt = Dec | Hex

type field =
    Mattr
  | Aattr
  | Battr
  | Wingran
  | Winsize
  | Winaseg
  | Winbseg
  | Winfunc
  | Pitch
  | Xres
  | Yres
  | Xchar
  | Ychar
  | Planes
  | Bpp
  | Memmodel
  | Banksize
  | Imageplanes
  | Rmasksize
  | Rfieldpos
  | Gmasksize
  | Gfieldpos
  | Bmasksize
  | Bfieldpos
  | Amasksize
  | Afieldpos
  | Dcattr

let all_fields =
  [Mattr; Aattr; Battr; Wingran; Winsize;
   Winaseg; Winbseg; Winfunc; Pitch; Xres; Yres; Xchar; Ychar;
   Planes; Bpp; Memmodel; Banksize; Imageplanes; Rmasksize;
   Rfieldpos; Gmasksize; Gfieldpos; Bmasksize; Bfieldpos;
   Amasksize; Afieldpos; Dcattr]

let field_names =
  let hash = Hashtbl.create 20 in
  Hashtbl.add hash Mattr       ("vbe.mode.attr",              Hex);
  Hashtbl.add hash Wingran     ("vbe.mode.win.gran",          Dec);
  Hashtbl.add hash Winsize     ("vbe.mode.win.size",          Dec);
  Hashtbl.add hash Aattr       ("vbe.mode.wina.attr",         Hex);
  Hashtbl.add hash Battr       ("vbe.mode.winb.attr",         Hex);
  Hashtbl.add hash Winaseg     ("vbe.mode.wina.seg",          Hex);
  Hashtbl.add hash Winbseg     ("vbe.mode.winb.seg",          Hex);
  Hashtbl.add hash Pitch       ("vbe.mode.bpsl",              Dec);
  Hashtbl.add hash Xres        ("vbe.mode.xres",              Dec);
  Hashtbl.add hash Yres        ("vbe.mode.yres",              Dec);
  Hashtbl.add hash Xchar       ("vbe.mode.xchar",             Dec);
  Hashtbl.add hash Ychar       ("vbe.mode.ychar",             Dec);
  Hashtbl.add hash Planes      ("vbe.mode.planes",            Dec);
  Hashtbl.add hash Bpp         ("vbe.mode.bpp",               Dec);
  Hashtbl.add hash Memmodel    ("vbe.mode.model",             Hex);
  Hashtbl.add hash Banksize    ("vbe.mode.banksize",          Dec);
  Hashtbl.add hash Imageplanes ("vbe.mode.imageplanes",       Dec);
  Hashtbl.add hash Rmasksize   ("vbe.mode.red.mask.size",     Dec);
  Hashtbl.add hash Rfieldpos   ("vbe.mode.red.field.pos",     Dec);
  Hashtbl.add hash Gmasksize   ("vbe.mode.green.mask.size",   Dec);
  Hashtbl.add hash Gfieldpos   ("vbe.mode.green.field.pos",   Dec);
  Hashtbl.add hash Bmasksize   ("vbe.mode.blue.mask.size",    Dec);
  Hashtbl.add hash Bfieldpos   ("vbe.mode.blue.field.pos",    Dec);
  Hashtbl.add hash Amasksize   ("vbe.mode.alpha.mask.size",   Dec);
  Hashtbl.add hash Afieldpos   ("vbe.mode.alpha.field.pos",   Dec);
  Hashtbl.add hash Dcattr      ("vbe.mode.direct.color.attr", Hex);
  hash

module Fset = Set.Make(struct type t = field let compare = compare end)
module Vset = Set.Make(struct type t = (field * int) let compare = compare end)

let print_field ppf (field, v) =
  let name, ifmt =
    Hashtbl.find field_names field in
  match ifmt with
    Dec -> fprintf ppf "%-25s = %d" name v
  | Hex -> fprintf ppf "%-25s = %#x" name v

let print_mode ppf fields =
  let rec f = function
      field :: [] ->
        fprintf ppf "%a" print_field field
    | field :: rest ->
        fprintf ppf "%a\n" print_field field;
        f rest
    | [] -> ()
  in
  f (List.filter (fun (_, v) -> v != 0) (List.sort compare fields))

let fset_from_list list =
  List.fold_left (fun set field -> Fset.add field set) Fset.empty list

let vset_from_list list =
  List.fold_left (fun set field -> Vset.add field set) Vset.empty list

let all_fields_set = fset_from_list all_fields

type layout =
    RGBA8888
  | BGRA8888
  | ARGB8888
  | ABGR8888
  | RGB888
  | BGR888
  | RGB555
  | BGR555
  | RGB565
  | BGR565
  | TEXT
  | INDEX4
  | INDEX8

let all_layouts =
  [RGBA8888;
   BGRA8888;
   ARGB8888;
   ABGR8888;
   RGB888;
   BGR888;
   RGB555;
   BGR555;
   RGB565;
   BGR565;
   TEXT;
   INDEX4;
   INDEX8]

let elts_of_layout layout =
    match layout with
      RGBA8888 ->
        32,
        8, 0,
        8, 8,
        8, 16,
        8, 24
    | BGRA8888 ->
        32,
        8, 16,
        8, 8,
        8, 0,
        8, 24
    | ARGB8888 ->
        32,
        8, 8,
        8, 16,
        8, 24,
        8, 0
    | ABGR8888 ->
        32,
        8, 24,
        8, 16,
        8, 8,
        8, 0
    | RGB888 ->
        24,
        8, 0,
        8, 8,
        8, 16,
        0, 0
    | BGR888 ->
        24,
        8, 16,
        8, 8,
        8, 0,
        0, 0
    | RGB555 ->
        15,
        5, 0,
        5, 5,
        5, 10,
        0, 0
    | BGR555 ->
        15,
        5, 10,
        5, 5,
        5, 0,
        0, 0
    | RGB565 ->
        16,
        5, 0,
        6, 5,
        5, 11,
        0, 0
    | BGR565 ->
        16,
        5, 11,
        6, 5,
        5, 0,
        0, 0
    | TEXT ->
        8,
        0, 0,
        0, 0,
        0, 0,
        0, 0
    | INDEX4 ->
        4,
        0, 0,
        0, 0,
        0, 0,
        0, 0
    | INDEX8 ->
        8,
        0, 0,
        0, 0,
        0, 0,
        0, 0

let make_mode xres yres xchar ychar layout =
  let bpp, rm, rf, gm, gf, bm, bf, am, af =
    elts_of_layout layout
  in
  let bytes_per_pixel = (succ bpp) / 8 in
  let mode_attr =
    hw_sup lor color_mode lor vga_compatible_memory_mode_sup lor lfb_sup lor
    (match layout with
      TEXT -> 0
    | _ -> graphics_mode)

  in
  let imgplanes =
     match layout with
       TEXT -> 0
     | INDEX4 -> 4
     | _ -> 1
  in
  let memmodel =
     match layout with
       INDEX4 -> planar_model
     | INDEX8 -> packed_pixel_model
     | TEXT -> text_model
     | _ -> direct_color_model
  in
  [ (Mattr, mode_attr);
    (Xres, xres);
    (Yres, yres);
    (Xchar, xchar);
    (Ychar, ychar);
    (Bpp, bpp);
    (Pitch, bytes_per_pixel * xres * (if layout == TEXT then xchar else 1));
    (Rfieldpos, rf);
    (Rmasksize, rm);
    (Gfieldpos, gf);
    (Gmasksize, gm);
    (Bfieldpos, bf);
    (Bmasksize, bm);
    (Afieldpos, af);
    (Amasksize, am);
    (Imageplanes, imgplanes);
    (Memmodel, memmodel);
  ]

let modes : (int * (field * int) list) list =
  let nsm = ref 0x80 in
  let nm () =
    incr nsm;
    !nsm - 1
  in

  let make_modes n xres yres layout_list =
    let n, lrev =
      List.fold_left begin fun (n, list) layout ->
        succ n, (n, make_mode xres yres 0 0 layout) :: list
      end (n, []) layout_list
    in
    n, List.rev lrev
    (*List.map (fun layout -> make_mode xres yres 0 0 layout) layout_list*)
  in

  let hard_modes =
    [0x100, make_mode  640  400 0  0 INDEX8;
     0x101, make_mode  640  480 0  0 INDEX8;

     0x102, make_mode  800  600 0  0 INDEX4;
     0x103, make_mode  800  600 0  0 INDEX8;

     0x104, make_mode 1024  768 0  0 INDEX4;
     0x105, make_mode 1024  768 0  0 INDEX8;

     0x106, make_mode 1280 1024 0  0 INDEX4;
     0x107, make_mode 1280 1024 0  0 INDEX8;

     0x108, make_mode   80   60 8 16 TEXT;
     0x109, make_mode  132   25 8 16 TEXT;
     0x10a, make_mode  132   43 8 16 TEXT;
     0x10b, make_mode  132   50 8 16 TEXT;
     0x10c, make_mode  132   60 8 16 TEXT;
   ]
  in
  let a = [RGB555; RGB565; RGB888] in
  let b =
    if !use_bgr then
      [BGR555; BGR565; BGR888; ARGB8888; ABGR8888; RGBA8888; BGRA8888]
    else
      [ARGB8888; RGBA8888]
  in
  let c = a @ b in

  let n = 0x10d in
  let n, sm320x200 = make_modes n 320 200 a in
  let n, sm640x480 = make_modes n 640 480 a in
  let n, sm800x600 = make_modes n 800 600 a in
  let n, sm1024x768 = make_modes n 1024 768 a in
  let n, sm1280x1024 = make_modes n 1280 1024 a in
  let n, sm1600x1200 = make_modes 0x120 1600 1200 a in

  (*let n = 0x80 in*)
  let n, m320x200   = make_modes n 320 200 b in
  let n, m320x240   = make_modes n 320 240 c in
  let n, m320x400   = make_modes n 320 400 c in
  let n, m512x384   = make_modes n 512 384 c in
  let n, m640x400   = make_modes n 640 400 c in
  let n, m640x480   = make_modes n 640 480 b in
  let n, m800x600   = make_modes n 800 600 b in
  let n, m1024x768  = make_modes n 1024 768 b in
  let n, m1280x1024 = make_modes n 1280 1024 b in
  let n, m1600x1200 = make_modes n 1600 1200 b in

  hard_modes @
  sm320x200 @ sm640x480 @ sm800x600 @ sm1024x768 @ sm1280x1024 @ sm1600x1200 @

  m320x200 @ m320x240 @ m320x400 @ m512x384 @ m640x400 @
  m640x480 @ m800x600 @ m1024x768 @ m1280x1024 @ m1600x1200

let print_modes ppf list =
  let rec f = function
      (nmode, mode) :: [] ->
        fprintf ppf "vbe.mode = %#x\n%a\n" nmode print_mode mode
    | (nmode, mode) :: rest ->
        fprintf ppf "vbe.mode = %#x\n%a\n\n" nmode print_mode mode;
        f rest
    | [] -> ()
  in
  f list

let _ =
  printf "%a\n" print_modes modes
