#include <stdio.h>
#include <stdlib.h>

#include <inc/int.h>
#include <inc/red.h>
#include <inc/dma.h>
#include <inc/dos.h>
#include <inc/eio.h>
#include <inc/aud.h>
#include <inc/bit.h>
#include <inc/pic.h>
#include <inc/gus.h>
#include <inc/log.h>
#include <inc/mup.h>

#include <gus/gusemu.h>

#define LOG_PREFIX "GUS"
static int log_level;

static struct {
  int port;
  int irq;
  int dma;
  int estimate;
  int mix_block;
  int hack;
} gus = {0x220, 5, 1, 0, 0, 0};

static int hack_count;
static int noirq;

static int mixlive, mixwpos, mixrpos;
static ub *prebuf;
static int prebufsize;
static int read_buf_end;

static ub *himem;

enum access_type {
  PORT_READ = 0,
  PORT_WRITE = 1
};

enum access_size {
  PORT_8BIT = 0,
  PORT_16BIT = 1
};

static ui read_gus (int nport, int size)
{
  /* EIO_LOG_READ; */

  _accesssize = size - 1;
  _accesstype = PORT_READ;
  _registerno = nport;

  gus_emu ();

  if (gus.hack
      && (0x305 == (nport & 0xf0f))
      && (hack_count == gus.hack)) {

    hack_count = 0;
    return _accvalue | 1;
  }
  else {
    if (hack_count)
      hack_count++;
    return _accvalue;
  }
}

static void write_gus (int nport, int size, ui val)
{
  /* EIO_LOG_WRITE; */

  if (gus.hack
      && (0x303 == (nport & 0xf0f))
      && (val == 0x8d)) {

    if (0 == hack_count)
      hack_count = 1;
  }
  else {
    hack_count = 0;
  }

  _accesssize = size - 1;
  _accesstype = PORT_WRITE;
  _registerno = nport;
  _accvalue = val;
  gus_emu ();
}

static void write_gus2 (int nport, int size, ui val)
{
  if (4 == size) {
    write_gus (nport, 2, 0xffff & (val >> 16));
    write_gus (nport, 2, 0xffff & val);
  }
  else write_gus (nport, size, val);
}

static int nirqs;
void GUS_gen_irqs (int _nirqs);
void GUS_gen_irqs (int _nirqs)
{
  if (0 == noirq) {
    ldebug ("requesting %d irqs\n", _nirqs);
    if (_nirqs) {
      PIC_request_service (gus.irq);
      nirqs = _nirqs - 1;
    }
  }
}

void GUS_start_DMA (void);
void GUS_start_DMA ()
{
  DMA_hold_DREQ (gus.dma);
}

static void mix (int size)
{
  int to_mix;

  to_mix = MIN (size, prebufsize - mixlive);

  while (to_mix) {
    int re_mix_now, al_mix_now;
    int left;

    ldebug ("mix size %d mix %d live %d wpos %d\n",
            size, to_mix, mixlive, mixwpos);

    left = prebufsize - mixwpos;
    re_mix_now = MIN (to_mix, left);
    al_mix_now = re_mix_now & ~3;

    if (al_mix_now) {
      mem_writeposition = (ui) (prebuf + mixwpos);
      mem_buffersize = al_mix_now;
      soundcalccallback ();
    }

    if (re_mix_now == left) {
      read_buf_end = mixwpos + al_mix_now;
      mixwpos = 0;
    }
    else
      mixwpos += al_mix_now;

    mixlive += al_mix_now;
    to_mix -= al_mix_now;
  }
  ldebug ("mixed\n");
}

static void play (int size)
{
  int to_play;

  to_play = size;

  while (to_play) {
    int left;
    int play_now;

    if (0 == mixlive) {
      mix (to_play);
    }

    ldebug ("play size %d play %d live %d rpos %d\n",
            size, to_play, mixlive, mixrpos);

    left = read_buf_end - mixrpos;
    play_now = MIN (left, to_play);
    play_now = MIN (play_now, mixlive);

    AUD_write ((void *) (prebuf + mixrpos), play_now);

    if (play_now == left) {
      mixrpos = 0;
    }
    else
      mixrpos += play_now;

    mixlive -= play_now;
    to_play -= play_now;
  }
  ldebug ("played %d live %d\n", size, mixlive);
}

void GUS_run ()
{
  int left;

  DMA_run ();
  AUD_run ();
  left = AUD_get_free ();

  if (gus.estimate) {
    if (left)
      play (left);
    else
      mix (gus.mix_block);
  }
  else {
    if (left) {
      mem_writeposition = (ui) prebuf;
      mem_buffersize = left;
      soundcalccallback ();
      AUD_write (prebuf, left);
    }
  }
}

static void pic_handler (int which)
{
  if (which) {
    if (nirqs) {
      PIC_request_service (gus.irq);
      nirqs -= 1;
    }
  }
  else {
    nirqs = 0;
  }
}

static int dma_misc_handler (int moo)
{
  DEREF (moo);
  return -1;
}

static int GUS_read_DMA (ui addr, int size, int *_irq)
{
  int mode;

  DEREF (_irq);
  mode = DMA_get_channel_mode (gus.dma);
  linfo ("addr %d size %d mode %d\n", addr, size, mode);

  VDD_DMA_ADDRESS (DMAdata) = addr;
  VDD_DMA_PAGE (DMAdata) = addr >> 16;
  VDD_DMA_COUNT (DMAdata) = size;
  VDD_DMA_MODE (DMAdata) = mode;
  DMAthread ();
  if (0 == ((mode >> 4) & 1)) {
    DMA_release_DREQ (gus.dma);
  }
  return size;
}

void GUS_pre_init ()
{
  static struct MUP_key keys[] = {
    {"port"      , TAG_INT , {.ip = &gus.port}},
    {"irq"       , TAG_INT , {.ip = &gus.irq}},
    {"dma"       , TAG_INT , {.ip = &gus.dma}},
    {"estimate"  , TAG_INT , {.ip = &gus.estimate}},
    {"mix.block" , TAG_INT , {.ip = &gus.mix_block}},
    {"hack"      , TAG_INT , {.ip = &gus.hack}},
  };

  static struct MUP_group group =
    {"gus.", LENOFA (keys), keys, NULL};

  MUP_add_group (&group);
}

static void GUSIRQ (INT_ARGS)
{
  DEREF (r);
}

void GUS_init ()
{
  int i;
  char env_var[256];

  log_level = MUP_get_log_level ("gus");

  himem = malloc (1024 * 1024 + 32 + 4096);

  if (NULL == himem) {
    ERRFail ("failed to get GUS memory");
  }

  noirq = MUP_get_int ("audio.no.irq");

  gusirq = gus.irq;
  gusdma = gus.dma;

  if (0 == gus.mix_block) {
    gus.mix_block = 1764;
  }

  if (0 == MUP_get_int ("audio.no.env")) {
    snprintf (env_var, 256, "ULTRASND=%3x,%d,3,%d,5",
              gus.port, gus.dma, gus.irq);
    DOS_put_env (env_var);
  }

  EIO_register_write_handler (gus.port, EIO_MASK_BWD, write_gus2);
  EIO_register_read_handler ((gus.port + 0x100) & 0xf00, EIO_MASK_BWD, read_gus);

  for (i = 6; i < 16; i++) {
    EIO_register_read_handler (gus.port + i, EIO_MASK_BWD, read_gus);
    EIO_register_write_handler (gus.port + i, EIO_MASK_BWD, write_gus2);
  }

  for (i = 0; i < 8; i++) {
    EIO_register_read_handler (gus.port + 0x100 + i, EIO_MASK_BWD, read_gus);
    EIO_register_write_handler (gus.port + 0x100 + i, EIO_MASK_BWD, write_gus2);
  }

  AUD_open (44100, 2, AUD_FMT_S16);

  prebufsize = AUD_get_buffer_size ();
  prebuf = malloc (prebufsize);
  if (NULL == prebuf) {
    ERRFail ("failed allocate %d bytes for GUS prebuf", prebufsize);
  }
  read_buf_end = prebufsize;

  INT_set (gus.irq > 8 ? (gus.irq - 8) + 0x70 : gus.irq + 8, GUSIRQ);

  PIC_register (gus.irq, pic_handler);
  DMA_register_channel (gus.dma, GUS_read_DMA, dma_misc_handler);

  himemaddr = himem;
  flatmemofs = himem;
  gusdatapos = himemaddr + 1024 * 1024 + 32;
  gusregofs = himemaddr + 1024 * 1024 + 32;
}
