/* -*- mode: c; c-basic-offset: 4 -*- */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <inc/dpmi.h>
#include <inc/mem.h>
#include <inc/tid.h>
#include <inc/vid.h>
#include <inc/ldt.h>
#include <inc/pic.h>
#include <inc/sel.h>
#include <inc/dos.h>
#include <inc/kbd.h>
#include <inc/vga.h>
#include <inc/int.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "I21"
static int log_level = -2;

#define NULL_ ((void*)MEM_REAL(0))
#define PTR(a,b) ((char*)MEM_REAL((a)*16+((b)&0xffff)))
#define IS_DATA_BASED_NULL(x) (NULL_==x)

#define WARN_NULL(name,var) do {		\
    if (IS_DATA_BASED_NULL (var)) {		\
	warn_null (name);			\
	SET_AX (r, FORMAT_INVALID);		\
	goto error;				\
    }						\
} while (0)

static void warn_null (const char *cap)
{
    int z;
    size_t i;
    char *np;
    static char str[74];

    DEREF (cap);
    np = (char *) 0;
    i = 0;
    z = 0;

    while ((sizeof (str) - 1) > i) {
	if (0 == *np) {
	    z = 1;
	    if (0 == i) {
		strcpy (str, "<empty>");
		i = strlen (str);
	    }
	    break;
	}
	else if (isprint (*np)) {
	    str[i++] = *np;
	}
	else {
	    if ((sizeof (str) - 5) > i) {
		snprintf (str + i, 4, "\\x%2x", *np);
		i += 4;
	    }
	    else {
		break;
	    }
	}
    }
    str[i] = 0;

    lwarn ("%s called with NULL argument.\n"
	   "0=`%s'%s\n",
	   cap, str, z ? "" : "...");
}

void I21 (INT_ARGS)
{
    int func;
    static int cached;
    char *data;
    int res;

    if (-2 == log_level)
	log_level = MUP_get_log_level ("i21");

    func = AH (r);

    if ((0xb != func)
	&& (0x02 != func)
	&& (0x09 != func)
	&& (0x06 != func)
	&& (0x40 != func)
	&& (0x44 != func)
	&& (0x2c != func)
	&& (0x2a != func)) {

	if (log_level < LOG_DEBUG) {
	    linfo ("ax = %#06x ", AX (r));
	}
	else {
	    ldebug ("ax = %#06x ", AX (r));
	    INT_log_regs (r);
	}
    }

    res = 0;

 reswitch:
    switch (func) {
/* DOS 1+ - WRITE CHARACTER TO STANDARD OUTPUT */
	case 0x02:
	{
	    char c;

	    c = DL (r);
	    DOS_con_out (c);
	    write (2, &c, 1);
	    SET_AL (r, 0x09 == c ? 0x20 : c);
	    break;
	}

/* DOS 1+ - DIRECT CONSOLE OUTPUT */
/* --------D-2106--DLFF------------------------- */
/* DOS 1+ - DIRECT CONSOLE INPUT */
	case 0x06:
	{			/* XXX con in */
	    if (0xff != DL (r))
		DOS_con_out (DL (r));
	    else {
		break;
		/* UNHANDLED_INT (r); */
	    }
	    break;
	}

/* DOS 1+ - DIRECT CHARACTER INPUT, WITHOUT ECHO */
	case 0x07:
/* DOS 1+ - CHARACTER INPUT WITHOUT ECHO */
	case 0x08:
	{
	    int key, was_enabled;

	    linfo ("DOS 1+ - CHARACTER INPUT WITHOUT ECHO");
	    was_enabled = INT_is_enabled ();
	    INT_enable ();
	    key = DOS_con_in ();
	    cached = 0 == key;
	    SET_AL (r, key);
	    if (0 == was_enabled) INT_disable ();
	    break;
	}

/* DOS 1+ - WRITE STRING TO STANDARD OUTPUT */
	case 0x09:linfo ("DOS 1+ - WRITE STRING TO STANDARD OUTPUT");
	    data = PTR (DS (r), DX (r));
	    while ('$' != *data) {
		DOS_con_out (*data);
		write (1, data, 1);
		data++;
	    }
	    break;

/* DOS 1+ - GET STDIN STATUS */
	case 0x0b:linfo ("DOS 1+ - GET STDIN STATUS");
	{
	    int got;

	    got = cached || KBD_check_key ();
	    SET_AL (r, !got - 1);
	    break;
	}

/* DOS 1+ - FLUSH BUFFER AND READ STANDARD INPUT */
	case 0x0c:                  /* XXX */
	    linfo ("U DOS 1+ - FLUSH BUFFER AND READ STANDARD INPUT");
	    switch (AL (r)) {
		case 0x01:
		case 0x06:
		case 0x07:
		case 0x08:
		case 0x0a:
		    /* AH (r) = AL (r); */
		    func = AL (r);
		    goto reswitch;
	    }
	    break;

/* DOS 1+ - DISK RESET */
	case 0x0d:
	    break;

/* DOS 1+ - SELECT DEFAULT DRIVE */
	case 0x0e:
	    linfo ("DOS 1+ - SELECT DEFAULT DRIVE %d", DL (r));
	    DOS_set_drive (DL (r));
	    SET_AL (r, DOS_get_highest_drive ());
	    break;

/* DOS 1+ - CLOSE FILE USING FCB */
	case 0x10:
	    data = PTR (DS (r), DX (r));
	    linfo ("DOS 1+ - CLOSE FILE USING FCB %p\n", data);
	    goto error_ok;

/* DOS 1+ - GET CURRENT DEFAULT DRIVE */
	case 0x19:
	    linfo ("DOS 1+ - GET CURRENT DEFAULT DRIVE");
	    SET_AX (r, DOS_get_drive ());
	    break;

/* DOS 1+ - SET DISK TRANSFER AREA ADDRESS */
	case 0x1a:
	    data = PTR (DS (r), DX (r));
	    linfo ("DOS 1+ - SET DISK TRANSFER AREA ADDRESS %04x:%04x %p",
		   DS (r), DX (r), data);

	    if (NULL_ == data) {
		lwarn ("SET DISK TRANSFER AREA ADDRESS called with NULL\n");
		SET_AX (r, FORMAT_INVALID);
		goto error;
	    }

	    DOS_set_DTA (data);
	    break;

/* DOS 1+ - GET ALLOCATION INFORMATION FOR DEFAULT DRIVE */
	case 0x1b:
/* DOS 1+ - GET ALLOCATION INFORMATION FOR SPECIFIC DRIVE */
	case 0x1c:
	    linfo ("DOS 1+ - GET ALLOCATION INFORMATION FOR SPECIFIC DRIVE");
	    SET_AL (r, 1);            /* sectors per cluster */
	    SET_CX (r, 16384);        /* bytes per sector */
	    SET_DX (r, 16384);        /* total number of clusters on disk */
	    DS (r) = 0;               /* DS:BX media ID */
	    SET_BX (r, 0);
	    break;

/* DOS 1+ - SET INTERRUPT VECTOR */
	case 0x25:
	    linfo ("DOS 1+ - SET INTERRUPT VECTOR %#x %04x:%04x",
		   AL (r), DS (r), DX (r));
	    INT_set_rm (AL (r), DS (r), DX (r));
	    break;

/* DOS 1+ - CREATE NEW PROGRAM SEGMENT PREFIX */
	case 0x26:
	    linfo ("DOS 1+ - CREATE NEW PROGRAM SEGMENT PREFIX %04x", DX (r));
	    DOS_create_PSP (DX (r));
	    break;

	case 0x29:
	    lwarn ("U DOS 1+ - PARSE FILENAME INTO FCB");
	    goto error_ok;

/* DOS 1+ - GET SYSTEM DATE */
	case 0x2a:linfo ("DOS 1+ - GET SYSTEM DATE");
	{
	    struct TID_tm tm;

	    TID_get_tm (&tm);
	    SET_CX (r, tm.year);
	    SET_DH (r, tm.mon);
	    SET_DL (r, tm.mday);
	    SET_AL (r, tm.wday);
	    break;
	}

/* DOS 1+ - SET SYSTEM DATE */
	case 0x2b:
/* DOS 1+ - SET SYSTEM TIME */
	case 0x2d:
	    lwarn ("U DOS 1+ - SET SYSTEM DATE/TIME %#x\n", AX (r));
	    break;

/* DOS 1+ - GET SYSTEM TIME */
	case 0x2c:
	{
	    struct TID_tm tm;
	    TID_get_tm (&tm);
	    SET_CH (r, tm.hour);
	    SET_CL (r, tm.min);
	    SET_DH (r, tm.sec);
	    SET_DL (r, tm.nsec / (TID_NSEC / 100));
	    linfo ("DOS 1+ - GET SYSTEM TIME = %02d:%02d:%02d:%02d",
		   tm.hour, tm.min, tm.sec, DH (r));
	    break;
	}

/* GET DISK TRANSFER AREA ADDRESS */
	case 0x2f:
	{
	    ui dta;

	    dta = MEM_USER ((ui) DOS_get_DTA ());

	    ES (r) = dta / 16;
	    SET_BX (r, dta - ES (r) * 16);
	    linfo ("GET DISK TRANSFER AREA ADDRESS = %#x %04x:%04x",
		   dta, ES (r), BX (r));
	    break;
	}

/* GET DOS VERSION */
	case 0x30:
	    SET_AH (r, 0);
	    SET_AL (r, 6);
	    SET_BX (r, 0);
	    linfo ("GET DOS VERSION = %#x", AX (r));
	    break;

/* EXTENDED BREAK CHECKING */
	case 0x33:
	{
	    linfo ("U EXTENDED BREAK CHECKING");
	    /* SET_BX (r, 0x0006); */
	    /* SET_DX (r, 0); */
	    /* SET_AX (r, -1); */
	    break;
	}

/* GET ADDRESS OF INDOS FLAG */
	case 0x34:
	{
	    lwarn ("U GET ADDRESS OF INDOS FLAG\n");
	    ES (r) = 0x50;
	    SET_BX (r, 0);
	    break;
	}

/* GET INTERRUPT VECTOR */
	case 0x35:
	{
	    us seg, off;

	    INT_get_rm (AL (r), &seg, &off);
	    ES (r) = seg;
	    SET_BX (r, off);
	    linfo ("GET INTERRUPT VECTOR %#x = %04x:%04x",
		   AL (r), ES (r), BX (r));
	    break;
	}

/* GET FREE DISK SPACE */
	case 0x36:
	    SET_AX (r, 1);            /* sectors per cluster */
	    SET_BX (r, 16384);        /* number of free clusters */
	    SET_CX (r, 16384);        /* bytes per sector */
	    SET_DX (r, 16384);        /* total number of clusters on disk */
	    linfo ("GET FREE DISK SPACE = spc(%d) fc(%d) bps(%d) c(%d)",
		   AX (r), BX (r), CX (r), DX (r));
	    break;

/* GET COUNTRY-SPECIFIC INFORMATION */
	case 0x38:
	    lwarn ("U GET COUNTRY-SPECIFIC INFORMATION\n");
	    SET_AX (r, 2);
	    goto error_ok;

/* "MKDIR" - CREATE SUBDIRECTORY */
	case 0x39:
	    data = PTR (DS (r), DX (r));
	    WARN_NULL ("MKDIR", data);

	    linfo ("\"MKDIR\" - CREATE SUBDIRECTORY `%s'", data);
	    res = DOS_make_dir (data);
	    if (-1 == res) {
		lwarn ("MKDIR `%s' failed\n", data);
		goto error;
	    }
	    break;

/* "RMDIR" - REMOVE SUBDIRECTORY */
	case 0x3a:
	    data = PTR (DS (r), DX (r));
	    WARN_NULL ("RMDIR", data);

	    linfo ("\"RMDIR\" - REMOVE SUBDIRECTORY `%s'", data);
	    res = DOS_remove_dir (data);

	    if (-1 == res) {
		lwarn ("RMDIR for `%s' failed\n", data);
		goto error_code;
	    }
	    break;

/* "CHDIR" - SET CURRENT DIRECTORY */
	case 0x3b:
	    data = PTR (DS (r), DX (r));
	    WARN_NULL ("CHDIR", data);

	    linfo ("\"CHDIR\" - SET CURRENT DIRECTORY `%s'", data);
	    res = DOS_set_cur_dir (data);

	    if (-1 == res) {
		lwarn ("CHDIR for `%s' failed\n", data);
		goto error_code;
	    }
	    break;

/* "CREAT" - CREATE OR TRUNCATE FILE */
	case 0x3c:
	    data = PTR (DS (r), DX (r));
	    WARN_NULL ("CREAT", data);

	    linfo ("\"CREAT\" - CREATE OR TRUNCATE FILE `%s'", data);
	    res = DOS_create (data, CX (r));

	    if (-1 == res) {
		lwarn ("CREAT for `%s' failed\n", data);
		goto error_code;
	    }

	    SET_AX (r, res);
	    break;

/* "OPEN" - OPEN EXISTING FILE */
	case 0x3d:
	    data = PTR (DS (r), DX (r));
	    WARN_NULL ("OPEN", data);

	    linfo ("\"OPEN\" - OPEN EXISTING FILE `%s'", data);
	    res = DOS_open (data, AL (r) & 3);

	    if (res == -1) {
		lwarn ("OPEN for `%s' failed\n", data);
		goto error_code;
	    }

	    SET_AX (r, res);
	    break;

/* "CLOSE" - CLOSE FILE */
	case 0x3e:
	    linfo ("\"CLOSE\" - CLOSE FILE %d", BX (r));
	    DOS_close (BX (r));
	    break;

/* "READ" - READ FROM FILE OR DEVICE */
	case 0x3f:
	{
	    ui len;

	    len = CX (r);
	    data = PTR (DS (r), DX (r));
	    WARN_NULL ("READ", data);

	    linfo ("\"READ\" - READ  FROM FILE OR  DEVICE %d %d %p",
		   BX (r), len, data);
	    res = DOS_read (BX (r), data, len);

	    if (-1 == res) {
		lwarn ("fd=%d data=%p len=%d\n", BX (r), data, len);
		goto error_code;
	    }

	    SET_AX (r, res);
	    break;
	}

/* "WRITE" - WRITE TO FILE OR DEVICE */
	case 0x40:
	{
	    ui len;

	    len = CX (r);
	    data = PTR (DS (r), DX (r));
	    WARN_NULL ("WRITE", data);

	    linfo ("\"WRITE\" - WRITE TO FILE OR DEVICE %d %d %p",
		   BX (r), len, data);
	    res = DOS_write (BX (r), data, len);

	    if (-1 == res)
		goto error;

	    SET_AX (r, res);
	    break;
	}

/* "UNLINK" - DELETE FILE */
	case 0x41:
	    data = PTR (DS (r), DX (r));
	    WARN_NULL ("UNLINK", data);

	    linfo ("\"UNLINK\" - DELETE FILE `%s'", data);
	    lwarn ("UNLINK `%s'\n", data);
	    if (*data) {
		res = DOS_unlink (data);
	    }
	    else {
		lwarn ("UNLINK passed an empty string\n");
		res = -1;
	    }
	    if (-1 == res) {
		lwarn ("UNLINK for `%s' failed\n", data);
		goto error_code;
	    }
	    break;

/* "LSEEK" - SET CURRENT FILE POSITION */
	case 0x42:
	{
	    int pos;

	    pos = (int) MKDWORD (CX (r), DX (r));

	    linfo ("\"LSEEK\" - SET CURRENT FILE POSITION %d %d %d",
		   BX (r), pos, AL (r));
	    res = DOS_seek (BX (r), pos, AL (r));

	    if (res == -1)
		goto error_code;

	    SET_DX (r, res >> 16);
	    SET_AX (r, res);
	    break;
	}

/* GET FILE ATTRIBUTES */
	case 0x43:
	    data = PTR (DS (r), DX (r));

	    if (IS_DATA_BASED_NULL (data)) {
		warn_null ("GET FILE ATTRIBUTES");
		SET_AX (r, FORMAT_INVALID);
		goto error_ok;          /* FC/SPERM */
	    }

	    linfo ("GET FILE ATTRIBUTES `%s'", data);
	    res = DOS_get_file_attr (data);

	    if (-1 == res) {
		lwarn ("GET FILE ATTRIBUTES for `%s' failed\n", data);
		goto error_code;
	    }

	    SET_CX (r, res);
	    break;

/* IOCTL - GET DEVICE INFORMATION */
	case 0x44:
	    linfo ("IOCTL - GET DEVICE INFORMATION %d", AL (r));
	    switch (AL (r)) {
		case 0x00:
		{
		    res = DOS_ioctl_get_info (BX (r));

		    if (-1 == res)
			goto error;

		    SET_DX (r, res);
		    break;
		}

		case 0x01:
		    SET_CF (r);
		    break;

		case 0x08:
		    lwarn ("U DOS 3.0+ - IOCTL - CHECK IF BLOCK DEVICE REMOVABLE\n"
			   "reporting %d(%c) as fixed", BL (r), BL (r) + 'A');
		    SET_AX (r, 0);
		    break;

		case 0x09:
		    lwarn ("U DOS 3.1+ - IOCTL - CHECK IF BLOCK DEVICE REMOTE\n"
			   "DRIVE %d(%c)",
			   BL (r), BL (r) + 'a');
		    goto error_ok;

		case 0x0d:
		{
		    if ((3 == BL (r))
			&& (8 == CH (r))
			&& (0x66 == CL (r))) {

			data = PTR (DS (r), DX (r));
			((us *)(data))[0] = 0;
			((ui *)(data + 2))[0] = 0xdeadbeef;
			memcpy (data + 6, "NO NAME    ", 11);
			memcpy (data + 17, "FAT16   ", 8);
			break;
		    }
		    else {
		    }
		}

		case 0xe:
		    lwarn ("U DOS 3.2+ - IOCTL - GET LOGICAL DRIVE MAP\n");
		    goto error_ok;

		default:
		    UNHANDLED_INT (r);
	    }
	    break;

/* "DUP" - DUPLICATE FILE HANDLE */
	case 0x45:
	    linfo ("\"DUP\" - DUPLICATE FILE HANDLE %d", BX (r));
	    res = DOS_dup_handle (BX (r));

	    if (-1 == res)
		goto error_code;

	    SET_AX (r, res);
	    break;

/* "CWD" - GET CURRENT DIRECTORY */
	case 0x47:
	    data = PTR (DS (r), SI (r));
	    WARN_NULL ("CWD", data);

	    linfo ("\"CWD\" - GET CURRENT DIRECTORY %d %p", DL (r), data);
	    DOS_get_cur_dir (DL (r), data);
	    SET_AX (r, 0x100);
	    break;

/* ALLOCATE MEMORY */
	case 0x48:
	{
	    ui base;

	    linfo ("ALLOCATE MEMORY %d", BX (r));
	    res = DOS_alloc (BX (r), &base);
	    if (-1 == res) {
		SET_BX (r, DOS_largest_block ());
		goto error_code;
	    }
	    SET_AX (r, MEM_USER (base) >> 4);
	    break;
	}

/* FREE MEMORY */
	case 0x49:
	    linfo ("FREE MEMORY %#x", ES (r));
	    if (0 == ES (r))
		lwarn ("attempt to free zero segment\n");
	    else
		DOS_free (MEM_REAL (ES (r) << 4));
	    break;

/* RESIZE MEMORY BLOCK */
	case 0x4a:
	{
	    int lin, nparas;

	    lin = MEM_REAL (ES (r) << 4);
	    nparas = BX (r);

	    linfo ("RESIZE MEMORY BLOCK %#x %d", ES (r), nparas);
	    if (-1 == DOS_resize (lin, nparas)) {
		lwarn ("failed to resize dos mem %#x to %d paras\n",
		       lin, nparas);
		SET_BX (r, DOS_largest_block ());
		goto error_code;
	    }
	    break;
	}

/* "EXEC" - LOAD AND/OR EXECUTE PROGRAM */
	case 0x4b:
	    data = PTR (DS (r), DX (r));
	    lwarn ("U \"EXEC\" - LOAD AND/OR EXECUTE PROGRAM %#x\n"
		   "`%s'\n", AX (r), data);
	    break;

/* GET RETURN CODE (ERRORLEVEL) */
	case 0x4d:linfo ("GET RETURN CODE (ERRORLEVEL)");
	    /* 0 - ok, 1 - cc, 2 - crit err, 3 - tsr */
	    SET_AX (r, 2);
	    break;

/* "EXIT" - TERMINATE WITH RETURN CODE */
	case 0x4c:
	    linfo ("\"EXIT\" - TERMINATE WITH RETURN CODE %d", AL (r));
	    finished ("normal exit via INT 21H AH=4Ch");
	    break;

/* "FINDFIRST" - FIND FIRST MATCHING FILE */
	case 0x4e:
	    data = PTR (DS (r), DX (r));

	    linfo ("\"FINDFIRST\" - FIND FIRST MATCHING FILE `%s'", data);
	    WARN_NULL ("FINDFIRST", data);

	    res = DOS_find_first (data, CX (r));
	    if (-1 == res) {
		lwarn ("FINDFIRST for `%s' failed\n", data);
		goto error_code;
	    }
	    break;

/* "FINDNEXT" - FIND NEXT MATCHING FILE */
	case 0x4f:
	    linfo ("\"FINDNEXT\" - FIND NEXT MATCHING FILE");
	    res = DOS_find_next ();
	    if (-1 == res) {
		goto error_code;
	    }
	    break;

/* DOS 2+ internal - SET CURRENT PROCESS ID (SET PSP ADDRESS) */
	case 0x50:
	    linfo ("DOS 2+ internal - SET CURRENT PROCESS ID (SET PSP ADDRESS) %#x",
		   BX (r));
	    DOS_set_PSP (BX (r));
	    break;

/* DOS 2+ internal - GET CURRENT PROCESS ID (GET PSP ADDRESS) */
	case 0x51:
	    linfo ("DOS 2+ internal - GET CURRENT PROCESS ID (GET PSP ADDRESS)");
	    SET_BX (r, MEM_USER (DOS_get_PSP ()) >> 4);
	    break;

/* U - DOS 2+ internal - "SYSVARS" - GET LIST OF LISTS */
	case 0x52:
	{
#if 0
	    us es, bx;
#endif

	    lwarn ("U U - DOS 2+ internal - \"SYSVARS\" - GET LIST OF LISTS\n");
	    break;
#if 0
	    DOS_get_list_of_lists (&es, &bx);
	    ES (r) = es;
	    SET_BX (r, bx);
	    break;
#endif
	}

/* DOS 2+ internal - CREATE CHILD PSP */
	case 0x55:
	    linfo ("DOS 2+ internal - CREATE CHILD PSP %#x %#x",
		   DX (r), SI (r));
	    DOS_create_child_PSP (MEM_REAL (DX (r) << 4), SI (r));
	    break;

/* "RENAME" - RENAME FILE */
	case 0x56:
	{
	    char *oldpath, *newpath;

	    oldpath = PTR (DS (r), DX (r));
	    newpath = PTR (ES (r), DI (r));

	    linfo ("\"RENAME\" - RENAME FILE `%s' -> `%s'",
		   oldpath, newpath);
	    if (IS_DATA_BASED_NULL (oldpath) && IS_DATA_BASED_NULL (newpath)) {
		warn_null ("RENAME(both)");
		SET_AX (r, FORMAT_INVALID);
		goto error;
	    }
	    WARN_NULL ("RENAME(old)", oldpath);
	    WARN_NULL ("RENAME(new)", newpath);

	    res = DOS_rename (oldpath, newpath);
	    if (-1 == res) {
		lwarn ("RENAME failed for `%s'=`%s'\n", oldpath, newpath);
		goto error_code;
	    }
	    break;
	}

/* GET FILE'S LAST-WRITTEN DATE AND TIME */
	case 0x57:
	    linfo ("GET FILE'S LAST-WRITTEN DATE AND TIME");
	    SET_AX (r, 0);
	    break;

	case 0x58:
	    switch (AL (r)) {
		case 0x00:
		    lwarn ("U DOS 2.11+ - GET OR SET MEMORY ALLOCATION STRATEGY\n");
		    SET_AX (r, 0);
		    break;

		case 0x01:
		    lwarn ("U DOS 2.11+ - GET OR SET MEMORY ALLOCATION STRATEGY\n");
		    goto error_ok;

		case 0x02:
		    lwarn ("U DOS 5+ - GET OR SET UMB LINK STATE\n");
		    SET_AL (r, 0);
		    break;

		case 0x03:
		    lwarn ("U DOS 5+ - GET OR SET UMB LINK STATE\n");
		    goto error_ok;

		default:
		    UNHANDLED_INT (r);
	    }
	    break;

/* DOS 3.0+ - "TRUENAME" - CANONICALIZE FILENAME OR PATH */
	case 0x60:
	{
	    char *in_path, *out_path;

	    in_path = PTR (DS (r), SI (r));
	    out_path = PTR (ES (r), DI (r));

	    lwarn ("U DOS 3.0+ - \"TRUENAME\" - CANONICALIZE FILENAME OR PATH\n"
		   "in %04x:%04x `%s'\n"
		   "out %04x:%04x",
		   DS (r), SI (r), in_path,
		   ES (r), DI (r));

	    strcpy (out_path, in_path);
	    break;
	}

/* DOS 3.0+ - GET CURRENT PSP ADDRESS */
	case 0x62:
	    linfo ("DOS 3.0+ - GET CURRENT PSP ADDRESS");
	    SET_BX (r, MEM_USER (DOS_get_PSP ()) >> 4);
	    break;

/* DOS 3.2+ - GET DOUBLE UB CHARACTER SET LEAD-BYTE TABLE */
	case 0x63:
	    lwarn ("U DOS 3.2+ - GET DOUBLE UB CHARACTER SET LEAD-BYTE TABLE\n");
	    SET_AL (r, 0xff);
	    goto error_ok;

/* DOS 3.3+ - GET EXTENDED COUNTRY INFORMATION */
	case 0x65:
	    lwarn ("U DOS 3.3+ - GET EXTENDED COUNTRY INFORMATION %#x\n", AX (r));
	    goto error_ok;

/* DOS 3.3+ - "FFLUSH" - COMMIT FILE */
	case 0x68:
	    lwarn ("U DOS 3.3+ - \"FFLUSH\" - COMMIT FILE\n");
	    break;

/* Rational Systems DOS/4G - INSTALLATION CHECK */
	case 0xff:
	    goto error;

	default:
	    UNHANDLED_INT (r);
    }

    linfo (" = %d\n", res);
    CLR_CF (r);
    return;

 error:
    UNHANDLED_INT (r);

 error_code:
    lwarn ("%#x completed with error code %d\n", func, DOS_get_error ());
    SET_AX (r, DOS_get_error ());
    SET_CF (r);
    return;

 error_ok:
    lwarn ("%#x completed with error possible code %d, %d\n",
	   func, AX (r), DOS_get_error ());
    SET_CF (r);
}
