open Zz
open Types
open Format

exception Bad_operand_index of int
exception Bad_spec of (int * string)
exception Write_to of zz

let reject_list =
  ["[GE]m"; "db"; ".byte";
   "^b[ts][srcf ]"; "[ls]\\(msw\\)";
   "[ls]\\([gli]dt\\)"; "invlpg"; "wbinvd"; "^rd";
   "^wr"; "umov"; "loadall"; "emms"; "esc";
   "xadd"; "ud2"; "ver"; "ltr"; "rsm"; "cmpxchg8b";
   "cpuid"; "clts"]

let string_of_size = function
    S8 -> "8" | S16 -> "16" | S32 -> "32"

let gcc_regs = ref false
let speed = ref true

let indent = 4
let snil = ""
let eff_lines = ref false
let fprintf_debug = ref false
let glob_lineno = ref 0
let each_cc = ref false
let each_reg_bwd = ref false
let singlerm = ref false
let _W_re = Str.regexp "~W"
let _O_re = Str.regexp "~O"
let _B_re = Str.regexp "~B"
let _D_re = Str.regexp "~D"
let _T_re = Str.regexp "~T"
let _V_re = Str.regexp "~V"
let _A_re = Str.regexp "~A"
let _X_re = Str.regexp "~X"
let _Q_re = Str.regexp "~Q"
let _C_re = Str.regexp "~C"
let _N_re = Str.regexp "~N"
let eff_name = ref "eff.tab"
let opc_name = ref "opcod.tab"
let commentre = Str.regexp ";.*$"
let trailwsre = Str.regexp " +$"
let traildashre = Str.regexp "[ -]+$"
let splitre = Str.regexp "[ \t]+"
let groupre = Str.regexp "\\(group\\) +\\([0-9]+\\)"
let prefixre = Str.regexp "\\(prefix\\) +\\([^ ]+\\)"
let specre = Str.regexp
    ("[rw]\\([123f]\\|\\(e?[abcds][ipxlh]\\)\\)\\(:\\([us][bwd]?\\)\\)?" ^
     "\\(=\\([a-zA-Z_][a-zA-Z_0-9]*\\)\\)?$")

let _ =
  let set_names s =
    eff_name := Filename.concat s "eff.tab";
    opc_name := Filename.concat s "opcod.tab"
  in
  let flip r () = r := not !r in
  Arg.parse
    ["-d", Arg.String set_names, "";
     "-fprintf_debug", Arg.Unit (flip fprintf_debug), "";
     "-eff_lines", Arg.Unit (flip eff_lines), "";
     "-ecc", Arg.Unit (flip each_cc), "";
     "-ereg", Arg.Unit (flip each_reg_bwd), "";
     "-srm", Arg.Unit (flip singlerm), "";
     "-confirm", Arg.Unit (flip Pp.confirm), ""]
    (fun s -> eprintf "dont know what to do with %s@." s)
    "Usage: iconvert [options]"

(*let fmt_out s p n =*)
(*  for i = p to pred (p + n) do*)
(*    if s.[i] = '\n' then incr glob_lineno;*)
(*    print_char s.[i]*)
(*  done*)

(*let fmt_flush = print_newline*)

(*let _ =*)
(*  Format.set_formatter_output_functions fmt_out fmt_flush*)


let ccs = [|
  "(cc & OF)";
  "(0 == (cc & OF))";
  "(cc & CF)";
  "(0 == (cc & CF))";
  "(cc & ZF)";
  "(0 == (cc & ZF))";
  "(cc & (CF | ZF))";
  "(0 == (cc & (CF | ZF)))";
  "(cc & SF)";
  "(0 == (cc & SF))";
  "(cc & PF)";
  "(0 == (cc & PF))";
  "((cc & SF) ^ ((cc & OF) >> (OF_SHIFT - SF_SHIFT)))";
  "(0 == ((cc & SF) ^ ((cc & OF) >> (OF_SHIFT - SF_SHIFT))))";
  "((cc & ZF) | ((cc & SF) ^ ((cc & OF) >> (OF_SHIFT - SF_SHIFT))))";
  "(0 == ((cc & ZF) | ((cc & SF) ^ ((cc & OF) >> (OF_SHIFT - SF_SHIFT)))))";
|]

let is_cc mnemonic =
  match mnemonic with
    "setcc" | "cmovcc" | "jcc" -> true
  | _ -> false

let has_reg_bwd zz_list =
  List.exists (function Regb | Regv -> true | _ -> false) zz_list

type operand = zz * size

type lcode = int * string list
type path =
    Pany of lcode
  | Pspec of (lcode * lcode)

type decl1 =
    Doper of (int * sign option * size option)
  | Dreg of reg * sign option * size option
  | Dflag
  | Deip

type decl = decl1 * string

type def =
    Def of (string * zz list)
  | Group of int
  | Empty

let defs = Array.create (512 + 25 * 8) Empty
let prefixes = Array.create 20 None
let nprefixes = ref 0
let decl_hash = Hashtbl.create 10

let size_of_v = function
    Vw -> S16
  | Vd -> S32

type group =
    ((dep * zz list * (var * string) list * (var * string) list)
       * ((int * string * path) list))

type mapval =
    I of group
  | S of (group * group)

let key_of_zz_list zz_list =
  String.concat "" (List.map Zz.name zz_list)

let pp_zz_list ppf zz_list =
  fprintf ppf "%s" (String.concat "," (List.map Zz.name zz_list))

let pp_spec_zz_list ppf (dep, zz_list) =
  let rec pp = function
      zz :: [] -> fprintf ppf "%s" (Zz.spec_name dep zz)
    | zz :: zzs ->
	fprintf ppf "%s, " (Zz.spec_name dep zz);
	pp zzs
    | [] -> ()
  in
  pp zz_list

let iter_f ~get_line ~f ~init =
  let rec g accu =
    let line = get_line () in
    match line with
      Some line ->
	let line = Str.global_replace commentre "" line in
	let line = Str.global_replace trailwsre "" line in
	let accu = f accu line in
	g accu
    | None -> ()
  in
  g init

let parse_opcodes chan =
  let parse s =
    let fields = Str.split splitre s in
    match fields with
      mnemonic :: operands -> mnemonic, List.map Zz.parse operands
    | [] -> assert false in

  let get_line () =
    try
      Some (input_line chan)
    with End_of_file ->
      None in

  let rejectre =
    let re = String.concat "\\)\\|\\(" reject_list in
    Str.regexp ("\\(" ^ re ^ "\\)") in

  let reject line =
    try
      let _ = Str.search_forward rejectre line 0 in (); true
    with Not_found ->
      false in

  iter_f ~init:0 ~get_line ~f:begin fun i line ->
    if line = snil
    then i
    else
      if reject line then (succ i)
      else begin
	let line = Str.global_replace traildashre "" line in
	if line = snil
	then (succ i)
	else begin
	  if Str.string_match groupre line 0 then begin
	    let ngroup = int_of_string (Str.matched_group 2 line) in
	    defs.(i) <- Group ngroup
	  end else if Str.string_match prefixre line 0 then begin
	    prefixes.(!nprefixes) <- Some (i, (Str.matched_group 2 line));
	    incr nprefixes
	  end else
	    defs.(i) <- Def (parse line);
	  succ i
	end
      end
  end

let parse_descr chan =
  let lineno = ref 0 in
  let get_line () =
    try
      let line = input_line chan in
      incr lineno;
      Some line
    with End_of_file ->
      None in

  let parse_vars specs =

    let parse_var is_read s =
      if not (Str.string_match specre s 0) then raise (Bad_spec (!lineno, s));
      let orf = Str.matched_group 1 s in

      let get_sign_size () =
	try
	  let t = Str.matched_group 4 s in
	  let sign = Some (if t.[0] = 'u' then Iunsigned else Isigned) in
	  let size =
	    if String.length t = 1
	    then None
	    else
	      match t.[1] with
		'b' -> Some S8
	      | 'w' -> Some S16
	      | 'd' -> Some S32
	      | _ -> assert false
	  in
	  sign, size
	with Not_found ->
	  None, None
      in

      let get_name name =
	try
	  Str.matched_group 6 s
	with Not_found ->
	  name
      in

      let pref = if is_read then "r" else "w" in

      match orf.[0] with
	'f' -> Dflag, get_name (pref ^ "F")
      | '1' | '2' | '3' ->
	  let pos = Char.code orf.[0] - Char.code '1' in
	  let sign, size = get_sign_size () in
	  Doper (pos, sign, size), get_name (pref ^ string_of_int (pos + 1))
      | _ ->
	  let reg = Reg.parse orf in
	  let sign, usize = get_sign_size () in
	  Dreg (reg, sign, usize), get_name (pref ^ Reg.name reg) in

    let parse_spec s =
      match s.[0] with
	'r' -> 0, parse_var true s
      | 'w' -> 1, parse_var false s
      | _ -> raise (Bad_spec (!lineno, s)) in

    let sort (rw1, (a, _)) (rw2, (b, _)) =
      if rw1 = 0 && rw2 = 1
      then -1
      else
	match a, b with
	  Doper (pos1, _, _), Doper (pos2, _, _) ->
	    compare pos2 pos1
	| _, _ -> 0 in

    List.sort sort (List.map parse_spec specs)
  in

  let rec g stop lines =
    let rec g lines = function
	Some line ->
	  if line = stop then List.rev lines
	  else
	    g (line :: lines) (get_line ())
      | None ->
	  failwith "malformed specs"
    in
    g lines
  in

  let gi line =
    if line.[0] != '{' then raise (Bad_spec (!lineno -  1, line));
    let lineno = !lineno in
    lineno, g "}" [] (get_line ()) in

  let gd line =
    if line.[0] != '[' then raise (Bad_spec (!lineno, line));
    let lineno = !lineno in
    lineno, g "]" [] (get_line ()) in

  let rec f specs = function
      Some line ->
	let specs =
	  try
	    let line = Str.global_replace commentre "" line in
	    if line = snil then specs else
	    if line.[0] = '{' then begin
	      let code = gi line in
	      List.iter begin fun (key, rl, wl) ->
		Hashtbl.add decl_hash key (rl, wl, Pany code)
	      end specs;
	      []
	    end else if line.[0] = '[' then begin
	      let code_mem = gd line in
	      let mline = get_line () in
	      begin match mline with
		Some line ->
		  let code_reg = gd line in
		  List.iter begin fun (key, rl, wl) ->
		    eprintf "spec code for key %s@." key;
		    Hashtbl.add decl_hash key (rl, wl,
					       Pspec (code_mem, code_reg))
		  end specs;
	      | None -> raise (Bad_spec (!lineno -  1, line))
	      end;
	      []
	    end else if line.[0] = '>' then begin
	      let rec skip () =
		let line = get_line () in
		match line with
		  Some line ->
		    if line = "<" then () else skip ()
		| None -> failwith "unterminated preprocessing directive"
	      in
	      match line with
		"> speed" ->
		  if not !speed then skip ();
		  specs
	      | "> size" ->
		  if !speed then skip ();
		  specs
	      | otherwise ->
		  failwith "bad preprocessing directive"
	    end else
	      begin
		let fields = Str.split splitre line in
		let spec =
		  match fields with
		    key :: decl ->
		      let vars = parse_vars decl in

		      let foldf fb = fun accu (b, var) ->
			if b = fb then var :: accu else accu
		      in

		      let rl = List.fold_left (foldf 0) [] vars
		      and wl = List.fold_left (foldf 1) [] vars in
		      (key, rl, wl)
		  | [] ->
		      assert false
		in
		spec :: specs
	      end
	  with e ->
	    eprintf "bad spec line: `%s'@." line;
	    raise e
	in
	f specs (get_line ())
    | None -> ()
  in

  f [] (get_line ())

let pp_if_else cond ppf pp_if_else =
  pp_open_vbox ppf indent;
  pp_print_string ppf "if (";
  pp_print_string ppf cond;
  pp_print_string ppf ") {";
  pp_print_cut ppf ();
  pp_if_else ppf ~if_clause:true;
  pp_close_box ppf ();
  pp_print_cut ppf ();
  pp_print_string ppf "}";
  pp_print_cut ppf ();
  pp_open_vbox ppf indent;
  pp_print_string ppf "else {";
  pp_print_cut ppf ();
  pp_if_else ppf ~if_clause:false;
  pp_close_box ppf ();
  pp_print_cut ppf ();
  pp_print_string ppf "}"

let pp_if_else2 cond ppf pp_if_else =
  pp_open_vbox ppf indent;
  pp_print_string ppf "if (";
  pp_print_string ppf cond;
  pp_print_string ppf ") {";
  pp_print_cut ppf ();
  pp_if_else ppf true;
  pp_close_box ppf ();
  pp_print_cut ppf ();
  pp_print_string ppf "}";
  pp_print_cut ppf ();
  pp_open_vbox ppf indent;
  pp_print_string ppf "else {";
  pp_print_cut ppf ();
  pp_if_else ppf false;
  pp_close_box ppf ();
  pp_print_cut ppf ();
  pp_print_string ppf "}"

let pp_block ppf (pp, fmt) =
  (*pp_print_string ppf "{";*)
  pp_open_vbox ppf indent;
  pp_print_cut ppf ();
  pp ppf fmt;
  pp_close_box ppf ();
  pp_print_cut ppf ();
  pp_print_cut ppf ()
  (*pp_print_string ppf "}"*)

let pp_modrm ppf in_group =
  if not in_group then
    fprintf ppf "modrm = FETCH_BYTE (env);@,"

let cast_string size sign usize =
  if size = usize && sign = Iunsigned then snil
  else
    match sign, size with
      Isigned, S8 -> "(SBYTE) "
    | Isigned, S16 -> "(SWORD) "
    | Isigned, S32 -> snil
    | _, _ -> snil

let pp_slist ppf l =
  let rec pp = function
      s :: [] -> fprintf ppf "%s" s
    | s :: rest -> fprintf ppf "%s,@ " s; pp rest
    | [] -> ()
  in
  pp l

let pp_alist pp_a ppf l =
  let rec pp = function
      x :: [] -> fprintf ppf "%a" pp_a x
    | x :: rest -> fprintf ppf "%a,@ " pp_a x; pp rest
    | [] -> ()
  in
  pp l

let pp_alist2 pp_a ppf l =
  let rec pp = function
      x :: [] -> fprintf ppf "%a" pp_a x
    | x :: rest -> fprintf ppf "%a@ " pp_a x; pp rest
    | [] -> ()
  in
  pp l

let pp_code ppf l =
  let rec pp = function
      s :: [] -> fprintf ppf "%s" s
    | s :: rest -> fprintf ppf "%s@," s; pp rest
    | [] -> ()
  in
  pp l

let key_of_var = function
  | Operand (pos, zz, size, cast) ->
      let sign, size =
	match cast with
	  Cast (sign, size) -> sign, size
	| Raw -> Iunsigned, size
      in
      if Zz.dsensitive zz then
	sprintf "%d:%s:%c:v"
	  pos (Zz.name zz) (char_of_sign sign)
      else
	sprintf "%d:%s:%c:%c"
	  pos (Zz.name zz) (char_of_sign sign) (char_of_size size)
  | Reg (reg, cast) ->
      let sign, size =
	match cast with
	  Cast (sign, size) -> sign, size
	| Raw -> Iunsigned, Reg.size reg
      in
      sprintf "%s:%c:%c"
	(Reg.name reg) (char_of_sign sign) (char_of_size size)
  (*| _ -> snil*)
  | Flag -> "f"
  | EIP -> "i"

let eq_var_lists ll lr =
  let compare_vars vl vr =
    match vl, vr with
    | Operand(posl, zzl, sizel, castl),
      Operand(posr, zzr, sizer, castr) ->
	if zzl = zzr then begin
	  if Zz.dsensitive zzl then
	    let sign_cc =
	      match castl, castr with
		Cast (sl, _), Cast (sr, _) -> sl = sr
	      | Raw, Raw -> true
	      | otherwise -> false
	    in
	    posl = posr
	  else
	    posl = posr && castl = castr
	end else false
    | Reg(regl, castl), Reg(regr, castr) ->
	  regl = regr && castl = castr
    | Flag, Flag -> true
    | EIP, EIP -> true
    | otherwise -> false
  in
  let compl vl vr = not (compare_vars vl vr) in
  if List.length ll = List.length lr then
    let rlr =
      List.fold_left begin fun rlr vl ->
	List.filter (compl vl) rlr
      end lr ll
    in
    rlr = []
  else false

module Sset = Set.Make (struct type t = string let compare = compare end)
module Iset = Set.Make (struct type t = int let compare = compare end)
module Smap = Map.Make (struct type t = string let compare = compare end)

let dep_size zz = function
    Indep -> Zz.size zz
  | Dep v ->
      begin match v with
	Vw -> S16
      | Vd -> S32
      end

type insn = string * zz list * var list * var list * string list
type gclass =
    Gsensetive of (v * insn list)
  | Ginsensetive of insn list

let make_insn mnemonic zz_list dep =
  let key_mnemonic =
    Zz.specialize_mnemonic mnemonic zz_list (dep = Dep Vd) in
  let spec_zz_list =
    match dep with
      Dep v ->
	let spec_zz zz =
	  let zz_v = Zz.specialize_operand v zz in
	  zz, zz_v
	in
	List.map spec_zz zz_list
    | Indep -> List.map (fun zz -> zz, Indep) zz_list
  in
  let spec_name (zz, dep) = Zz.spec_name dep zz in
  let key =
    key_mnemonic ^ String.concat "" (List.map spec_name spec_zz_list) in

  let (rl, wl, code) = Hashtbl.find decl_hash key in
  let convert (decl1, name) =
    let var =
      match decl1 with
	Doper (pos, sign, size) ->
	  let zz, dep = List.nth spec_zz_list pos in
	  let zz_size = dep_size zz dep in
	  let cast =
	    match sign, size with
	      Some sign, None -> Cast (sign, zz_size)
	    | Some sign, Some size -> Cast (sign, size)
	    | otherwise -> Raw
	  in
	  Operand (pos, zz, zz_size, cast)
      | Dreg (reg, sign, size) ->
	  let reg_size = Reg.size reg in
	  let cast =
	    match sign, size with
	      Some sign, None -> Cast (sign, reg_size)
	    | Some sign, Some size -> Cast (sign, size)
	    | otherwise -> Raw
	  in
	  Reg (reg, cast)
      | Dflag -> Flag
      | Deip -> EIP
    in
    var, name
  in
  let rvars = List.map convert rl
  and wvars = List.map convert wl in

  let insn = mnemonic, zz_list, rvars, wvars, code in
  insn

let var_reads_modrm (var, _) =
  match var with
    Operand(_, zz, _, _) -> Zz.has_modrm zz
  | otherwise -> false

let var_neads_rm (var, _) =
  match var with
    Operand(_, zz, _, _) -> Zz.needs_rm zz
  | otherwise -> false

let var_fetches (var, _) =
  match var with
    Operand(_, zz, _, _) -> Zz.fetches zz
  | otherwise -> false

let var_all prefix accu (var, vname) =
  match var with
    Operand (pos, zz, size, cast) ->
      (*let s = prefix ^ (string_of_int (pos + 1)) in*)
      let s = vname in
      let accu =
	match zz with
	| CL | Regb | Gb | Ib | Jb | Const1 | AL | Eb ->
	    (s, ctype_code size cast) :: accu
	| Rd | Td | Cd | Dd -> (s, ctype_code S32 Raw) :: accu
	| DX | Rw | Ew | Sw | Iw -> (s, ctype_code size cast) :: accu
	| M | Ov | Ev | Fv | Regv
	| ES | CS | DS | SS | FS | GS
	| Iv | Jv | Gv | Ob | EAX ->
	    (s, ctype_code size cast) :: accu
	| Ma ->
	    (s ^ "_lo", ctype_code size Raw) ::
	    (s ^ "_hi", ctype_code size Raw) :: accu
	| Mp | Ap | Ep ->
	    (s ^ "_seg", ctype_code S16 Raw) ::
	    (s ^ "_off", ctype_code size Raw) :: accu
	| Xb | Yb | Xv | Yv | Ms -> accu
      in accu
  | Reg (reg, cast) ->
      (vname, ctype_code (Reg.size reg) cast) :: accu
      (*(prefix ^ Reg.name reg, ctype_code (Reg.size reg) cast) :: accu*)
  | Flag ->
      (vname, ctype_code S32 Raw) :: accu
      (*(prefix ^ "F", ctype_code S32 Raw) :: accu*)
  | EIP ->
      (vname, ctype_code S32 Raw) :: accu
      (*(prefix ^ "EIP", ctype_code S32 Raw) :: accu*)

let pp_var ~rm_state ~is_read ~assume ~rm_is_reg ppf (var, vname) =
  let inv = assume = Anone or assume = Afree in
  match var with
  | Operand op ->
      fprintf ppf "%a" (Pp.pp ~rm_state ~assume ~is_read ~rm_is_reg) (op, vname)
  | Flag ->
      if inv then begin
	if is_read then begin
	  Pp.pp_decl ppf ~name:vname ~size:S32 ~cast:Raw ~is_read;
	  fprintf ppf " = READ_DWORD_FLAGS (env);"
	end else
	  fprintf ppf "WRITE_DWORD_FLAGS (env, %s);" vname
      end
  | EIP -> assert false;
      if is_read then
	fprintf ppf "%s = READ_EIP (env);" vname
      else
	fprintf ppf "%s = WRITE_EIP (env);" vname
  | Reg (reg, cast) ->
      if inv then begin
	let name = Reg.name reg in
	let cast = Pp.string_of_cast (Reg.size reg) cast in
	if is_read then begin
	  Pp.pp_decl ppf ~name:vname ~size:S32 ~cast:Raw ~is_read
	end;
	begin match reg with
	  Reg8 n ->
	    if is_read then
	      fprintf ppf " = %sREAD_BYTE_REG (env, _REG_%s);" cast name
	    else
	      fprintf ppf "WRITE_BYTE_REG (env, _REG_%s, %s);" name vname
	| Reg16 n ->
	    if is_read then
	      fprintf ppf " = %sREAD_WORD_REG (env, _REG_%s);" cast name
	    else
	      fprintf ppf "WRITE_WORD_REG (env, _REG_%s, %s);" name vname
	| Reg32 n ->
	    if is_read then
	      fprintf ppf " = %sREAD_DWORD_REG (env, _REG_%s);" cast name
	    else
	      fprintf ppf "WRITE_DWORD_REG (env, _REG_%s, %s);" name vname
	| Seg n ->
	    if is_read then
	      fprintf ppf " = READ_SEGMENT_REG (env, _REG_%s);" name
	    else
	      fprintf ppf "WRITE_SEGMENT_REG (env, _REG_%s, %s);" name vname
	end
      end

let pp_group in_group ppf group =
  let ((dep, zz_list, rvars, wvars), bodies) = group in
  let rmodrm = List.exists var_reads_modrm rvars in
  let wmodrm = List.exists var_reads_modrm wvars in
  let rfetch = List.exists var_fetches rvars in
  let wfetch = List.exists var_fetches wvars in

  let decls =
    let rlist =
      if !Pp.decl_init then
	let filt (var, vname) =
	  match var with
	  | Operand (_, (Ob|Ov), _, _) -> true
	  | Operand (_, zz, _, _) -> Zz.needs_rm zz
	  | otherwise -> false
	in
	List.filter filt rvars
      else rvars
    in
    List.fold_left (var_all "r") [] rlist
  in

  let decls = List.fold_left (var_all "w") decls wvars in

  let decl s_code decls =
    let filt accu (name, code) =
      if code = s_code then name :: accu else accu
    in
    let names = List.fold_left filt [] decls in
    if names != [] then
      fprintf ppf "%s @[%a@];@," (ctype_name s_code) pp_slist names
  in

  let rm_is_reg = in_group in

  let rneedsrm = List.exists var_neads_rm rvars in
  let wneedsrm = List.exists var_neads_rm wvars in

  let rm_state, rm_local =
    if rneedsrm && wneedsrm
    then Rm_rw, false
    else
      if rneedsrm
       then Rm_r, true
      else
	if wneedsrm
	then
	  if rfetch or wfetch
	  then Rm_wf, false
	  else Rm_w, true
	else Rm_nil, true
  in

  let rm_state =
    match rm_state with
      Rm_nil when wmodrm or rmodrm -> Rm_reg
    | otherwise -> rm_state
  in
  let decls =
    match rm_state with
      Rm_nil when in_group = false -> decls
    | _ ->
	let rm = if rm_local then decls else ("rm", 2) :: decls in
	if in_group then rm else ("modrm", 0) :: rm
  in

  decl 0 decls;
  decl 1 decls;
  decl 2 decls;
  decl 3 decls;
  decl 4 decls;
  decl 5 decls;

  if (rmodrm or wmodrm) && not in_group
  then fprintf ppf "modrm = FETCH_BYTE (env);@,";

  let pp_assume ppf assume =
    if rvars != [] then
      fprintf ppf "%a@,"
	(pp_alist2 (pp_var
		      ~assume
		      ~rm_state
		      ~is_read:true
		      ~rm_is_reg)) rvars;

    let pp_body ppf (opcode, mnemonic, path) =
      let substs =
	let rec find = function
	    (Operand(0, zz, size, cast), _) :: tl -> (zz, size, cast)
	  | hd :: tl -> find tl
	  | [] -> raise Not_found in

	let substs =
	  try
	    let size =
	      match mnemonic with
		"enter_16" -> Some S16
	      | "enter_32" -> Some S32
	      | "leave_16" -> Some S16
	      | "leave_32" -> Some S32
	      | otherwise ->
		  let dep_size zz_list =
		    match dep with
		      Dep Vw -> Some S16
		    | Dep Vd -> Some S32
		    | Indep ->
			begin match zz_list with
			  zz :: _ -> Some (Zz.size zz)
			| [] -> None
			end
		  in
		  match zz_list with
		    Jb :: [] ->
		      let _, size, cast = find rvars in
		      begin match cast with
			Cast(_, size) -> Some size
		      | Raw -> Some size
		      end
		  | otherwise -> dep_size zz_list
	    in
	    begin match size with
	      Some S8 ->
		[_W_re, "BYTE";
		 _B_re, "8";
		 _D_re, "WORD";
		 _T_re, "BYTE";
		 _V_re, "0";
		 _A_re, "AL";
		 _Q_re, "q"]
	    | Some S16 ->
		[_W_re, "WORD";
		 _B_re, "16";
		 _D_re, "DWORD";
		 _T_re, "WORD";
		 _V_re, "1";
		 _A_re, "AX";
		 _Q_re, "r"]
	    | Some S32 ->
		[_W_re, "DWORD";
		 _B_re, "32";
		 _D_re, "QWORD";
		 _T_re, "DWORD";
		 _V_re, "2";
		 _A_re, "EAX";
		 _Q_re, "r"]
	    | None -> []
	    end
	  with Not_found ->
	    fprintf ppf "not found!!!@,";
	    [] in
	let substs =
	  (_N_re, string_of_int opcode) :: (_O_re, mnemonic) :: substs in
	if is_cc mnemonic then
	  (_C_re, "cc") :: substs
	else
	  substs
      in

      let pp_line ppf line =
	let line = Utils.remove_head_ws 4 line in
	fprintf ppf "%s" (Utils.subst line substs)
      in

      let rec pp_lines = function
	  line :: [] -> pp_line ppf line
	| line :: lines ->
	    fprintf ppf "%a@," pp_line line;
	    pp_lines lines
	| [] -> ()
      in
      let lineno, code =
	match path with
	  Pany (lineno, code) -> lineno, code
	| Pspec ((memline, memcode), (regline, regcode)) ->
	    begin match assume with
	      Anone ->
		eprintf "spec path for %s %a and nothing assumed@."
		  mnemonic pp_spec_zz_list (dep, zz_list);
		failwith "abort"
	    | Areg -> (regline, regcode)
	    | Amem -> (memline, memcode)
	    | Afree -> assert false
	    end
      in
      if !eff_lines
      then fprintf ppf "# %d \"%s\"@\n" lineno !eff_name;
      pp_lines code;
      if !fprintf_debug then
	fprintf ppf "@,if (NULL != turtle_log)@,    %a@,"
	  Pp.fprintf_debug (dep, opcode, mnemonic, zz_list, rvars, wvars)
    in

    let get_opno (opcode, _, _) opno =
      if in_group then (opcode - 512) mod 8 else opno in

    begin match bodies with
    | a :: [] ->
	pp_body ppf a
    | a :: b :: [] when not in_group ->
	let pp_if ppf ~if_clause =
	  if if_clause then
	    pp_body ppf a
	  else
	    pp_body ppf b
	in
	pp_if_else "0 == opno" ppf pp_if

    | list ->
	let pp_bodies ppf bodies =
	  let rec pp opno = function
	      [] -> ()
	    | body :: [] ->
		let opno = get_opno body opno in
		fprintf ppf
		  "case %d:@,@[<v 4>{@,%a@,break;@]@,}"
		  opno pp_body body;
	    | body :: rest ->
		let opno = get_opno body opno in
		fprintf ppf
		  "case %d:@,@[<v 4>{@,%a@,break;@]@,}@,"
		  opno pp_body body;

		pp (succ opno) rest
	  in
	  pp 0 bodies
	in
	if (not in_group) or (List.length bodies = 8) then
	  fprintf ppf "@[<v 4>switch (opno) {@,default:@,%a@]@,}"
	    pp_bodies list
	else
	  let pp_error ppf () =
	    fprintf ppf "NOT_IMPLEMENTED (env); return;@,"
	  in
	  fprintf ppf "@[<v 4>switch (opno) {@,%a@,default: %a@]@,}"
	    pp_bodies list pp_error ()
    end;

    if wvars != [] then
      fprintf ppf "@,%a"
	(pp_alist2 (pp_var
		      ~assume
		      ~rm_state
		      ~is_read:false
		      ~rm_is_reg)) wvars;
  in

  let has_spec_path =
    List.exists (fun (_, mnem, path) ->
      match path with
	Pspec _ -> true
      | otherwise -> false) bodies in

  if !singlerm or has_spec_path then begin
    if rm_local && not has_spec_path then
      fprintf ppf "%a" pp_assume Anone
    else begin
(*      eprintf "%b %b %b@." !singlerm has_spec_path rm_local;*)
(*      assert false;*)

      fprintf ppf "%a@,"
	(pp_alist2 (pp_var
		      ~assume:Afree
		      ~rm_state
		      ~is_read:true
		      ~rm_is_reg)) rvars;

      let pp_if ppf ~if_clause =
	let assume = if if_clause then Amem else Areg in
	fprintf ppf "%a" pp_assume assume
      in
      pp_if_else "IS_MODRM_MEM (modrm)" ppf pp_if;

      fprintf ppf "@,%a"
	(pp_alist2 (pp_var
		      ~assume:Afree
		      ~rm_state
		      ~is_read:false
		      ~rm_is_reg)) wvars;
    end
  end else begin
    fprintf ppf "%a" pp_assume Anone
  end

type switch =
    Cbad
  | Cone of int
  | Coneof of int * int
  | Cgroup of int

let pp_switch ppf = function
    Cbad -> fprintf ppf "Bad"
  | Cone n -> fprintf ppf "One(%d)" n
  | Coneof (n, opno) -> fprintf ppf "Oneof(%d, %d)" n opno
  | Cgroup n -> fprintf ppf "Group(%d)" n

type x =
    Xbad
  | Xone of string
  | Xoneof of string * int
  | Xgroup of int

let switch = Array.create 512 Cbad
let xwitch = Array.create 512 Xbad

let pat_iter, pat_jiter =
  let var_name sensetive (var, name) =
    match var with
    | Operand (pos, zz, size, cast) ->
	let sign, size =
	  match cast with
	    Cast (sign, size) -> sign, size
	  | Raw -> Iunsigned, size
	in
	if sensetive then
	  sprintf "%d%s%c"
	    pos (Zz.name zz) (char_of_sign sign)
	else
	  sprintf "%d%s%c%c"
	    pos (Zz.name zz) (char_of_sign sign) (char_of_size size)
    | Reg (reg, cast) ->
	let sign, size =
	  match cast with
	    Cast (sign, size) -> sign, size
	  | Raw -> Iunsigned, Reg.size reg
	in
	sprintf "%s%c%c"
	  (Reg.name reg) (char_of_sign sign) (char_of_size size)
    | Flag -> "f"
    | EIP -> "i"
  in
  let ngroup = ref 0 in

  let iter ppf mkey group =
    let ((dep, zz_list, rvars, wvars) as inv), bodies = group in
    let string_of_insn (opcode, mnemonic, _) =
      let name zz = Zz.spec_name dep zz in
      sprintf "%#x: %s %s"
	opcode mnemonic (String.concat "," (List.map name zz_list))
    in

    let mnems = List.map string_of_insn bodies in

    begin match dep with
      Indep -> ()
    | Dep v ->
	eprintf "key %s@." mkey;
	eprintf "D %s sensetive insns @[%a@]@."
	  (string_of_size (size_of_v v))
	  pp_slist mnems;
	failwith "abort"
    end;

    let bits = "any" in
    printf "/* %s: @[%a@] */@\n" bits pp_slist mnems;

    let do_one ppf one =
      let zz_name zz = Zz.spec_name dep zz in
      let (opcode, mnemonic, _) = one in
      let name =
	mnemonic ^ (String.concat "" (List.map zz_name zz_list)) in
      xwitch.(opcode) <- Xone name;
      fprintf ppf "OP1 (%s)@\n" name;
      fprintf ppf "@[<v 4>{@,%a@]@\n}@\n" (pp_group false) (inv, [one])
    and do_many ppf group =
      let rpart =
	"r" ^ (String.concat "" (List.map (var_name false) rvars))
      and wpart =
	"w" ^ (String.concat "" (List.map (var_name false) wvars)) in
      let name = rpart ^ "_" ^ wpart in
      let rec mark i = function
	  (opcode, _, _) :: rest ->
	    xwitch.(opcode) <- Xoneof (name, i);
	    mark (succ i) rest
	| [] -> ()
      in
      mark 0 (snd group);
      fprintf ppf "OP2 (%s)@\n" name;
      fprintf ppf "@[<v 4>{@,%a@]@\n}@\n" (pp_group false) group
    in

    begin match bodies with
      one :: [] -> do_one ppf one
    | one :: two :: [] ->
	do_one ppf one; do_one ppf two
    | otherwise -> do_many ppf group
    end;

(*    let pp decl ppf group =*)
(*      fprintf ppf "%s (group%d)@\n" decl !ngroup;*)
(*      fprintf ppf "@[<v 4>{@,%a@]@\n}@\n" (pp_group false) group;*)
(*      incr ngroup*)
(*    in*)
(*    match bodies with*)
(*      (opcode, _, _) :: [] ->*)
(*	switch.(opcode) <- Cone !ngroup;*)
(*	pp "OP1" ppf group*)

(*    | ((opa, _, _) as a) :: ((opb, _, _) as b) :: [] ->*)
(*	switch.(opa) <- Cone !ngroup;*)
(*	pp "OP1" ppf (inv, [a]);*)

(*	switch.(opb) <- Cone !ngroup;*)
(*	pp "OP1" ppf (inv, [b]);*)

(*    | l ->*)
(*	let n = ref 0 in*)
(*	List.iter begin fun (opcode, _, _) ->*)
(*	  switch.(opcode) <- Coneof (!ngroup, !n);*)
(*	  incr n*)
(*	end l;*)
(*	pp "OP2" ppf group*)

  and jiter ppf mkey groupw groupd =
    let ((depw, zz_list, rvars, wvars) as invw), bodiesw = groupw in
    let ((depd, _, _, _) as invd), bodiesd = groupd in

    let string_of_insn dep (opcode, mnemonic, _) =
      let name zz = Zz.spec_name dep zz in
      let b =
	match dep with
	  Dep Vw -> false
	| Dep Vd -> true
	| Indep -> false
      in
      let mnemonic = Zz.specialize_mnemonic mnemonic zz_list b in
      sprintf "%#x: %s %s"
	opcode mnemonic (String.concat "," (List.map name zz_list))
    in

    let mnemsw = List.map (string_of_insn depw) bodiesw in
    let mnemsd = List.map (string_of_insn depd) bodiesd in

    printf "/* 16: @[%a@] */@\n" pp_slist mnemsw;
    printf "/* 32: @[%a@] */@\n" pp_slist mnemsd;

    let do_one ppf onew oned =
      let (opcode, mnemonic, _) = onew in
      let name =
	mnemonic ^ (String.concat "" (List.map Zz.name zz_list)) in
      let (opcode, mnemonic, _) = oned in
      xwitch.(opcode) <- Xone name;
      fprintf ppf "OP1 (%s)@\n" name;
      let pp_if ppf ~if_clause =
	if if_clause then
	  pp_group false ppf (invd, [oned])
	else
	  pp_group false ppf (invw, [onew])
      in

      printf "@[<v 4>{@,%a@]@\n}@\n"
	(pp_if_else "IS_OPERAND_SIZE_32 (env)") pp_if;
    and do_many ppf groupw groupd =
      let rpart =
	"r" ^ (String.concat "" (List.map (var_name true) rvars))
      and wpart =
	"w" ^ (String.concat "" (List.map (var_name true) wvars)) in
      let name = rpart ^ "_" ^ wpart in
      let rec mark i = function
	  (opcode, _, _) :: rest ->
	    xwitch.(opcode) <- Xoneof (name, i);
	    mark (succ i) rest
	| [] -> ()
      in
      mark 0 (snd groupw);
      fprintf ppf "OP2 (%s)@\n" name;
      let pp_if ppf ~if_clause =
	if if_clause then
	  pp_group false ppf groupd
	else
	  pp_group false ppf groupw
      in

      printf "@[<v 4>{@,%a@]@\n}@\n"
	(pp_if_else "IS_OPERAND_SIZE_32 (env)") pp_if;
    in

    match bodiesw, bodiesd with
      onew :: [], oned :: [] -> do_one ppf onew oned
    | onew :: twow :: [], oned :: twod :: [] ->
	do_one ppf onew oned;
	do_one ppf twow twod;

    | otherwise ->
	do_many ppf groupw groupd
(*    let pp decl ppf groupw groupd =*)
(*      printf "%s (group%d)@\n" decl !ngroup ;*)

(*      let pp_if ppf ~if_clause =*)
(*	if if_clause then*)
(*	  pp_group false ppf groupd*)
(*	else*)
(*	  pp_group false ppf groupw*)
(*      in*)

(*      printf "@[<v 4>{@,%a@]@\n}@\n"*)
(*	(pp_if_else "IS_OPERAND_SIZE_32 (env)") pp_if;*)
(*      incr ngroup*)
(*    in*)
(*    match bodiesw, bodiesd with*)
(*      (opcode, _, _) :: [], _ :: [] ->*)
(*	switch.(opcode) <- Cone !ngroup;*)
(*	pp "OP1" ppf groupw groupd*)

(*    | ((opa, _, _) as wa) :: ((opb, _, _) as wb) :: [], da :: db :: [] ->*)
(*	switch.(opa) <- Cone !ngroup;*)
(*	pp "OP1" ppf (invw, [wa]) (invd, [da]);*)

(*	switch.(opb) <- Cone !ngroup;*)
(*	pp "OP1" ppf (invw, [wb]) (invd, [db]);*)

(*    | otherwise ->*)
(*	let n = ref 0 in*)
(*	List.iter begin fun (opcode, _, _) ->*)
(*	  switch.(opcode) <- Coneof (!ngroup, !n);*)
(*	  incr n*)
(*	end bodiesw;*)
(*	pp "OP2" ppf groupw groupd*)
  in
  iter, jiter


let pp_key ppf (mnemonic, zz_list) =
  fprintf ppf "%s %a" mnemonic pp_zz_list zz_list

let main () =
  printf "#include <stdio.h>@\n";
  printf "#include <stdint.h>@\n";
  printf "#include <stdlib.h>@\n";
  printf "#include <byteswap.h>@\n";
  printf "#include \"turtle.h\"@\n";

  if !fprintf_debug then
    printf "FILE *turtle_log;@\n";

  let pp_lines ppf lines =
    let rec pp = function
	line :: [] -> fprintf ppf "%s" line
      | line :: lines -> fprintf ppf "%s@\n" line; pp lines
      | [] -> ()
    in
    pp lines
  in

  let pp_lines_if ppf (lines_gcc, lines_any) =
    if lines_any = [] then
      fprintf ppf "@[<v 2>#ifdef __GNUC__@,%a@]@,#endif"
	pp_lines lines_gcc
    else
      fprintf ppf "@[<v 2>#ifdef __GNUC__@,%a@]@,@[<v 2>#else@,%a@]@,#endif"
	pp_lines lines_gcc
	pp_lines lines_any in

  let gcc_pro =
    ["register X86 *env __asm__(\"esi\");";
     "#define OP1(f) static void f(void);static void f()";
     "#define OP2(f) static void f(int)REGPARM(1);static void f(int opno)";
     "#define C1(f) f()";
     "#define C2(f,n) f(n)";
   ]
  and any_pro = [
    "#define OP1(f) static void f(X86*)REGPARM(1);static void f(X86*env)";
    "#define OP2(f) static void f(X86*,int)REGPARM(2);"
    ^ "static void f(X86*env,int opno)";
    "#define C1(f) f(env)";
    "#define C2(f,n) f(env,n)";]
  in
  printf "@[%a@]@\n" pp_lines_if (gcc_pro, any_pro);

  let make_insns start len =
    Utils.sub_foldi begin fun opcode accu ->
      let i_accu, sw_accu, sd_accu = accu in
      let def = defs.(opcode) in

      let accu =
	match def with
	  Def (mnemonic, zz_list) ->
	    let ikey = key_of_zz_list zz_list in
	    if Zz.insn_dsensitive mnemonic zz_list then begin
	      let insn16 = make_insn mnemonic zz_list (Dep Vw) in
	      let elem16 = (opcode, insn16) in

	      let insn32 = make_insn mnemonic zz_list (Dep Vd) in
	      let elem32 = (opcode, insn32) in
	      i_accu, elem16 :: sw_accu, elem32 :: sd_accu
	    end else begin
	      let insn = make_insn mnemonic zz_list Indep in
	      let elem = (opcode, insn) in
	      elem :: i_accu, sw_accu, sd_accu
	    end
	| otherwise -> accu
      in
      accu
    end start len ([], [], []) defs
  in

  let ilist, slist =
    Utils.sub_foldi begin fun opcode accu ->
      let (ilist, slist) = accu in
      let def = defs.(opcode) in
      let accu =
	match def with
	  Def (mnemonic, zz_list) ->
	    if Zz.insn_dsensitive mnemonic zz_list then
	      ilist, (opcode, mnemonic, zz_list) :: slist
	    else
	      (opcode, mnemonic, zz_list) :: ilist, slist
	| otherwise -> accu
      in
      accu
    end 0 512 ([], []) defs
  in

  let insns_i = List.map begin fun (opcode, mnemonic, zz_list) ->
    let insn = make_insn mnemonic zz_list Indep in
    (opcode, insn)
  end ilist in

  let make_map list =
    let fold map (opcode, mnemonic, zz_list) =
      if is_cc mnemonic or has_reg_bwd zz_list
      then map
      else
	let is_sensetive = Zz.insn_dsensitive mnemonic zz_list in

	let dep = if is_sensetive then Dep Vw else Indep in

	let insn = make_insn mnemonic zz_list dep in
	let (_, _, rvars, wvars, code) = insn in

	let key_of_vars vl =
	  String.concat "," (List.map (fun (var, _) -> key_of_var var) vl)
	in

	let vkey =
	  if not !speed then
	    if zz_list = [] then
	      sprintf "%d%b" opcode is_sensetive
	    else
	      let zkey = key_of_zz_list zz_list in
	      zkey ^ ":" ^ key_of_vars rvars ^ "  " ^ key_of_vars wvars
	  else
	    sprintf "%d" opcode
	in

	try
	  begin match (Smap.find vkey map) with
	    I (i, l) when not is_sensetive ->
	      let l = (opcode, mnemonic, code) :: l in
	      Smap.add vkey (I (i, l)) map

	  | S ((iw, lw), (id, ld)) when is_sensetive ->
	      let insnd = make_insn mnemonic zz_list (Dep Vd) in
	      let (_, zz_listd, rvarsd, wvarsd, coded) = insnd in
	      let (_, zz_listi, rvarsi, wvarsi) = id in
	      if
		zz_listd <> zz_listi or
		0 != compare rvarsd rvarsi or
		0 != compare wvarsd wvarsi
	      then begin
		eprintf "%#x %s %a@." opcode mnemonic pp_zz_list zz_list;
		let zkey = key_of_zz_list zz_list in
		let vkeyd =
		  zkey ^ ":" ^ key_of_vars rvarsd ^ "  " ^ key_of_vars wvarsd in
		eprintf "%s@." vkey;
		eprintf "%s@." vkeyd;

		eprintf "wvarsi '%s'@." (key_of_vars wvarsi);
		eprintf "wvarsd '%s'@." (key_of_vars wvarsd);

		eprintf "%b %b %b@."
		  (zz_listd <> zz_listi)
		  (rvarsd <> rvarsi)
		  (wvarsd <> wvarsi);
		eprintf "%a@." pp_zz_list zz_listd;
		failwith "whoops"
	      end;
	      let lw = (opcode, mnemonic, code) :: lw in
	      let ld = (opcode, mnemonic, coded) :: ld in
	      Smap.add vkey (S ((iw, lw), (id, ld))) map

	  | I (i, l) ->
	      assert false

	  | otherwise ->
	      eprintf "%b@." is_sensetive;
	      eprintf "%#x %s %a@." opcode mnemonic pp_zz_list zz_list;
	      eprintf "%s@." vkey;
	      failwith "unexpected"
	  end
	with Not_found ->
	  if is_sensetive then begin
	    let iw = (dep, zz_list, rvars, wvars) in
	    let lw = [opcode, mnemonic, code] in
	    let insnd = make_insn mnemonic zz_list (Dep Vd) in
	    let (_, zz_listd, rvarsd, wvarsd, coded) = insnd in
	    let id = (Dep Vd, zz_listd, rvarsd, wvarsd) in
	    let ld = [opcode, mnemonic, coded] in
	    Smap.add vkey (S ((iw, lw), (id, ld))) map
	  end else begin
	    let ii = (dep, zz_list, rvars, wvars) in
	    let li = [opcode, mnemonic, code] in
	    Smap.add vkey (I (ii, li)) map
	  end
    in
    List.fold_left fold Smap.empty list
  in

  let fold_map dep map elem =
    let (opcode, (mnemonic, zz_list, rvars, wvars, code)) = elem in
    if is_cc mnemonic or has_reg_bwd zz_list
    then map
    else
      let key_of_vars vl =
	String.concat "," (List.map (fun (var, _) -> key_of_var var) vl) in
      let ikey = key_of_zz_list zz_list in
      let mkey =
	ikey ^ ":" ^ key_of_vars rvars ^ "  " ^ key_of_vars wvars in

      let mkey =
	if opcode > 511
	then sprintf "%s_group%d" mkey ((opcode - 512) / 8)
	else mkey in
      try
	let (i, v) = Smap.find mkey map in
	let v = (opcode, mnemonic, code) :: v in
	Smap.add mkey (i, v) map
      with Not_found ->
	let i = (dep, zz_list, rvars, wvars) in
	let v = [opcode, mnemonic, code] in
	Smap.add mkey (i, v) map
  in

  let make_maps (i, w, d) =
    let mapi = List.fold_left (fold_map Indep) Smap.empty i in
    let mapw = List.fold_left (fold_map (Dep Vw)) Smap.empty w in
    let mapd = List.fold_left (fold_map (Dep Vd)) Smap.empty d in
    mapi, mapw, mapd
  in

(*  let insns = make_insns 0 512 in*)
(*  let (mapi, mapw, mapd) = make_maps insns in*)
  (*Smap.iter pat_iter mapi;*)
  (*Smap.iter pat_jiter mapi;*)

  let all =
    Utils.sub_foldi begin fun i accu ->
      match defs.(i) with
	Def (mnemonic, zz_list) -> (i, mnemonic, zz_list) :: accu
      | otherwise -> accu
    end 0 512 [] defs
  in
  let map = make_map all in
  Smap.iter begin fun mkey -> function
      I (i, l) -> pat_iter std_formatter mkey (i, l)
    | S (a, b) -> pat_jiter std_formatter mkey a b
  end map;
(*  let join =*)
(*    Smap.fold begin fun mkey (_, group) set ->*)
(*      let fold1 set (opcode, _, _) =*)
(*	Iset.add opcode set*)
(*      in*)
(*      List.fold_left fold1 set group*)
(*(*      if Smap.mem mkey mapw then*)*)
(*(*	Sset.add mkey set*)*)
(*(*      else*)*)
(*(*	set*)*)
(*    end mapd Iset.empty in*)

(*  let get_vi (i, bodies) list =*)
(*    let set = List.fold_left begin fun (opcode, _, _) ->*)
(*      Iset.add opcode set*)
(*    end Iset.empty bodies*)
(*    in*)
(*    let filt = List.filter (fun (opcode, _, _) -> Iset.mem opcode set) list in*)
(*    let map = List.fold_left (fold_map (Dep Vd)) Smap.empty filt in*)
(*    let mapd = Smap.fold begin fun mkey elem accu ->*)
(*      elem :: accu*)
(*    end map [] in*)
(*    match mapd with*)
(*      mapd :: [] -> mapd*)
(*    | otherwise -> assert false*)
(*  in*)

(*  Smap.iter begin fun mkey group ->*)
(*    if Sset.mem mkey join*)
(*    then pat_jiter mkey group (Smap.find mkey mapd)*)
(*    else pat_iter mkey group*)
(*  end mapw;*)

(*  Smap.iter begin fun mkey group ->*)
(*    if not (Sset.mem mkey join) then pat_iter mkey group*)
(*  end mapd;*)

  let groups =
    Utils.sub_foldi begin fun opcode accu ->
      match defs.(opcode) with
	Group n ->
	  (n, opcode) :: accu
      | otherwise -> accu
    end 0 512 [] defs
  in

  List.iter begin fun (n, opcode) ->
    let insns = make_insns (512 + n * 8) 8 in
    let (mapi, mapw, mapd) = make_maps insns in

    let fold mkey subgr accu = subgr :: accu in
    let li = Smap.fold fold mapi []
    and lw = Smap.fold fold mapw []
    and ld = Smap.fold fold mapd [] in

    let optab = Array.create 8 false in
    let mark l =
      List.iter (fun (_, bodies) ->
	List.iter
	  (fun (opcode, _, _) -> optab.((opcode - 512) mod 8) <- true)
	  bodies)
	l
    in
    mark li;
    mark lw;
    mark ld;

    if li != [] or lw != [] or ld != [] then
      xwitch.(opcode) <- Xgroup n;
    (*switch.(opcode) <- Cgroup n;*)

    let pp_init ppf () =
      let init = [
	"BYTE modrm = FETCH_BYTE (env);";
	"BYTE opno = GET_MODRM_REG (modrm);";
      ] in
      List.iter (fun s -> fprintf ppf "%s@," s) init
    in

    let pp_group_body ppf l =
      let pp_elem ppf elem =
	let (((dep, zz_list, rvars, wvars) as inv), bodies) = elem in

	let pp_body ppf ((opcode, _, _) as body) =
	  fprintf ppf "case %#x:@,{@[<v 4>@,%a@,break;@]@,}"
	    ((opcode - 512) mod 8)
	    (pp_group true) (inv, [body])
	in

	let rec pp = function
	    body :: [] ->
	      fprintf ppf "%a" pp_body body
	  | body :: rest ->
	      fprintf ppf "%a@," pp_body body;
	      pp rest
	  | [] -> ()
	in
	pp bodies
      in

      let pp_case ppf l =
	let rec pp = function
	    elem :: [] ->
	      fprintf ppf "%a" pp_elem elem
	  | elem :: rest ->
	      fprintf ppf "%a@," pp_elem elem;
	      pp rest
	  | [] -> ()
	in
	pp l
      in
      match l with
	a :: [] ->
	  fprintf ppf "%a" (pp_group true) a
      | l ->
	  fprintf ppf "@[<v 4>switch (opno) {@,%a@]@,}"
	    pp_case l
    in

    let pp_group_def ppf (pp_fn, arg) =
      fprintf ppf "OP1 (group_%d)@\n@[<v 4>{@,%a%a@]@\n}@\n"
	n pp_init () pp_fn arg in

    match li, lw, ld with
      [], [], [] -> ()
    | li, [], [] ->
	printf "%a" pp_group_def (pp_group_body, li)

    | [], lw, [] ->
	printf "%a" pp_group_def (pp_group_body, lw)

    | [], [], ld ->
	printf "%a" pp_group_def (pp_group_body, ld)

    | [], lw, ld ->
	let pp_if ppf ~if_clause =
	  if if_clause then
	    pp_group_body ppf ld
	  else
	    pp_group_body ppf lw
	in
	printf "%a"
	  pp_group_def ((pp_if_else "IS_OPERAND_SIZE_32 (env)"), pp_if)

    | otherwise ->
	assert false
  end groups;

  let pp_case ppf i =
(*    match switch.(i) with*)
(*      Cbad -> ()*)
(*    | Cone n ->*)
(*	fprintf ppf "C1 (group%d); break;" n*)
(*    | Coneof (n, opno) ->*)
(*	fprintf ppf "C2 (group%d, %d); break;" n opno*)
(*    | Cgroup n ->*)
(*	fprintf ppf "C1 (group_%d); break;" n*)
    match xwitch.(i) with
      Xbad -> ()
    | Xone s ->
	fprintf ppf "C1 (%s); break;" s
    | Xoneof (s, opno) ->
	fprintf ppf "C2 (%s, %d); break;" s opno
    | Xgroup n ->
	fprintf ppf "C1 (group_%d); break;" n
  in

  let pp_cc ppf (opcode, mnemonic, zz_list) =
    if !each_cc or (opcode land 0xf = 0xf) then begin
      let is_sensetive = Zz.insn_dsensitive mnemonic zz_list in
      let pp ppf () =
	if is_sensetive then begin
	  let (_, _, rvarsw, wvarsw, codew) =
	    make_insn mnemonic zz_list (Dep Vw)
	  and (_, _, rvarsd, wvarsd, coded) =
	    make_insn mnemonic zz_list (Dep Vd) in

	  let groupw =
	    ((Dep Vw), zz_list, rvarsw, wvarsw),
	    [opcode, mnemonic, codew]
	  and groupd =
	    ((Dep Vd), zz_list, rvarsd, wvarsd),
	    [opcode, mnemonic, coded] in

	  let pp_if ppf ~if_clause =
	    if if_clause then
	      fprintf ppf "%a" (pp_group false) groupd
	    else
	      fprintf ppf "%a" (pp_group false) groupw
	  in
	  pp_if_else "IS_OPERAND_SIZE_32 (env)" ppf pp_if
	end else begin
	  let (_, _, rvars, wvars, code) =
	    make_insn mnemonic zz_list Indep in
	  let group = (Indep, zz_list, rvars, wvars),
	    [opcode, mnemonic, code] in
	  pp_group false ppf group
	end
      in
      let pp_label ppf () =
	if not !each_cc then
	  fprintf ppf "do_%s%s:@,"
	    mnemonic (String.concat "" (List.map Zz.name zz_list)) in
      fprintf ppf
	"@[cc = READ_DWORD_FLAGS (env);@,cc = %s;@]@,%a@[<v 4>{@,%a@,break;@]@,}"
	ccs.(opcode mod 16) pp_label ()
	pp ()
    end else begin
      fprintf ppf
	"@[<v>cc = READ_DWORD_FLAGS (env);@,cc = %s;@,goto do_%s%s;@]"
	ccs.(opcode mod 16)
	mnemonic (String.concat "" (List.map Zz.name zz_list))
    end
  in

  let pp_reg_bwd ppf (opcode, mnemonic, zz_list) =
    if !each_reg_bwd or (opcode land 0x7 = 7) then begin
      let is_sensetive = Zz.insn_dsensitive mnemonic zz_list in
      let pp ppf () =
	if is_sensetive then begin
	  let (_, _, rvarsw, wvarsw, codew) =
	    make_insn mnemonic zz_list (Dep Vw)
	  and (_, _, rvarsd, wvarsd, coded) =
	    make_insn mnemonic zz_list (Dep Vd) in

	  let groupw =
	    ((Dep Vw), zz_list, rvarsw, wvarsw),
	    [opcode, mnemonic, codew]
	  and groupd =
	    ((Dep Vd), zz_list, rvarsd, wvarsd),
	    [opcode, mnemonic, coded] in

	  let pp_if ppf ~if_clause =
	    if if_clause then
	      fprintf ppf "%a" (pp_group false) groupd
	    else
	      fprintf ppf "%a" (pp_group false) groupw
	  in
	  pp_if_else "IS_OPERAND_SIZE_32 (env)" ppf pp_if;
	end else begin
	  let (_, _, rvars, wvars, code) =
	    make_insn mnemonic zz_list Indep in
	  let group = (Indep, zz_list, rvars, wvars),
	    [opcode, mnemonic, code] in
	  pp_group false ppf group
	end
      in
      if !each_reg_bwd then fprintf ppf "@\n#define opcode %d@\n" opcode;
      fprintf ppf "@,{@[<v 4>@,%a@,break;@]@,}"
	pp ();
      if !each_reg_bwd then fprintf ppf "@\n#undef opcode@\n";
    end else begin
      fprintf ppf "/* %s %a */" mnemonic pp_zz_list zz_list
    end
  in

  let pp_cases ppf () =
    let rec pp_prefixes i =
      let opcode, name =
	match prefixes.(i) with
	  Some (opcode, name) -> opcode, name
	| _ -> assert false
      in
      fprintf ppf "case %#05x: HANDLE_PREFIX_%s (env);"
	opcode (String.uppercase name);
(*      if i = !nprefixes - 1 then begin*)
(*	fprintf ppf*)
(*	  "@,@[<v 4>after_prefix:@,opcode = FETCH_BYTE (env);@,goto reswitch;@]@,"*)
(*      end*)
(*      else begin*)
(*	fprintf ppf " goto after_prefix;@,";*)
(*	pp_prefixes (succ i)*)
(*      end*)
      fprintf ppf " goto refetch;@,";
      if i = !nprefixes - 1 then ()
      else pp_prefixes (succ i)
    in

    let rec pp i =
      if i = 512 then ()
      else begin
(*	if switch.(i) != Cbad then*)
(*	  fprintf ppf "case %#05x: %a@," i pp_case i*)
	if xwitch.(i) != Xbad then
	  fprintf ppf "case %#05x: %a@," i pp_case i
	else begin
	  match defs.(i) with
	    Def (mnemonic, zz_list) ->
	      if is_cc mnemonic then
		fprintf ppf "case %#05x: %a@,"
		  i pp_cc (i, mnemonic, zz_list)
	      else if has_reg_bwd zz_list then
		fprintf ppf "case %#05x: %a@,"
		  i pp_reg_bwd (i, mnemonic, zz_list)
	      else ()
	  | otherwise -> ()
	end;
	pp (succ i)
      end
    in
    pp_prefixes 0;
    fprintf ppf
      "case 0x00f: opcode = FETCH_BYTE (env) | 0x100; goto reswitch;@,";
    pp 0
  in
  let pp_default ppf () =
    fprintf ppf "default: NOT_IMPLEMENTED (env); return;"
  in
  let pp_switch ppf () =
    fprintf ppf "refetch:@,";
    fprintf ppf "opcode = FETCH_BYTE (env);@,";
    fprintf ppf "reswitch:@,";
    fprintf ppf "@[<v 4>switch (opcode) {@,%a%a@]@,}"
      pp_cases () pp_default ()
  in

  let init_gcc =
    ["DWORD cc, save, opcode;";
     (*"DWORD rF;";*)
     "save = (DWORD) env;";
     "env = _env;"]
     (*"rF = READ_DWORD_FLAGS (env);"]*)
  and init_any =
    ["DWORD cc, opcode;";
     (*"DWORD rF;";*)
     "#define env _env";]
     (*"rF = READ_DWORD_FLAGS (env);"]*)
  in

  let fini_gcc = ["env = (X86 *) save;"]
  and fini_any = [] in

  printf "void turtle (X86 *_env)@\n{@[<v 4>@,%a@,%a@,%a@]@\n}@\n"
    pp_lines_if (init_gcc, init_any)
    pp_switch ()
    pp_lines_if (fini_gcc, fini_any);
  ()

let _ =
  let ic = open_in !opc_name in
  parse_opcodes ic;
  close_in ic;

  let ic = open_in !eff_name in
  parse_descr ic;
  close_in ic;
  main ()
