open Types

exception Bad_zz of string
exception No_size of zz

let dsensitive = function
    Ov | Ma | Ev | Fv | Mp | Regv | CS
  | Gv | Ap | DS | EAX | ES | Iv | FS | Jv
  | SS | GS | Xv | Ep | Yv | Jb -> true
  | _ -> false

let size zz =
  match zz with
    CL | Regb | Gb | Xb | Yb | Ob | AL | Eb | Const1 | Ib | Jb -> S8
  | DX | Rw | Ew | Sw | Iw | CS | DS | ES | SS | FS | GS -> S16
  | Rd | Td | Cd | Dd -> S32
  | _ -> raise (No_size zz)

let insn_dsensitive mnemonic operands =
  let mnemonic_dsensitive =
    match mnemonic with
      "ret" | "iret" | "retf" | "call" | "enter" | "leave"
    | "pusha" | "popa" | "cbw" | "cwd" | "pushf" -> true
    | "push" ->
	begin match operands with
	  (Ib | CS | DS | ES | SS | FS | GS) :: _ -> true
	| _ -> false
	end
    | _ -> false
  in
  mnemonic_dsensitive or (List.exists dsensitive operands)

let sensitive_names =
  let h = Hashtbl.create 10 in
  let a = Hashtbl.add h in
  let b s = Hashtbl.add h s (s, s ^ "d") in
  let c s = Hashtbl.add h s (s ^ "_16", s ^ "_32") in
  a "cbw" ("cbw", "cwde");
  a "cwd" ("cwd", "cdq");
  b "pusha";
  b "popa";
  c "call";
  c "ret";
  c "iret";
  c "retf";
  c "call";
  c "enter";
  c "leave";
  h

let specialize_operand v = function
    Ov | Ev | Fv | Regv | Gv | EAX | Iv | Jv | Xv | Yv | M -> Dep v
  | Ms -> raise (No_size Ms)
  | Ma -> Dep v
  | Ep | Mp -> Dep v
  | Ap -> Dep v
  | zz -> Indep

let specialize_mnemonic mnemonic operands o32 =
  try
    let (n16, n32) = Hashtbl.find sensitive_names mnemonic in
    if o32
    then n32
    else n16
  with Not_found ->
    begin match mnemonic, operands with
      "push", operand :: [] ->
	begin match operand with
	  Ib | CS | DS | ES | SS | FS | GS ->
	    if o32
	    then "pushd"
	    else "pushw"
	| _ -> mnemonic
	end
    |	"pop", operand :: [] ->
	begin match operand with
	  CS | DS | ES | SS | FS | GS ->
	    if o32
	    then "popd"
	    else "popw"
	| _ -> mnemonic
	end
    | s, Jb :: [] ->
	if o32
	then s ^ "d"
	else s ^ "w"
    | _ -> mnemonic
    end

(*let specialize operand_size mnemonic operands =*)
(*  let o32 =*)
(*    match operand_size with*)
(*      S32 -> true*)
(*    | S16 -> false*)
(*    | _ -> failwith "cant specialize on S8" in*)

(*  let new_mnemonic = specialize_mnemonic mnemonic operands o32 in*)
(*  let specialized_operands =*)
(*    List.map (fun operand -> operand, specialize_operand operand_size operand)*)
(*      operands*)
(*  in*)
(*  new_mnemonic, specialized_operands*)

let parse s =
  match s with
  | "CL" -> CL
  | "Rd" -> Rd
  | "DX" -> DX
  | "regb" -> Regb
  | "M" -> M
  | "Gb" -> Gb
  | "Td" -> Td
  | "Ov" -> Ov
  | "Ib" -> Ib
  | "Jb" -> Jb
  | "Xb" -> Xb
  | "Yb" -> Yb
  | "Ma" -> Ma
  | "Ev" -> Ev
  | "Fv" -> Fv
  | "Rw" -> Rw
  | "Ew" -> Ew
  | "Mp" -> Mp
  | "Sw" -> Sw
  | "regv" -> Regv
  | "CS" -> CS
  | "Gv" -> Gv
  | "Ob" -> Ob
  | "Ap" -> Ap
  | "DS" -> DS
  | "eAX" -> EAX
  | "ES" -> ES
  | "Iv" -> Iv
  | "Ms" -> Ms
  | "1" -> Const1
  | "FS" -> FS
  | "Jv" -> Jv
  | "Iw" -> Iw
  | "SS" -> SS
  | "GS" -> GS
  | "Cd" -> Cd
  | "Xv" -> Xv
  | "AL" -> AL
  | "Eb" -> Eb
  | "Ep" -> Ep
  | "Dd" -> Dd
  | "Yv" -> Yv
  | _ -> raise (Bad_zz s)

let name = function
  | CL -> "CL"
  | Rd -> "Rd"
  | DX -> "DX"
  | Regb -> "regb"
  | M -> "M"
  | Gb -> "Gb"
  | Td -> "Td"
  | Ov -> "Ov"
  | Ib -> "Ib"
  | Jb -> "Jb"
  | Xb -> "Xb"
  | Yb -> "Yb"
  | Ma -> "Ma"
  | Ev -> "Ev"
  | Fv -> "Fv"
  | Rw -> "Rw"
  | Ew -> "Ew"
  | Mp -> "Mp"
  | Sw -> "Sw"
  | Regv -> "regv"
  | CS -> "CS"
  | Gv -> "Gv"
  | Ob -> "Ob"
  | Ap -> "Ap"
  | DS -> "DS"
  | EAX -> "eAX"
  | ES -> "ES"
  | Iv -> "Iv"
  | Ms -> "Ms"
  | Const1 -> "1"
  | FS -> "FS"
  | Jv -> "Jv"
  | Iw -> "Iw"
  | SS -> "SS"
  | GS -> "GS"
  | Cd -> "Cd"
  | Xv -> "Xv"
  | AL -> "AL"
  | Eb -> "Eb"
  | Ep -> "Ep"
  | Dd -> "Dd"
  | Yv -> "Yv"

let spec_name dep zz =
  let v =
    match dep with
      Dep v ->
	begin match v with
	  Vw -> "w"
	| Vd -> "d"
	end
    | Indep ->
	begin match size zz with
	  S8 -> "b"
	| S16 -> "w"
	| S32 -> "d"
	end
  in
  match zz with
  | CL -> "CL"
  | Rd -> "Rd"
  | DX -> "DX"
  | Regb -> "regb"
  | M -> "M"
  | Gb -> "Gb"
  | Td -> "Td"
  | Ov -> "O" ^ v
  | Ib -> "Ib"
  | Jb -> "Jb"
  | Xb -> "Xb"
  | Yb -> "Yb"
  | Ma -> "Ma"
  | Ev -> "E" ^ v
  | Fv -> "F" ^ v
  | Rw -> "Rw"
  | Ew -> "Ew"
  | Mp -> "Mp"
  | Sw -> "Sw"
  | Regv -> "reg" ^ v
  | CS -> "CS"
  | Gv -> "G" ^ v
  | Ob -> "Ob"
  | Ap -> "Ap"
  | DS -> "DS"
  | EAX -> if dep = (Dep Vd) then "EAX" else "AX"
  | ES -> "ES"
  | Iv -> "I" ^ v
  | Ms -> "Ms"
  | Const1 -> "1"
  | FS -> "FS"
  | Jv -> "J" ^ v
  | Iw -> "Iw"
  | SS -> "SS"
  | GS -> "GS"
  | Cd -> "Cd"
  | Xv -> "X" ^ v
  | AL -> "AL"
  | Eb -> "Eb"
  | Ep -> "Ep"
  | Dd -> "Dd"
  | Yv -> "Y" ^ v

let has_modrm = function
  | Rd | M | Gb | Td | Ma | Ev | Rw | Ew | Mp | Sw | Gv | Ms | Cd | Eb
  | Ep | Dd -> true
  | _ -> false

let needs_rm = function
  | M | Ma | Ev | Ew | Mp | Ms | Eb | Ep -> true
  | _ -> false

let requires_address = function
  | Rd | M | Gb | Td | Ma | Ev | Rw | Ew | Mp | Sw | Gv | Ms | Cd | Eb
  | Ep | Dd -> true
  | _ -> false

let fetches = function
  | Ov | Ib | Jb | Iv | Jv | Iw -> true
  | zz -> false (*has_modrm zz*)
