#include <stdlib.h>
#include <string.h>

#include <inc/mem.h>
#include <inc/x86.h>
#include <inc/asm.h>
#include <inc/sig.h>
#include <inc/int.h>
#include <inc/sel.h>
#include <inc/red.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "INT"
static int log_level;

#define PCREL(at, to) do{*(ui*)(at)=(ui)(to)-(ui)(at)-4;}while(0)

void int_trampoline (int nint, intregs r);
void irq_trampoline (int nint, intregs r);

static struct idte {
  us sel;
  ui off;
  int_handler handler;
} int_tab[256];

static int enabled;
static ub iret[] = {0xcf};

void INT_log_regs (INT_ARGS)
{
  log ("intregs:\n"
       "eax=%#010x ebx=%#010x ecx=%#010x edx=%#010x\n"
       "esi=%#010x edi=%#010x ebp=%#010x esp=%#010x\n"
       "eip=%#010x efl=%#010x\n"
       "cs=%#06x ds=%#06x es=%#06x ss=%#06x "
       "fs=%#06x gs=%#06x\n",
       EAX(r), EBX(r), ECX(r), EDX(r), ESI(r), EDI(r), EBP(r),
       ESP(r), EIP(r), EFL(r),
       CS (r), DS(r), ES(r), FS (r), ASM_ss, GS (r)
    );
}

static void unhandled_int (INT_ARGS)
{
  char *eip;

  eip = (char *) SELLIN (CS (r), EIP (r));
  lwarn ("unhandled int %#x called from %p\n",
         eip[-1], eip - 2);
  INT_log_regs (r);
  Fail ("exiting");
}

void INT_enable ()
{
  /* log ("%p\n", __builtin_return_address (0)); */
  enabled = 1;
}

void INT_disable ()
{
  /* log ("%p\n", __builtin_return_address (0)); */
  enabled = 0;
}

int INT_is_enabled ()
{
  return enabled;
}

static void get_pm (int nint, us *sel, ui *off)
{
  int e;

  e = enabled;
  enabled = 0;
  *sel = int_tab[nint].sel;
  *off = int_tab[nint].off;
  enabled = e;
}

static void get_rm (int nint, us *seg, us *off)
{
  int e;
  us *p;

  p = (us *) MEM_REAL (nint * 4);
  e = enabled;
  enabled = 0;
  *off = p[0];
  *seg = p[1];
  enabled = e;
}

static void set_rm (int nint, us seg, us off)
{
  int e;
  us *p;

  p = (us *) MEM_REAL (nint * 4);
  e = enabled;
  enabled = 0;
  p[0] = off;
  p[1] = seg;
  enabled = e;
}

static ui create_trampoline (int nint, int is_irq)
{
  ui tramp;
  ui addr;
  int int_trampoline_size;

  int_trampoline_size = ASM_int_trampoline_end - ASM_int_trampoline;

  addr = (ui) (is_irq ? irq_trampoline : int_trampoline);
  ASM_int_trampoline_nint = nint;
  ASM_int_trampoline_addr = addr;

  tramp = (ui) xmalloc (int_trampoline_size);
  memcpy ((void *) tramp, ASM_int_trampoline, int_trampoline_size);

  return tramp;
}

void INT_get_pm (int nint, us *sel, ui *off)
{
  get_pm (nint, sel, off);
  linfo ("get %#x, %#x:%#x\n", nint, *sel, *off);
}

void INT_set_pm (int nint, us sel, ui off)
{
  int e;

  e = enabled;
  enabled = 0;
  int_tab[nint].sel = sel;
  int_tab[nint].off = off;
  enabled = e;
  linfo ("set %#x, %#x:%#x\n", nint, sel, off);
}

void INT_get_rm (int nint, us *sel, us *off)
{
  get_rm (nint, sel, off);
}

void INT_set_rm (int nint, us sel, us off)
{
  set_rm (nint, sel, off);
}

void INT_set (int nint, int_handler handler)
{
  int e;
  int is_irq;

  is_irq = ((nint >= 8) && (nint < 16)) || ((nint >= 0x70) && (nint < 0x78));

  e = enabled;
  enabled = 0;
  int_tab[nint].sel = ASM_cs;
  int_tab[nint].off = create_trampoline (nint, is_irq);
  int_tab[nint].handler = handler;
  set_rm (nint, RED_SEGMENT, RED_INTERRUPT_00 + nint);
  enabled = e;
}

void INT_simulate (int nint, regs86 *r)
{
  ui *sp;
  ui off;
  us sel;

  if (LOG_DEBUG == log_level) {
    log ("simulate int %#x\n", nint);
    SIG_log_regs (r);
  }

  SIGESP (r) -= 12;
  if (ASM_ss == SIGSS (r))
    sp = (ui *) SIGESP (r);
  else
    sp = (ui *) SELLIN (SIGSS (r), SIGESP (r));

  sp[0] = SIGEIP (r);
  sp[1] = SIGCS (r);
  sp[2] = SIGEFL (r);

  get_pm (nint, &sel, &off);

  SIGCS (r) = sel;
  SIGEIP (r) = off;
}

void INT_simulate_real (int nint, intregs regs)
{
  if (NULL == int_tab[nint].handler) {
    log ("no handler for int %#x\n", nint);
    UNHANDLED_INT (regs);
  }

  ldebug ("simulate real %#x eax = %#lx\n", nint, (long) EAX (regs));

  if (LOG_DEBUG == log_level) {
    INT_log_regs (regs);
  }

  int_tab[nint].handler (regs);
}

void irq_trampoline (int nint, intregs r)
{
  us seg, off;
  struct vm86_regs vr;

  DEREF (r);
  get_rm (nint, &seg, &off);

  vr.cs = seg;
  vr.eip = off;
  vr.eflags = 0;
  SIG_vm86 (&vr, 1);
}

void int_trampoline (int nint, intregs r)
{
  ldebug ("enter int trampoline %d %p\n", nint, (void *) r);

  if (0x31 == nint) {
    int_tab[nint].handler (r);
  }
  else {
    us seg, off;
    struct vm86_regs vr;

    get_rm (nint, &seg, &off);

    vr.cs = seg;
    vr.eip = off;
    vr.eax = r[INT_REG_EAX];
    vr.ebx = r[INT_REG_EBX];
    vr.ecx = r[INT_REG_ECX];
    vr.edx = r[INT_REG_EDX];
    vr.esi = r[INT_REG_ESI];
    vr.edi = r[INT_REG_EDI];
    vr.ebp = r[INT_REG_EBP];
    vr.eflags = r[INT_REG_EFL] & ~(X86_EFLAGS_TF | X86_EFLAGS_IF);

    SIG_vm86 (&vr, 1);

    r[INT_REG_EAX] = vr.eax;
    r[INT_REG_EBX] = vr.ebx;
    r[INT_REG_ECX] = vr.ecx;
    r[INT_REG_EDX] = vr.edx;
    r[INT_REG_ESI] = vr.esi;
    r[INT_REG_EDI] = vr.edi;
    r[INT_REG_EBP] = vr.ebp;
    r[INT_REG_EFL] = vr.eflags & ~(X86_EFLAGS_VM | X86_EFLAGS_TF);
  }
  ldebug ("leave int trampoline %d %p\n", nint, (void *) r);
}

static void ignore (INT_ARGS)
{
  lwarn ("ignoring\n");
  INT_log_regs (r);
}

void INT_init ()
{
  int i;
  us cs;
  ui unhandled_int_handler;
  ub *redir;

  DEREF (ignore);

  log_level = MUP_get_log_level ("int");

  cs = ASM_cs;

  unhandled_int_handler = create_trampoline (0, 0);
  int_tab[0].handler = unhandled_int;

  redir = (ub *) RED_FLAT (RED_INTERRUPT_00);

  for (i = 0; i < 256; i++) {
    INT_set_pm (i, cs, unhandled_int_handler);
    INT_set_rm (i, RED_SEGMENT, RED_INTERRUPT_00 + i);
    redir[i] = RED_ICE;
  }

  INT_set_rm (0x66, 0, 0);
  INT_set_rm (0x67, 0, 0);
  INT_set_rm (0x68, 0, 0);
  INT_set_rm (0x69, 0, 0);

  INT_set (0x10, I10);
  INT_set (0x11, I11);
  INT_set (0x12, I12);
  INT_set (0x15, I15);
  INT_set (0x16, I16);
  INT_set (0x1A, I1A);
  INT_set (0x21, I21);
  INT_set (0x29, I29);
  INT_set (0x2f, I2F);
  INT_set (0x31, I31);
  INT_set (0x33, I33);
  INT_set (0x4b, I4B);

  INT_set (0x13, ignore);

  redir[8 + 5] = 0xcf;
  redir[8 + 7] = 0xcf;
  redir[8 + 11] = 0xcf;
  INT_set_pm (8 + 5, ASM_cs, (ui) &iret);
  INT_set_pm (8 + 7, ASM_cs, (ui) &iret);
  INT_set_pm (8 + 11, ASM_cs, (ui) &iret);

  enabled = 1;
}
