#include <string.h>
#include <unistd.h>
#include <sys/syscall.h>

#include <inc/sel.h>
#include <inc/ldt.h>
#include <inc/xdt.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "LDT"
static int log_level;

#if 0
#include <asm/ldt.h>
#else
/* Maximum number of LDT entries supported. */
#define LDT_ENTRIES     8192
/* The size of each LDT entry. */
#define LDT_ENTRY_SIZE  8

typedef struct modify_ldt_ldt_s {
        unsigned int  entry_number;
        unsigned long base_addr;
        unsigned int  limit;
        unsigned int  seg_32bit:1;
        unsigned int  contents:2;
        unsigned int  read_exec_only:1;
        unsigned int  limit_in_pages:1;
        unsigned int  seg_not_present:1;
        unsigned int  useable:1;
} modify_ldt_t;

#define MODIFY_LDT_CONTENTS_DATA        0
#define MODIFY_LDT_CONTENTS_STACK       1
#define MODIFY_LDT_CONTENTS_CODE        2
#endif

#define SEL2IDX(x) ((x)>>3)
#define IDX2SEL(x) (((x)<<3)|7)

static char LDT_map[LDT_ENTRIES];
static modify_ldt_t LDT_tab[LDT_ENTRIES];

static void checked_modify_ldt (int index)
{
  int count;

  count = syscall (SYS_modify_ldt, 1, LDT_tab + index, sizeof (modify_ldt_t));

  if (0 != count) {
    ERRFail ("modify_ldt %d failed %d", index, count);
  }
}

#if 0
static void read_ldt (int index, void *buf)
{
  int count;
  char map[LDT_ENTRIES * LDT_ENTRY_SIZE];

  count = syscall (SYS_modify_ldt, 0, map, sizeof (map));

  if (sizeof (map) != count) {
    ERRFail ("modify_ldt(read) %d failed, count %d", index, count);
  }

  memcpy (buf, map + LDT_ENTRY_SIZE * index, LDT_ENTRY_SIZE);
}
#endif

static void log_ldt_entry (int index)
{
  static const char *content_names[] = {"data", "stack", "code"};

  log ("%d\n"
       "limit           %d, %#x\n"
       "base_addr       %ld, %#lx\n"
       "seg_not_present %d\n"
       "read_exec_only  %d\n"
       "limit_in_pages  %d\n"
       "seg_32bit       %d\n"
       "contents        %d %s\n",
       index,
       LDT_tab[index].limit, LDT_tab[index].limit,
       LDT_tab[index].base_addr, LDT_tab[index].base_addr,
       LDT_tab[index].seg_not_present,
       LDT_tab[index].read_exec_only,
       LDT_tab[index].limit_in_pages,
       LDT_tab[index].seg_32bit,
       LDT_tab[index].contents,
       LDT_tab[index].contents < 3
       ? content_names[LDT_tab[index].contents]
       : "bogus"
    );
}

static const char *names[] =
{"dro", "drw", "sro", "srw", "ceo", "cer", "Ceo", "Cer"};

enum {DRO, DRW, SRO, SRW, CEO, CER, CCEO, CCER};

static void log_descriptor (const char *cap, void *buf)
{
  int g, D, x, u, p, dpl, s, typ, a;
  unsigned long base_addr;
  unsigned int limit;
  ul d;

  d = *(ul *) buf;
  limit = (d & 0xffff) | (((d >> 48) & 0xf) << 16);
  base_addr =
    ((d >> 16) & 0xffff)
    | (((d >> 32) & 0xff) << 16)
    | (((d >> 56) & 0xff) << 24);

  g = (d >> 55) & 1;
  D = (d >> 54) & 1;
  x = (d >> 53) & 1;
  u = (d >> 52) & 1;

  p = (d >> 47) & 1;
  dpl = (d >> 45) & 3;
  s = (d >> 44) & 1;
  typ = (d >> 41) & 7;
  a = (d >> 40) & 1;

  log ("%s\ndescriptor at %p %#llx\n"
       "limit %d, %#x\n"
       "base %ld, %#lx\n"
       "%c%c%c%c %c dpl=%d %c type=%d:%s %c\n\n",
       cap,
       buf, d,
       limit, limit,
       base_addr, base_addr,
       g ? 'G' : 'g',
       D ? 'D' : 'd',
       x ? 'X' : 'x',
       u ? 'U' : 'u',
       p ? 'P' : 'p',
       dpl,
       s ? 'S' : 's',
       typ, names[typ],
       a ? 'A' : 'a');
}

int LDT_alloc_selectors (int count)
{
  int i, index;
  int region = 0;

  for (i = 0; i < LDT_ENTRIES; i++) {
    if (LDT_map[i])
      region = 0;
    else
      if (++region == count)
        break;
  }

  if (region < count)
    return -1;

  index = i - count + 1;

  for (i = index; i < index + count; i++) {
    LDT_map[i] = 1;

    LDT_tab[i].entry_number = i;
    LDT_tab[i].base_addr = 0;
    LDT_tab[i].limit = 0;
    LDT_tab[i].seg_32bit = 0;
    LDT_tab[i].contents = MODIFY_LDT_CONTENTS_DATA;
    LDT_tab[i].read_exec_only = 0;
    LDT_tab[i].limit_in_pages = 0;
    LDT_tab[i].seg_not_present = 0;
    LDT_tab[i].useable = 1;
  }

  linfo ("allocated %d selectors starting from %d %#x\n",
         count, index, IDX2SEL (index));
  return IDX2SEL (index);
}

static int check_selector (int selector, char *operation, int fatal)
{
  int index;

  index = selector >> 3;

  if (index >= LDT_ENTRIES) {
    Fail ("attempt to %s bogus selector %#x", operation, selector);
  }

  if (IS_GDT_SELECTOR (selector)) {
    Fail ("attempt to %s GDT selector %#x", operation, selector);
  }

  switch (LDT_map[index]) {
    case 0:
      if (fatal)
        Fail ("attempt to %s free selector %#x", operation, selector);
      else {
        lwarn ("attempt to %s free selector %#x\n", operation, selector);
      }
      break;

    case 2:
      if (fatal)
        Fail ("attempt to %s system selector %#x", operation, selector);
      else {
        lwarn ("attempt to %s system selector %#x\n", operation, selector);
      }
      break;

    case 3:
      lwarn ("%s permanent selector %#x\n", operation, selector);
      break;

    case 4:
      lwarn ("%s immutable selector %#x\n", operation, selector);
      break;
  }

  return index;
}

static int check_selector2 (int selector, char *operation)
{
  int index;

  index = selector >> 3;

  if (index >= LDT_ENTRIES) {
    Fail ("attempt to %s bogus selector %#x", operation, selector);
  }

  if (IS_GDT_SELECTOR (selector)) {
    Fail ("attempt to %s GDT selector %#x", operation, selector);
  }

  return index;
}

void LDT_free_selector (int selector)
{
  int index;

  index = selector >> 3;

  if (index >= LDT_ENTRIES) {
    Fail ("attempt to free bogus selector %#x", selector);
  }

  if (IS_GDT_SELECTOR (selector)) {
    Fail ("attempt to free GDT selector %#x", selector);
  }

  switch (LDT_map[index]) {
    case 0:
      lwarn ("freeing free selector %#x\n", selector);
      break;

    case 1:
      LDT_map[index] = 0;
      break;

    case 2:
      lwarn ("attempt to free system selector %#x\n", selector);
      break;

    case 3:
      lwarn ("attempt to free permanent selector %#x\n", selector);
      break;

    case 4:
      lwarn ("attempt to free immutable selector %#x\n", selector);
      break;
  }
  /* memset (LDT_tab + index, 0, sizeof (modify_ldt_t)); */
  LDT_tab[index].useable = 0;
  /* LDT_tab[index].seg_not_present = 1; */
  /* checked_modify_ldt (index); */
}

void LDT_set_selector_base (int selector, unsigned long base)
{
  int index;

  index = check_selector (selector, "rebase", 1);

  LDT_tab[index].base_addr = base;
  checked_modify_ldt (index);
}

void LDT_set_selector_limit (int selector, unsigned int limit)
{
  int index;

  index = check_selector (selector, "change limit of", 1);

  if (limit > 0xfffff) {
    LDT_tab[index].limit_in_pages = 1;
    LDT_tab[index].limit = limit >> 12;
  }
  else {
    LDT_tab[index].limit_in_pages = 0;
    LDT_tab[index].limit = limit;
  }

  checked_modify_ldt (index);
}

void LDT_make_selector_permanent (int selector)
{
  int index;

  index = check_selector (selector, "set permanent", 1);
  LDT_map[index] = 3;
}

void LDT_make_selector_immutable (int selector)
{
  int index;

  index = check_selector (selector, "set immutable", 1);
  LDT_map[index] = 4;
}

unsigned long LDT_get_selector_base (int selector)
{
  int index;

  /* index = check_selector (selector, "get base of", 1); */
  index = check_selector2 (selector, "get base of");
  return LDT_tab[index].base_addr;
}

unsigned int LDT_get_selector_limit (int selector)
{
  int index;

  index = check_selector (selector, "get limit of", 1);

  if (LDT_tab[index].limit_in_pages)
    return LDT_tab[index].limit << 12;
  else
    return LDT_tab[index].limit;
}

void LDT_get_selector_descriptor (int selector, void *buf)
{
  int index;
  unsigned short access;

  /* index = check_selector (selector, "get descriptor of", 0); */
  index = check_selector2 (selector, "get descriptor of");

  access = ((LDT_tab[index].limit_in_pages != 0) << 15)
    | ((LDT_tab[index].seg_32bit != 0) << 14)
    | ((LDT_tab[index].seg_not_present == 0) << 7)
    | (3 << 5)
    | (1 << 4);

  switch (LDT_tab[index].contents) {
    case MODIFY_LDT_CONTENTS_DATA:
      if (0 == LDT_tab[index].read_exec_only) {
        access |= 2;
      }
      break;

    case MODIFY_LDT_CONTENTS_CODE:
      if (0 == LDT_tab[index].read_exec_only) {
        access |= 8;
      }
      else {
        access |= 10;
      }
      break;

    case MODIFY_LDT_CONTENTS_STACK:
      access |= 6;
      break;

    default:
      Aldri ();
  }

  XDT_set_descriptor_base (buf, LDT_tab[index].base_addr);
  XDT_set_descriptor_limit (buf, LDT_tab[index].limit);
  XDT_set_descriptor_access (buf, access);

#if 0
  if (LOG_DEBUG == log_level) {
    log_descriptor (__func__, buf);
  }
#endif
}

static const unsigned int contents_tab[] = {
  MODIFY_LDT_CONTENTS_DATA,
  MODIFY_LDT_CONTENTS_STACK,
  MODIFY_LDT_CONTENTS_CODE
};

static void set_typ (int index, int present, unsigned short ar)
{
  if (present) {
    if (((ar >> 2) & 3) == 3)
      Fail ("conforming segments not supported index %d ar %#x",
            index, ar);

    if (((ar >> 1) & 7) == 2)
      Fail ("read only stack is not supproted index %d ar %#x",
            index, ar);
  }

  LDT_tab[index].read_exec_only = 0 == ((ar >> 1) & 1);
  LDT_tab[index].contents = contents_tab[(ar >> 2) & 3];
}

void LDT_set_selector_descriptor (int selector, void *buf)
{
  int index;
  int present;
  unsigned long long d;

  index = check_selector (selector, "set descriptor of", 1);

  d = *(unsigned long long *) buf;

  present = 0 != ((d >> 47) & 1);

  if (present && (0 == ((d >> 44) & 1))) {
    Fail ("system descriptors not supported %#llx", d);
  }

  LDT_tab[index].limit           = XDT_get_descriptor_limit (buf);
  LDT_tab[index].base_addr       = XDT_get_descriptor_base (buf);
  LDT_tab[index].limit_in_pages  = (d >> 55) & 1;
  LDT_tab[index].seg_32bit       = (d >> 54) & 1;
  LDT_tab[index].seg_not_present = 0 == present;

  set_typ (index, present, (d >> 40) & 0xffff);
  checked_modify_ldt (index);

  if (LOG_DEBUG == log_level) {
    log_descriptor (__func__, buf);
    log_ldt_entry (index);
  }
}

unsigned int LDT_get_selector_access_rights (int selector)
{
  int index;
  unsigned int ar;

  index = check_selector (selector, "get access rights of", 1);

  ar = 0;
  ar |= LDT_tab[index].limit_in_pages << 15;
  ar |= LDT_tab[index].seg_32bit << 14;
  ar |= 0x7 << 4;
  ar |= (LDT_tab[index].seg_not_present == 0) << 7;
  switch (LDT_tab[index].contents) {
    case MODIFY_LDT_CONTENTS_CODE:
      ar |= 1 << 3;
      break;

    case MODIFY_LDT_CONTENTS_STACK:
      ar |= 1 << 2;
      break;
  }
  ar |= (LDT_tab[index].read_exec_only == 0) << 1;
  return ar;
}

static void set_access (int index, int ar)
{
  int present, dpl;

  present = 0 != (ar & 0x80);
  set_typ (index, present, ar);

  dpl = (ar >> 5) & 3;
  if (3 != dpl) {
    lwarn ("DPL %d, setting to 3\n", dpl);
  }

  LDT_tab[index].limit_in_pages  = 0 != (ar & 0x8000);
  LDT_tab[index].seg_32bit       = 0 != (ar & 0x4000);
  LDT_tab[index].seg_not_present = 0 == present;
}

void LDT_set_selector_access_rights (int selector, int ar)
{
  /* From pmode/w 1.33 documentation pmw.txt

  ) The access rights/type word passed to the function in CX has the following
  format:

  Bit: 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
      +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
      | G |B/D| 0 | ? |       ?       | 1 |  DPL  | 1 |C/D|E/C|W/R| A |
      +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

      G   - 0=byte granular, 1=page granular
      B/D - 0=default 16bit, 1=default 32bit
      DPL - must be equal to caller's CPL
      C/D - 0=data, 1=code
      E/C - data: 0=expand-up, 1=expand-down
      code: must be 0 (non-conforming)
      W/R - data: 0=read, 1=read/write
      code: must be 1 (readable)
      A   - 0=not accessed, 1=accessed
      0   - must be 0
      1   - must be 1
      ?   - ignored
  */

  int index;
  int present;

  present = 0 != (ar & 0x80);

  if (present && (0 == (ar & (1 << 4)))) {
    Fail ("invalid access rights %#x for selector %#x (system descriptor)",
          ar, selector);
  }

  index = check_selector (selector, "set access rights of", 1);

  set_access (index, ar);

  if (LOG_DEBUG == log_level) {
    log_ldt_entry (index);
  }
  checked_modify_ldt (index);
}

int LDT_create_system_alias (int is_code, unsigned long base_addr)
{
  int alias;
  int index;

  alias = LDT_alloc_selectors (1);
  if (-1 == alias) {
    Fail ("LDT_alloc_selectors (1) failed for system %s descriptor",
          is_code ? "code" : "data");
  }

  index = SEL2IDX (alias);
  LDT_tab[index].limit_in_pages  = 1;
  LDT_tab[index].seg_not_present = 0;
  LDT_tab[index].seg_32bit       = 1;
  LDT_tab[index].read_exec_only  = 0;
  LDT_tab[index].useable         = 1;
  LDT_tab[index].base_addr       = base_addr;
  LDT_tab[index].limit           = 0xfffff;
  LDT_tab[index].contents        = (is_code
                                    ? MODIFY_LDT_CONTENTS_CODE
                                    : MODIFY_LDT_CONTENTS_DATA);

  if (LOG_DEBUG == log_level) {
    log_ldt_entry (index);
  }
  checked_modify_ldt (index);
  return alias;
}

int LDT_create_alias (int selector)
{
  int alias;
  int alias_index;
  int entry_number;
  int index;

  index = check_selector (selector, "create alias of", 1);
  alias = LDT_alloc_selectors (1);
  if (-1 == alias) {
    lwarn ("LDT_alloc_selectors (1) failed\n");
    return -1;
  }

  alias_index = SEL2IDX (alias);
  entry_number = LDT_tab[alias_index].entry_number;
  memcpy (LDT_tab + alias_index, LDT_tab + index, sizeof (*LDT_tab));
  LDT_tab[alias_index].entry_number = entry_number;
  LDT_tab[alias_index].contents = MODIFY_LDT_CONTENTS_DATA;
  checked_modify_ldt (alias_index);
  if (LOG_DEBUG == log_level) {
    log_ldt_entry (index);
    log_ldt_entry (alias_index);
  }
  return alias;
}

int LDT_is_selector_big (int selector)
{
  /* int index = check_selector (selector, "get big", 1); */
  int index = check_selector2 (selector, "get big");
  return LDT_tab[index].seg_32bit;
}

void LDT_init ()
{
  int i;
  int nread;
  int used;
  char tab[LDT_ENTRY_SIZE * LDT_ENTRIES];
  char empty[LDT_ENTRY_SIZE];

  XDT_init ();
  log_level = MUP_get_log_level ("ldt");

  memset (empty, 0, LDT_ENTRY_SIZE);
  nread = syscall (SYS_modify_ldt, 0, &tab, LDT_ENTRY_SIZE * LDT_ENTRIES);

  if (-1 == nread) {
    ERRFail ("modify_ldt");
  }

  used = 0;

  for (i = 0; i < nread / LDT_ENTRY_SIZE; i++) {
    char *buf;

    buf = tab + i * LDT_ENTRY_SIZE;
    if (memcmp (buf, empty, LDT_ENTRY_SIZE)) {
      LDT_map[i] = 2;
      LDT_tab[i].entry_number = i;
      LDT_tab[i].base_addr = XDT_get_descriptor_base (buf);
      LDT_tab[i].limit = XDT_get_descriptor_limit (buf);
      set_access (i, XDT_get_descriptor_access (buf));
      used++;
    }
  }

  linfo ("nread = %d used = %d\n", nread, used);
}
