#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>

#include <inc/mem.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "MEM"
static int log_level;

#define MAX_VIDBUF_SIZE (1600 * 1200 * 4)

static int ram_size = 0x3000000;
static int vid_map_size;
static int ram_fd;
static int vid_fd;
static int vid_off;
static int vid_addr;
static void *vid_buf;

void *MEM_video_buffer ()
{
  return vid_buf;
}

void MEM_unmap_mappings_to_video ()
{
  ldebug ("addr %#x, size %d\n", vid_addr, vid_map_size);
  if (vid_addr) {
    void *map, *vid_map;

    vid_map = (void *) vid_addr;
    ldebug ("vid_addr %p\n", vid_map);

    Fail_if (vid_map_size <= 0);

    if (-1 == munmap (vid_map, vid_map_size)) {
      ERRFail ("munmap: addr=%p, size=%d", vid_map, vid_map_size);
    }

    if (vid_addr < ram_size) {
      map = mmap (vid_map, vid_map_size,
                  PROT_READ | PROT_WRITE,
                  MAP_SHARED | MAP_FIXED,
                  ram_fd, vid_addr);

      if (vid_map != map) {
        ERRFail ("mmap: map=%p addr=%p, size=%d, ram_fd=%d, addr=%d\n",
                 map, vid_map, vid_map_size, ram_fd, vid_addr);
      }
    }
    vid_addr = 0;
    vid_map_size = -1;
  }
}

void MEM_map_address_to_video (int addr, int off, int size)
{
  void *map, *vid_map;

  ldebug ("(addr=%#x, off=%d, size=%d)\n", addr, off, size);

  if ((0 == addr) || (size <= 0)) {
    Fail ("addr=%#x size=%d\n", addr, size);
  }

  vid_map = (void *) addr;
  map = mmap (vid_map, size,
              PROT_READ | PROT_WRITE,
              MAP_SHARED | MAP_FIXED,
              vid_fd, off + vid_off);

  if (map != vid_map) {
    ERRFail ("mmap: map=%p addr=%p, size=%d, vid_fd=%d, off=%d",
             map, vid_map, size, vid_fd, off);
  }

  vid_addr = addr;
  vid_map_size = size;
}

static inline int mkprot (int mprot)
{
  switch (mprot) {
    case MEM_MAP_RDONLY:
      return PROT_READ;

    case MEM_MAP_WRONLY:
      return PROT_WRITE;

    case MEM_MAP_RDWR:
      return PROT_READ | PROT_WRITE;

    case MEM_MAP_NONE:
      return PROT_NONE;

    default:
      Fail ("wrong protection %#x", mprot);
  }
}

void MEM_map_address (int target, int addr, int size)
{
  void *map, *addr_map;

  ldebug ("(target=%#x, addr=%#x, size=%d)\n", target, addr, size);

  addr_map = (void *) addr;
  map = mmap (addr_map, size,
              PROT_READ | PROT_WRITE,
              MAP_SHARED | MAP_FIXED,
              ram_fd, target);
  if (map != addr_map) {
    ERRFail ("mmap: map=%p addr=%p, size=%d, vid_fd=%d, target=%#x",
             map, addr_map, size, ram_fd, target);
  }
}

void MEM_protect_memory (int addr, int size, int mprot)
{
  int res;

  res = mprotect ((void *) addr, size, mkprot (mprot));
  if (-1 == res) {
    ERRFail ("mprotect: addr=%#x, size=%d, mprot=%d", addr, size, mprot);
  }
}

void *MEM_map (int size, int mprot)
{
  return mmap (NULL, size, mkprot (mprot), MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
}

int MEM_ram_fd ()
{
  return ram_fd;
}

int MEM_ram_size ()
{
  return ram_size;
}

static int open_bs (const char *sh_name,
                    const char *name,
                    int size)
{
  int fd;
  int err;
  int is_shm;

  fd = shm_open (sh_name, O_RDWR | O_CREAT | O_TRUNC, 0600);
  if (-1 == fd) {
    err = errno;
    lwarn ("shm_open(%s): %s\n", sh_name, strerror (err));
    is_shm = 0;
  }
  else is_shm = 1;

  if (0 == is_shm) {
    fd = open (name, O_RDWR | O_CREAT | O_TRUNC, 0600);

    if (-1 == fd) {
      ERRFail ("backing store open failed (%s, %s, %d)",
               sh_name, name, size);
    }
  }

  if (-1 == ftruncate (fd, size)) {
    ERRFail ("ftruncate(%d, %d)", fd, size);
  }

  if (is_shm)
    err = shm_unlink (name);
  else
    err = unlink (name);

  if (-1 == err) {
    ERRFail ("%sunlink(%s)",
             is_shm ? "shm_" : "",
             is_shm ? sh_name : name);
  }

  return fd;
}

static void init_ram ()
{
  void *ram, *map;
  const char name[] = "/minde.ram";

  ram_fd = open_bs (name, name + 1, ram_size);

  ram = (void *) MEM_REAL (0);
  map = mmap (ram, ram_size,
              PROT_READ | PROT_WRITE,
              MAP_SHARED | MAP_FIXED,
              ram_fd, 0);

  ldebug ("ram=%p..%p size=%d\n",
          ram, (char *) ram + ram_size, ram_size);

  if (map != ram) {
    ERRFail ("mmap(%p, ..., fd=%d) = %p", map, ram_fd, ram);
  }
}

static void init_vram ()
{
  void *map;
  ui addr, aligned;
  const char name[] = "/minde.vram";

  vid_fd = open_bs (name, name + 1, MAX_VIDBUF_SIZE);

  map = mmap (NULL, MAX_VIDBUF_SIZE + 0xffff,
              PROT_READ | PROT_WRITE,
              MAP_SHARED, vid_fd, 0);

  if (MAP_FAILED == map) {
    ERRFail ("mmap(0, ..., fd=%d)", vid_fd);
  }

  addr = (ui) map;
  aligned = (addr + 0xffff) & ~0xffff;
  vid_off = aligned - addr;
  vid_buf = (void *) aligned;

  ldebug ("vid_buf=%p..%p vid_size=%d\n",
          vid_buf, (char *) vid_buf + MAX_VIDBUF_SIZE, MAX_VIDBUF_SIZE);
}

void MEM_pre_init ()
{
  static struct MUP_key keys[] = {
    {"size", TAG_INT, {.ip = &ram_size}},
  };

  static struct MUP_group group =
    {"mem.", LENOFA (keys), keys, NULL};

  MUP_add_group (&group);
}

void MEM_init ()
{
  log_level = MUP_get_log_level ("mem");

  ram_size += 0x100000;

  init_ram ();
  init_vram ();

  strcpy ((char *) MEM_REAL (0xffff5), "88/01/01");
}
