#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/poll.h>

#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "MUP"

#define MAXDRIVES ('Z' - 'A' + 1)

static struct fs_map {
  char letter;
  char *path;
} fs_map[MAXDRIVES] = {{0, 0}, };

static int default_log_level = -1;
static char *exe;
static char *dir;
static char *args;
static int ntopkeys;
static struct MUP_key *topkeys;
static struct MUP_group *last_group;
static struct MUP_group *log_group;
static int npreset, presets_alloc;

static struct MUP_key glob_keys[] = {
  {"exe"  , TAG_STRING , {.sp = &exe}},
  {"args" , TAG_STRING , {.sp = &args}},
  {"dir"  , TAG_STRING , {.sp = &dir}},
};

struct pathref {
  char *path;
  int refcount;
};

static struct preset {
  char *spec;
  char *key;
  char *val;
  int lineno;
  struct pathref *pathref;
} *presets;

void MUP_add_keys (struct MUP_key *keys, int nkeys)
{
  if (0 == ntopkeys) {
    topkeys = malloc (nkeys * sizeof (topkeys[0]));
  }
  else {
    topkeys = realloc (topkeys, (ntopkeys + nkeys) * sizeof (topkeys[0]));
  }

  if (NULL == topkeys) {
    ERRFail ("failed to allocate %d bytes for MUP keys",
             (ntopkeys + nkeys) * sizeof (topkeys[0]));
  }
  memcpy (topkeys + ntopkeys, keys, nkeys * sizeof (topkeys[0]));
  ntopkeys += nkeys;
}

void MUP_add_group (struct MUP_group *group)
{
  group->prev = last_group;
  last_group = group;

  if (0 == strcmp (group->prefix, "log.")) {
    log_group = group;
  }
}

static struct MUP_key *find_key (char *name, int nkeys,
                                 struct MUP_key *akeys)
{
  int i;

  for (i = 0; i < nkeys; i++) {
    if (0 == strcmp (name, akeys[i].name))
      return akeys + i;
  }
  return 0;
}

static struct MUP_key *get_key (char *name)
{
  struct MUP_group *group;
  struct MUP_key *key;

  for (group = last_group; group; group = group->prev) {
    int prefix_len;

    prefix_len = strlen (group->prefix);

    if (prefix_len && (0 == strncmp (group->prefix, name, prefix_len))) {

      key = find_key (name + prefix_len, group->nkeys, group->keys);

      if (key)
        return key;
    }
  }

  return find_key (name, ntopkeys, topkeys);
}

static int xstrtol (char *s, int *res)
{
  int ival;
  char *endptr;

  errno = 0;
  ival = strtol (s, &endptr, 0);
  if ((0 != endptr[0]) || (0 != errno)) {
    int err;

    err = errno;
    log ("`%s' is not an integer\n"
         "syserr %s\n",
         s, strerror (err));
    return -1;
  }

  *res = ival;
  return 0;
}

int MUP_get_log_level (char *name)
{
  struct MUP_key *key;

  Fail_if (NULL == log_group);

  key = find_key (name, log_group->nkeys, log_group->keys);

  if (NULL == key) {
    Fail ("log key `%s' does not exist", name);
  }

  if (TAG_LOG_LEVEL != key->tag) {
    Fail ("key `%s' %d is not log level", name, key->tag);
  }

  if (-1 == key->var.ll)
    return default_log_level;
  else
    return key->var.ll;
}

void MUP_add_fs_mapping (int letter, char *path)
{
  struct fs_map *map;

  if (letter < 0 || letter >= MAXDRIVES) {
    Fail ("bougs drive %c(%#x)", letter + 'A', letter + 'A');
  }

  map = fs_map + letter;
  if (map->path)
    free (map->path);

  map->letter = letter;
  map->path = strdup (path);
  if (NULL == map->path) {
    ERRFail ("strdup for mup path failed");
  }
}

char *MUP_get_fs_mapping (int letter)
{
  if (letter >= 0 && letter < MAXDRIVES) {
    return fs_map[letter].path;
  }
  else {
    Fail ("bougs drive %c(%#x)", letter + 'A', letter + 'A');
  }
}

char MUP_get_fs_first_available_letter (int letter_from)
{
  int i;

  if (letter_from < 0 || letter_from >= MAXDRIVES) {
    Fail ("bougs drive %c(%#x)", letter_from + 'A', letter_from + 'A');
  }

  for (i = letter_from; i < MAXDRIVES; i++)
    if (NULL == fs_map[i].path)
      return i + 'A';

  return -1;
}

int MUP_get_int (char *name)
{
  struct MUP_key *key;

  key = get_key (name);

  if (NULL == key) {
    Fail ("key `%s' does not exist", name);
  }

  switch (key->tag) {
    case TAG_INT:
      return *key->var.ip;

    case TAG_IN_INT:
      return key->var.ii;

    default:
      Fail ("type of `%s' %d is not int", name, key->tag);
  }
}

char *MUP_get_string (char *name)
{
  struct MUP_key *key;

  key = get_key (name);

  if (NULL == key) {
    Fail ("key `%s' does not exist", name);
  }

  if (TAG_STRING != key->tag) {
    Fail ("type of `%s' %d is not string", name, key->tag);
  }

  return *key->var.sp;
}

static int parse_log_level (char *s)
{
  if (0 == strcmp (s, "warn")) {
    return LOG_WARN;
  }
  else if (0 == strcmp (s, "info")) {
    return LOG_INFO;
  }
  else if (0 == strcmp (s, "verbose")) {
    return LOG_VERBOSE;
  }
  else if (0 == strcmp (s, "extra.verbose")) {
    return LOG_EXTRA_VERBOSE;
  }
  else if (0 == strcmp (s, "debug")) {
    return LOG_DEBUG;
  }
  else if (0 == strcmp (s, "none")) {
    return LOG_NONE;
  }
  return -1;
}

static char *join_name_dirname (const char *name, const char *template)
{
  int retlen;
  char *retval;
  char *last_slash;

  last_slash = strrchr (template, '/');
  if (NULL == last_slash) {
    return NULL;
  }

  retlen = strlen (name) + (last_slash - template) + 2;
  retval = malloc (retlen);
  if (NULL == retval) {
    ERRFail ("failed to allocate %d bytes", retlen);
  }

  strncpy (retval, template, last_slash - template + 1);
  retval[last_slash - template + 1] = 0;
  strcat (retval, name);

  return retval;
}

static void set_key (char *name, char *val, char *path, int lineno)
{
  struct MUP_key *key;

  key = get_key (name);

  if (NULL == key) {
    Fail ("parsing `%s' line %d,  key `%s' does not exist",
          path, lineno, name);
  }

  switch (key->tag) {
    case TAG_STRING:
    {
      char **sp;

      sp = key->var.sp;

      if (*sp) {
        free (*sp);
      }
      *sp = strdup (val);
      if (0 == *sp) {
        ERRFail ("strdup failed for string value");
      }
      break;
    }

    case TAG_IN_INT:
    case TAG_INT:
    {
      int ival;
      int res;

      res = xstrtol (val, &ival);
      if (-1 == res) {
        Fail ("parsing `%s' line %d `%s' is not an integer",
              path, lineno, val);
      }
      if (TAG_INT == key->tag)
        *key->var.ip = ival;
      else
        key->var.ii = ival;
      break;
    }

    case TAG_SETTER:
    {
      if (-1 == key->var.setter (val)) {
        Fail ("parsing `%s' line %d setter %s(`%s') failed",
              path, lineno, name, val);
      }
      break;
    }

    case TAG_LOG_LEVEL:
    {
      int ll;

      ll = parse_log_level (val);

      if (-1 == ll) {
        Fail ("parsing `%s' line %d `%s' is not valid log level",
              path, lineno, val);
      }

      if (0 == strcmp (name, "log.all")) {
        int i;

        default_log_level = ll;
        if (log_group) {
          for (i = 0; i < log_group->nkeys; i++) {
            struct MUP_key *key;

            key = log_group->keys + i;
            if (TAG_LOG_LEVEL == key->tag)
              key->var.ll = ll;
          }
        }
      }
      else {
        key->var.ll = ll;
      }
      break;
    }
  }
}

static void parse_mup (char *path);

static void parse_file (char *path, FILE *mup)
{
  int lineno;
  char *line;
  int line_size;
  struct pathref *pathref;

  pathref = NULL;
  lineno = 0;
  line_size = 80;

  /* log ("parsing `%s' `%s'\n", get_current_dir_name (), path); */

  line = malloc (line_size);

  if (NULL == line) {
    ERRFail ("parsing `%s' can not alloc %d bytes", path, line_size);
  }

  while (0 == feof (mup)) {
    char *s, *colon;
    char *name, *val;
    int offset;
    int last_non_space;

    lineno++;
    offset = 0;
    last_non_space = 0;

    while (1) {
      int c;

      if (offset == line_size) {
        line_size *= 2;
        line = realloc (line, line_size);

        if (NULL == line) {
          ERRFail ("parsing `%s' can not re-alloc %d bytes for line %d",
                   path, line_size, lineno);
        }
      }

      c = fgetc (mup);

      switch (c) {
        case EOF:
          if (ferror (mup)) {
            ERRFail ("parsing `%s' line %d fgetc failed", path, lineno);
          }

        case '\r':
        case '\n':
          goto got_line;

        case ';':
          do {
            c = fgetc (mup);
          } while (0 == ((EOF == c) || ('\n' == c) || ('\r' == c)));
          ungetc (c, mup);
          break;

        default:
          if (0 == isspace (c))
            last_non_space = offset;

          line[offset++] = c;
          break;
      }
    }

  got_line:
    if (last_non_space + 1 < offset)
      offset = last_non_space + 1;

    line[offset] = 0;

    if (0 == offset)
      continue;

    name = line;
    while (*name && isspace (*name))
      name++;

    Fail_if (0 == *name);

    s = val = strchr (line, '=');

    if ((0 == s) || (s == line)) {
      Fail ("parsing `%s' malformed line %d `%s'", path, lineno, line);
    }

    do {
      val++;
    } while (*val && isspace (*val));

    if (0 == val[0]) {
      Fail ("parsing `%s' malformed line %d `%s'", path, lineno, line);
    }

    do {
      s--;
    } while ((s != line) && isspace (*s));

    s[1] = 0;

    if (0 == name[1]) {
      char dl;

      dl = name[0];
      if (dl >= 'A' && dl <= 'Z') {
        MUP_add_fs_mapping (dl - 'A', val);
        continue;
      }
    }

    if (0 == strcmp (name, "include")) {

      if ('/' == val[0]) {
        parse_mup (val);
      }
      else {
        char *incl_path;

        incl_path = join_name_dirname (val, path);
        if (NULL == incl_path) { /* path does not contain slashes */
          parse_mup (val);
        }
        else {
          parse_mup (incl_path);
          free (incl_path);
        }
      }
      continue;
    }

    colon = strchr (line, ':');

    if (colon && (s - colon > 0)) {
      if (npreset == presets_alloc) {
        if (0 == presets_alloc)
          presets_alloc = 10;
        else
          presets_alloc *= 2;
        presets = realloc (presets, presets_alloc * sizeof (struct preset));

        if (NULL == presets) {
          ERRFail ("realloc of %d bytes failed",
                   presets_alloc * sizeof (struct preset));
        }
      }

      presets[npreset].spec = strndup (name, colon - name);
      if (NULL == presets[npreset].spec)
        ERRFail ("strndup failed for application name");

      name = colon + 1;
      presets[npreset].key = strdup (name);
      if (NULL == presets[npreset].key)
        ERRFail ("strdup failed for per application key");

      presets[npreset].val = strdup (val);
      if (NULL == presets[npreset].val)
        ERRFail ("strdup failed for per application value");

      if (NULL == pathref) {
        pathref = malloc (sizeof (struct pathref));
        if (NULL == pathref)
          ERRFail ("failed to allocate %d bytes for pathref",
                   sizeof (struct pathref));

        pathref->refcount = 0;
        pathref->path = strdup (path);

        if (NULL == pathref->path)
          ERRFail ("strdup for ref path failed");
      }

      presets[npreset].pathref = pathref;
      pathref->refcount += 1;

      npreset += 1;
      continue;
    }

    set_key (name, val, path, lineno);
  }

  /* vbe_mode_post_proc (); */
  free (line);
}

static void parse_mup (char *path)
{
  FILE *mup;

  mup = fopen (path, "r");
  if (NULL == mup) {
    ERRFail ("can not open `%s'", path);
  }

  parse_file (path, mup);
  fclose (mup);
}

static void process_presets ()
{
  int i;
  int len;
  char *spec;

  if (exe) {
    if (args) {
      char *exe_part;
      char *last_slash;

      last_slash = strrchr (exe, '/');
      if ((NULL == last_slash) || (0 == last_slash[1]))
        exe_part = exe;
      else
        exe_part = last_slash + 1;

      len = strlen (exe_part);
      len += strlen (args);
      spec = malloc (len + 2);
      if (NULL == spec) {
        ERRFail ("failed to allocate %d bytes for spec name", len + 1);
      }
      strcpy (spec, exe_part);
      strcat (spec, " ");
      strcat (spec, args);
    }
    else {
      char *exe_part;
      char *last_slash;

      last_slash = strrchr (exe, '/');
      if ((NULL == last_slash) || (0 == last_slash[1]))
        exe_part = exe;
      else
        exe_part = last_slash + 1;

      spec = strdup (exe_part);
      if (NULL == spec) {
        ERRFail ("strdup for spec name failed");
      }
    }
  }
  else spec = NULL;

  if (NULL == spec)
    goto exit;

  for (i = 0; i < npreset; i++) {
    int len;

    len = strlen (presets[i].spec);
    if (0 == strncmp (spec, presets[i].spec, len)) {
      set_key (presets[i].key, presets[i].val,
               presets[i].pathref->path,
               presets[i].lineno);

      free (presets[i].spec);
      free (presets[i].key);
      free (presets[i].val);

      presets[i].pathref->refcount -= 1;
      if (0 == presets[i].pathref->refcount) {
        free (presets[i].pathref->path);
        free (presets[i].pathref);
      }
    }
  }

  free (spec);
 exit:
  if (presets)
    free (presets);
}

/* This function is a nightmare */
void MUP_init (int argc, char **argv)
{
  int in_fd, res;
  struct pollfd pfd;
  char *mup_name;
  char *mup_path;
  char *real_exe;
  char *real_args;

  MUP_add_keys (glob_keys, LENOFA (glob_keys));

  mup_name = "minde.mup";
  real_exe = NULL;
  real_args = NULL;

  if (argc > 1) {
    int i;
    int cmdlen;

    real_exe = strdup (argv[1]);
    if (NULL == real_exe) {
      ERRFail ("strdup failed");
    }

    cmdlen = 0;
    for (i = 2; i < argc; i++) {
      cmdlen += strlen (argv[i]) + 1;
    }

    if (cmdlen) {
      real_args = malloc (cmdlen + 1);
      if (NULL == real_args) {
        ERRFail ("malloc (%d) failed", cmdlen + 1);
      }
      real_args[0] = 0;

      for (i = 2; i < argc; i++) {
        strcat (real_args, argv[i]);
        if (i < argc - 1)
          strcat (real_args, " ");
      }
    }
  }

  mup_path = join_name_dirname (mup_name, argv[0]);
  if (NULL != mup_path) {
    if (0 == access (mup_path, R_OK)) {
      parse_mup (mup_path);
    }
    free (mup_path);
  }

  if (NULL != real_exe) {
    if (NULL == strchr (real_exe, '/')) {
      if (0 == access (mup_name, R_OK)) {
        parse_mup (mup_name);
      }
    }
    else {
      mup_path = join_name_dirname (mup_name, real_exe);

      if (NULL != mup_path) {
        if (0 == access (mup_path, R_OK)) {
          parse_mup (mup_path);
        }
        free (mup_path);
      }
    }

    if (exe) {
      log ("overriding local exe `%s' with `%s'\n", exe, real_exe);
      free (exe);
    }
    exe = real_exe;

    if (NULL != real_args) {
      if (args) {
        log ("overriding local args `%s' with `%s'\n", args, real_args);
        free (args);
      }
      args = real_args;
    }
  }
  else {
    char *old_exe;
    char *old_args;

    if (NULL == exe) {
      if (0 == access (mup_name, R_OK)) {
        parse_mup (mup_name);
      }
      goto parse_stdin;
    }

    old_exe = exe;
    old_args = args;

    exe = NULL;
    args = NULL;

    if (NULL == strchr (old_exe, '/')) {
      if (0 == access (mup_name, R_OK)) {
        parse_mup (mup_name);
      }
    }
    else {
      mup_path = join_name_dirname (mup_name, old_exe);

      if (NULL != mup_path) {
        if (0 == access (mup_path, R_OK)) {
          parse_mup (mup_path);
        }
        free (mup_path);
      }
    }

    if (exe) {
      free (exe);
    }
    exe = old_exe;

    if (args) {
      free (args);
    }
    args = old_args;
  }

 parse_stdin:
  in_fd = fileno (stdin);
  pfd.fd = in_fd;
  pfd.events = POLLIN;
  pfd.revents = 0;

  res = poll (&pfd, 1, 10);
  if (res > 0) {
    parse_file ("<stdin>", stdin);
    fgetc (stdin);
  }
  else if (res == -1) {
    perror ("stdin poll failed");
    exit (EXIT_FAILURE);
  }

  process_presets ();
}
