#include <string.h>

#include <inc/pic.h>
#include <inc/int.h>
#include <inc/eio.h>
#include <inc/bit.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "PIC"
static int log_level;

struct pic_cont {
  int nicw;
  ui icw[4];
  ub rr;
};

static ui isr;
static ui irr;
static ui imr;

static struct pic_cont conts[2];
static int last_write_to;
static PIC_handler pic_handlers[16];

static void log_pic (const char *cap)
{
  DEREF (cap);
  ldebug ("%s\n"
          "1: nicw % d rr %d icw[%#x,%#x,%#x,%#x]\n"
          "2: nicw % d rr %d icw[%#x,%#x,%#x,%#x]\n"
          "  isr %#6x irr %#6x imr %#6x\n",
          cap,
          conts[0].nicw, conts[0].rr,
          conts[0].icw[0], conts[0].icw[1], conts[0].icw[2], conts[0].icw[3],

          conts[1].nicw, conts[1].rr,
          conts[1].icw[0], conts[1].icw[1], conts[1].icw[2], conts[1].icw[3],

          isr, irr, imr
    );
}

static ui read_even (int nport, int size)
{
  int retval;
  int ncont;

  DEREF (size);
  EIO_LOG_READ;

  ncont = (0xa0 == nport);

  if (conts[ncont].rr)
    retval = isr;
  else
    retval = irr;

  return (retval >> (ncont << 3)) & 0xff;
}

static ui read_odd (int nport, int size)
{
  DEREF (size);
  EIO_LOG_READ;

  return imr >> ((0xa1 == nport) << 3);
}

static void ICW1 (int ncont, ui val)
{
  conts[ncont].icw[0] = val;
  conts[ncont].nicw = 1;
  linfo ("ICW1 = 0x%02x\n", val);
}

#define OCW2_IRQ(v) ((v)&7)
#define OCW2_EOI(v) ((v)&0x20)
#define OCW2_RSL(v) ((v)>>6)

static void OCW2 (int ncont, ui val)
{
  int irq;

  if (0 == OCW2_EOI (val)) {
    Fail ("OCW2 is not EOI");
  }

  switch (OCW2_RSL (val)) {
    case 0: {                   /* EOI */
      int misr;
      ldebug ("highest priority EOI\n");
      misr = isr & (0xff << (ncont << 3));

      if (0 == misr) {
        linfo ("EOI ncont %d while no irq in service isr %#x, irr %#x, imr %#x\n",
               ncont, isr, irr, imr);
        return;
      }

      irq = lsbindex (misr);
      isr ^= (1 << irq);

      ldebug ("EOI ncont %d irq %d isr %#x, irr %#x, imr %#x\n",
              ncont, irq, isr, irr, imr);

      if (pic_handlers[irq]) {
        pic_handlers[irq] (1);
      }
      break;
    }

    case 1:                     /* specific EOI */
      irq = OCW2_IRQ (val);
      irq <<= (ncont << 3);

      ldebug ("specific EOI ncont %d for irq %d\n", ncont, irq);
      if (0 == (isr & (1 << irq))) {
        ldebug ("specific EOI ncont %d and irq %d not in service\n"
                "isr %#x, irr %#x, imr %#x\n",
                ncont, irq, isr, irr, imr);
        return;
      }

      isr ^= (1 << irq);

      ldebug ("specific EOI ncont %d irq %d isr %#x, irr %#x, imr %#x\n",
              ncont, irq, isr, irr, imr);

      if (pic_handlers[irq]) {
        pic_handlers[irq] (1);
      }
      break;

    default:
      Fail ("unimplemented OCW2");
  }
}

static void OCW3 (int ncont, ui val)
{
  ldebug ("OCW3 %#x\n", val);

  if (0 != (val >> 5)) {
    Fail ("unimplemented OCW3 %#x", val);
  }

  if (0 != (val & 4)) {
    Fail ("polling mode not supported");
  }

  switch (val & 7) {
    case 0:
    case 1:
      lwarn ("strange OCW3 %#x\n", val);
      break;

    case 2:
      ldebug ("read irr mode\n");
      conts[ncont].rr = 0;
      break;

    case 3:
      ldebug ("read isr mode\n");
      conts[ncont].rr = 1;
      break;
  }
}

static void write_even (int nport, int size, ui val)
{
  int ncont;

  DEREF (size);
  EIO_LOG_WRITE;

  last_write_to = nport;
  ncont = (0xa0 == nport);

  if (val & 0x10)
    ICW1 (ncont, val);
  else if (0 == ((val >> 3) & 3))
    OCW2 (ncont, val);
  else if (8 == (val & 0x98))
    OCW3 (ncont, val);
  else {
    Fail ("(nport=%#x, val=%#x) not ICW1,OCW2 or OCW3", nport, val);
  }
}

static void write_odd (int nport, int size, ui val)
{
  int ncont;
  struct pic_cont *cont;

  DEREF (size);
  EIO_LOG_WRITE;

  ncont = (0xa1 == nport);
  cont = conts + ncont;

  if ((cont->nicw > 0) && (1 == nport - last_write_to)) {
    if (cont->nicw < 0) {
      Fail ("next icw == 4");
    }

    linfo ("0x%02x ICW%d = 0x%02x\n", nport, cont->nicw + 1, val);

    if ((1 == ncont) && (1 == cont->nicw))
      val -= 8;

    cont->icw[cont->nicw++] = val;

    switch (cont->nicw) {
      case 2:
        if (cont->nicw & 2) {
          if (0 == (cont->nicw & 1))
            cont->nicw = -1;
          else
            cont->nicw = 3;
        }
        break;

      case 3:
        if (0 == (cont->nicw & 1))
          cont->nicw = -1;
    }
  }
  else {
    int shift;

    shift = ncont << 3;
    imr = (val << shift) | (imr & (0xff << (8 - shift)));;
  }

  log_pic (__func__);
  return;
}

void PIC_request_service (ub irq)
{
  int mask;

  mask = 1 << irq;

  ldebug ("isr %#x irr %#x imr %#x\n", isr, irr, imr);

  if (isr & mask) {
    linfo ("irq %d already in service\n", irq);
    return;
  }

  if (irr & mask) {
    linfo ("irq %d already requested\n", irq);
    return;
  }

  if (imr & mask) {
    linfo ("irq %d is masked\n", irq);
    return;
  }

  ldebug ("requesting %d mask %d\n", irq, mask);
  irr |= mask;
}

int PIC_irq_pending ()
{
  /* Assert (INT_is_enabled () == 0); */

  return irr && ((0 == isr) || (LEASTBIT (irr) < LEASTBIT (isr)));
}

int PIC_run ()
{
  int retval;
  static int in_pic;

  Fail_if (in_pic);
  in_pic = 1;
  retval = -1;

  if (irr && ((0 == isr) || (LEASTBIT (irr) < LEASTBIT (isr)))) {
    int irq;
    ui lowbitmask;

    lowbitmask = LEASTBIT (irr);
    irq = popcount (lowbitmask - 1);

    ldebug ("will service irq %d\n", irq);

    irr ^= lowbitmask;
    isr |= lowbitmask;

    if (pic_handlers[irq])
      pic_handlers[irq] (0);

    retval = irq + conts[irq > 7].icw[1];
  }

  in_pic = 0;
  ldebug ("int %#x\n", retval);
  return retval;
}

void PIC_eoi (int ncont)
{
  write_even (0x20 + (ncont << 7), 1, 0x20);
}

void PIC_register (ub irq, PIC_handler pic_handler)
{
  if (irq >= 16) {
    Fail ("(irq=%d) irq greater than 16", irq);
  }

  pic_handlers[irq] = pic_handler;
}

void PIC_init ()
{
  log_level = MUP_get_log_level ("pic");

  EIO_register_write_handler (0x20, EIO_MASK_B, write_even);
  EIO_register_write_handler (0xa0, EIO_MASK_B, write_even);
  EIO_register_write_handler (0x21, EIO_MASK_B, write_odd);
  EIO_register_write_handler (0xa1, EIO_MASK_B, write_odd);

  EIO_register_read_handler (0x20, EIO_MASK_B, read_even);
  EIO_register_read_handler (0xa0, EIO_MASK_B, read_even);
  EIO_register_read_handler (0x21, EIO_MASK_B, read_odd);
  EIO_register_read_handler (0xa1, EIO_MASK_B, read_odd);

  write_even (0x20, 1, 0x12);
  write_odd (0x21, 1, 0x08);

  write_even (0xa0, 1, 0x12);
  write_odd (0xa1, 1, 0x70);
}
