#include <stdio.h>
#include <string.h>

#include <inc/sig.h>
#include <inc/int.h>
#include <inc/sb16.h>
#include <inc/eio.h>
#include <inc/dma.h>
#include <inc/log.h>
#include <inc/pic.h>
#include <inc/mup.h>
#include <inc/aud.h>
#include <inc/bit.h>
#include <inc/dos.h>

#define LOG_PREFIX "SB16"
static int log_level;

static struct {
  int ver_lo;
  int ver_hi;
  int irq;
  int dma;
  int hdma;
  int port;
  int mix_block;
} sb = {4, 5, 5, 1, 5, 0x220, -1};

static int mix_block, noirq;

static struct mixer {
  int nreg;
  ub regs[0x83];
} mixer;

static struct dsp {
  int in_index;
  int out_data_len;
  int fmt_stereo;
  int fmt_signed;
  int fmt_bits;
  int dma_auto;
  int dma_buffer_size;
  int fifo;
  int freq;
  int time_const;
  int speaker;
  int needed_bytes;
  int cmd;
  int dma_pos;
  int use_hdma;

  int v2x6;

  ub in_data[10];
  ub out_data[10];

  int left_till_irq;
} dsp;

#define nocmd ~0

static void log_dsp (const char *cap)
{
  DEREF (cap);

  linfo ("%c:%c:%d:%c:dmabuf=%d:pos=%d:freq=%d:timeconst=%d:speaker=%d\n",
         dsp.fmt_stereo ? 'S' : 'M',
         dsp.fmt_signed ? 'S' : 'U',
         dsp.fmt_bits,
         dsp.dma_auto ? 'a' : 's',
         dsp.dma_buffer_size,
         dsp.dma_pos,
         dsp.freq,
         dsp.time_const,
         dsp.speaker);
}

static void control (int hold)
{
  linfo ("%d high %d\n", hold, dsp.use_hdma);
  if (hold) {
    if (dsp.use_hdma)
      DMA_hold_DREQ (sb.hdma);
    else
      DMA_hold_DREQ (sb.dma);
  }
  else {
    if (dsp.use_hdma)
      DMA_release_DREQ (sb.hdma);
    else
      DMA_release_DREQ (sb.dma);
  }
}

/* static void dma_cmd (ub cmd, ub d0, ub d1, ub d2) */
static void dma_cmd (ub cmd, ub d0, int dma_len)
{
  int bps;
  audfmt_e fmt;

  dsp.use_hdma = cmd < 0xc0;
  dsp.fifo = (cmd >> 1) & 1;
  dsp.dma_auto = (cmd >> 2) & 1;

  switch (cmd >> 4) {
    case 11:
      dsp.fmt_bits = 16;
      break;

    case 12:
      dsp.fmt_bits = 8;
      break;
  }

  dsp.fmt_signed = (d0 >> 4) & 1;
  dsp.fmt_stereo = (d0 >> 5) & 1;

  if (-1 != dsp.time_const) {
    int tmp;

    tmp = 256 - dsp.time_const;
    dsp.freq = (1000000 + (tmp / 2)) / tmp;
  }
  bps = 1 << (16 == dsp.fmt_bits);

  if (-1 != dma_len)
    dsp.dma_buffer_size = (dma_len + 1) * bps;

  linfo ("frequency %d, stereo %d, signed %d, bits %d, size %d, auto %d\n",
         dsp.freq, dsp.fmt_stereo, dsp.fmt_signed, dsp.fmt_bits,
         dsp.dma_buffer_size, dsp.dma_auto);

  if (16 == dsp.fmt_bits) {
    if (dsp.fmt_signed) {
      fmt = AUD_FMT_S16;
    }
    else {
      fmt = AUD_FMT_U16;
    }
  }
  else {
    if (dsp.fmt_signed) {
      fmt = AUD_FMT_S8;
    }
    else {
      fmt = AUD_FMT_U8;
    }
  }

  dsp.dma_pos = 0;
  dsp.left_till_irq = dsp.dma_buffer_size;

  if (sb.mix_block) {
    mix_block = sb.mix_block;
  }
  else {
    int align;

    align = bps << dsp.fmt_stereo;
    mix_block = ((dsp.freq * align) / 100) & ~(align - 1);
  }

  AUD_reset (dsp.freq, 1 << dsp.fmt_stereo, fmt);
  control (1);
  dsp.speaker = 1;
}

static void command (ub cmd)
{
  char *msg;

  msg = (char *)-1;

  linfo ("%#x\n", cmd);

  if (cmd > 0xaf && cmd < 0xd0) {
    if (cmd & 8)
      goto error;

    switch (cmd >> 4) {
      case 11:
      case 12:
        break;
      default:
        msg = "wrong bits";
        goto error;
    }
    dsp.needed_bytes = 3;
  }
  else {
    switch (cmd) {
      case 0x10:
        dsp.needed_bytes = 1;
        break;

      case 0x14:
        dsp.needed_bytes = 2;
        dsp.dma_buffer_size = 0;
        break;

      case 0x20:
        dsp.out_data[dsp.out_data_len++] = 0xff;
        break;

      case 0x35:
        lwarn ("MIDI commands not implemented\n");
        break;

      case 0x40:
        dsp.freq = -1;
        dsp.time_const = -1;
        dsp.needed_bytes = 1;
        break;

      case 0x41:
      case 0x42:
        dsp.freq = -1;
        dsp.time_const = -1;
        dsp.needed_bytes = 2;
        break;

      case 0x47:                /* Continue Auto-Initialize DMA 16bit */
        break;

      case 0x48:
        dsp.needed_bytes = 2;
        break;

      case 0x27:                /* ????????? */
      case 0x4e:
        return;

      case 0x80:
        cmd = nocmd;
        break;

      case 0x90:
      case 0x91:
      {
        ub d0;

        d0 = 4;
        if (dsp.fmt_signed) d0 |= 16;
        if (dsp.fmt_stereo) d0 |= 32;
        dma_cmd (cmd == 0x90 ? 0xc4 : 0xc0, d0, -1);
        cmd = nocmd;
        break;
      }

      case 0xd0:                /* XXX */
        control (0);
        return;

      case 0xd1:
        dsp.speaker = 1;
        break;

      case 0xd3:
        dsp.speaker = 0;
        return;

      case 0xd4:
        control (1);
        break;

      case 0xd5:
        control (0);
        break;

      case 0xd6:
        control (1);
        break;

      case 0xd9:
        control (0);
        dsp.dma_auto = 0;
        return;

      case 0xda:
        control (0);
        dsp.dma_auto = 0;
        break;

      case 0xe0:
        dsp.needed_bytes = 1;
        break;

      case 0xe1:
        dsp.out_data[dsp.out_data_len++] = sb.ver_lo;
        dsp.out_data[dsp.out_data_len++] = sb.ver_hi;
        return;

      case 0xf2:
        dsp.out_data[dsp.out_data_len++] = 0xaa;
        mixer.regs[0x82] |= 1;
        PIC_request_service (sb.irq);
        return;

      default:
        msg = "is unknown";
        goto error;
    }
  }
  dsp.cmd = cmd;
  return;

 error:
  Fail ("%#x %s", cmd, msg);
  return;
}

static void complete ()
{
  linfo ("complete command %#x, in_index %d, needed_bytes %d\n",
         dsp.cmd, dsp.in_index, dsp.needed_bytes);

  if (dsp.cmd > 0xaf && dsp.cmd < 0xd0) {
    int d0, d1, d2;

    d0 = dsp.in_data[0];
    d1 = dsp.in_data[1];
    d2 = dsp.in_data[2];

    ldebug ("d0 = %d, d1 = %d, d2 = %d\n",
            d0, d1, d2);
    dma_cmd (dsp.cmd, d0, d1 + (d2 << 8));
  }
  else {
    switch (dsp.cmd) {

      case 0x10:
        break;

      case 0x14:
      {
        int d0, d1;
        int save_left;
        int save_pos;

        d0 = dsp.in_data[0];
        d1 = dsp.in_data[1];

        save_left = dsp.left_till_irq;
        save_pos = dsp.dma_pos;
        dma_cmd (0xc0, 0, d0 + (d1 << 8));
        dsp.left_till_irq = save_left;
        dsp.dma_pos = save_pos;

        linfo ("set buffer size data[%d, %d] %d pos %d\n",
               d0, d1, dsp.dma_buffer_size, dsp.dma_pos);
        break;
      }

      case 0x40:
        dsp.time_const = dsp.in_data[0];
        linfo ("set time const %d\n", dsp.time_const);
        break;

      case 0x41:
      case 0x42:
        dsp.freq = dsp.in_data[1] + (dsp.in_data[0] << 8);
        linfo ("set freq %#x, %#x = %d\n",
               dsp.in_data[1], dsp.in_data[0], dsp.freq);
        break;

      case 0x48:
        dsp.dma_buffer_size = dsp.in_data[1] + (dsp.in_data[0] << 8);
        linfo ("set dma len %#x, %#x = %d\n",
               dsp.in_data[1], dsp.in_data[0], dsp.dma_buffer_size);
        break;

      case 0xe0:
        dsp.out_data_len = 1;
        linfo ("data = %#x\n", dsp.in_data[0]);
        dsp.out_data[0] = dsp.in_data[0] ^ 0xff;
        break;

      default:
        goto error;
    }
  }

  dsp.cmd = -1;
  return;

 error:
  Fail ("unrecognized command %#x", dsp.cmd);
}

static void DSP_write (int nport, int size, ui val)
{
  int iport;

  /* EIO_LOG_WRITE; */

  iport = nport - sb.port;

  switch (iport) {

    case 0x6:
      if (0 == val)
        dsp.v2x6 = 0;
      else if ((1 == val) && (0 == dsp.v2x6)) {
        dsp.v2x6 = 1;
        dsp.out_data[dsp.out_data_len++] = 0xaa;
      }
      else
        dsp.v2x6 = ~0;
      break;

    case 0xc:                   /* write data or command | write status */
      if (0 == dsp.needed_bytes) {
        command (val);
        if (0 == dsp.needed_bytes) {
          log_dsp (__func__);
        }
      }
      else {
        dsp.in_data[dsp.in_index++] = val;
        if (dsp.in_index == dsp.needed_bytes) {
          dsp.needed_bytes = 0;
          dsp.in_index = 0;
          complete ();
          log_dsp (__func__);
        }
      }
      break;

    default:
      Fail ("(nport=%#x, size=%d, val=%#x)", nport, size, val);
  }
}

static ui DSP_read (int nport, int size)
{
  char *msg;
  int iport, retval;

  msg = (char *) -1;
  iport = nport - sb.port;

  switch (iport) {

    case 0x6:                   /* reset */
      return 0;

    case 0xa:                 /* read data */
      if (dsp.out_data_len) {
        retval = dsp.out_data[--dsp.out_data_len];
      }
      else {
#if 1
        lwarn ("empty output buffer\n");
        retval = 0;
#else
        msg = "empty output buffer";
        goto error;
#endif
      }
      break;

    case 0xc:                 /* 0 can write */
      retval = 0;
      break;

    case 0xd:                 /* timer interrupt clear */
      goto error;

    case 0xe:                 /* data available status | irq 8 ack */
      retval = (0 == dsp.out_data_len) ? 0 : 0x80;
      break;

    case 0xf:                 /* irq 16 ack */
      retval = 0xff;
      mixer.regs[0x82] &= ~2;
      ldebug ("16 ack\n");
      break;

    default:
      goto error;
  }

  if ((0xc != iport) && (0xe != iport)) {
    ldebug ("(nport=%#x, size=%d) iport %#x = %#x\n",
            nport, size, iport, retval);
  }

  return retval;

 error:
  Fail ("(nport=%#x, size=%d) %s",  nport, size, msg);
}

static void MIXER_write_index (int nport, int size, ui val)
{
  DEREF (nport);

  EIO_LOG_WRITE;
  mixer.nreg = val & 0xff;
  if (2 == size) {
    mixer.regs[mixer.nreg] = val >> 16;
  }
}

/*
  Hack hack
  works only with irq 5 and dma 5
  factory settings so to speak
*/
static void MIXER_write (int nport, int size, ui val)
{
  DEREF (nport);
  DEREF (size);

  EIO_LOG_WRITE;
  mixer.regs[mixer.nreg] = val;
}

static ui MIXER_read (int nport, int size)
{
  ub retval;

  DEREF (nport);
  DEREF (size);

  EIO_LOG_READ;

  switch (mixer.nreg) {
    case 0x80:
      retval = 2;               /* for irq 5 */
      break;

    case 0x81:
      retval = 0x20 | (sb.dma << 1);
      break;

    case 0x82:
      retval = mixer.regs[0x82];
      ldebug ("retval = %d\n", retval);
      break;

    case 0x0e:
      retval = ~0;
      break;

    default:
      /* Fail ("unknown mixer register %#x", mixer.nreg); */
      lwarn ("unknown mixer register %#x\n", mixer.nreg);
      retval = mixer.regs[mixer.nreg];
      break;
  }
  return retval;
}

void SB16_run ()
{
  if (0 == dsp.speaker)
    return;

  AUD_run ();
}

static int write_audio (ui addr, int len, int size)
{
  int temp, net;

  temp = size;

  net = 0;

  while (temp) {
    int left_till_end;
    int to_copy;
    int copied;

    left_till_end = len - dsp.dma_pos;

    to_copy = MIN (temp, left_till_end);

    copied = AUD_write ((void *) (addr + dsp.dma_pos), to_copy);

    temp -= copied;
    dsp.dma_pos += copied;

    if (dsp.dma_pos == len) {
      dsp.dma_pos = 0;
    }

    net += copied;

    if (copied != to_copy)
      return net;
  }

  return net;
}

static int SB_read_DMA (ui addr, int size, int *_irq)
{
  int free, till, copy, written;

  if (0 == dsp.speaker)
    return 0;

  if (dsp.left_till_irq < 0) {
    dsp.left_till_irq += dsp.dma_buffer_size;
    return dsp.dma_pos;
  }

  free = AUD_get_free ();

  if ((free <= 0) || (0 == size)) {
    return dsp.dma_pos;
  }

  if (mix_block > 0) {
    copy = MIN (free, mix_block);
  }
  else {
    copy = free;
  }

  till = dsp.left_till_irq;

  ldebug ("addr:%#010x free:%d till:%d size:%d\n",
          addr, free, till, size);
/*   linfo ("pos %d free %d size %d till %d copy %d auto %d noirq %d\n", */
/*       dsp.dma_pos, free, size, till, copy, dsp.dma_auto, noirq); */
  if (till <= copy) {
    if (0 == dsp.dma_auto) {
      copy = till;
    }
  }

  written = write_audio (addr, size, copy);
  dsp.left_till_irq -= written;
  AUD_adjust_estimate (free - written);

  if (dsp.left_till_irq <= 0) {
    mixer.regs[0x82] |= 2;
    if (0 == noirq)
      *_irq = sb.irq;

    if (0 == dsp.dma_auto) {
      control (0);
    }
  }

  ldebug ("pos %5d free %5d size %5d till % 5d copy %5d dma size %5d\n",
          dsp.dma_pos, free, size, dsp.left_till_irq, copy,
          dsp.dma_buffer_size);

  if (dsp.left_till_irq <= 0) {
    dsp.left_till_irq += dsp.dma_buffer_size;
  }

  return dsp.dma_pos;
}

static int dma_misc_handler (int moo)
{
  DEREF (moo);

  return -1;
}

static void IRQ (INT_ARGS)
{
  DEREF (r);

  lwarn ("this handler should not be called\n");

  dsp.speaker = 0;
  control (0);
  PIC_eoi (0);
}

static void pic_handler (int which)
{
  if (which) {
    ldebug ("pic_handler\n");
  }
}

static ui nada (int nport, int size)
{
  DEREF (nport);
  DEREF (size);

  return 0;
}

void SB16_pre_init ()
{
  static struct MUP_key keys[] = {
    {"port"      , TAG_INT , {.ip = &sb.port}},
    {"irq"       , TAG_INT , {.ip = &sb.irq}},
    {"dma"       , TAG_INT , {.ip = &sb.dma}},
    {"hdma"      , TAG_INT , {.ip = &sb.hdma}},
    {"mix.block" , TAG_INT , {.ip = &sb.mix_block}},
    {"ver.lo"    , TAG_INT , {.ip = &sb.ver_lo}},
    {"ver.hi"    , TAG_INT , {.ip = &sb.ver_hi}},
  };

  static struct MUP_group group =
    {"sb.", LENOFA (keys), keys, NULL};

  MUP_add_group (&group);
}

void SB16_init ()
{
  int i;
  char env_var[256];
  static const ub dsp_write_ports[] = {0x6, 0xc};
  static const ub dsp_read_ports[] = {0x6, 0xa, 0xc, 0xd, 0xe, 0xf};

  log_level = MUP_get_log_level ("sb16");

  if (0 == MUP_get_int ("audio.no.env")) {
    snprintf (env_var, 256, "BLASTER=A%3x I%d D%d H%d T6",
              sb.port, sb.irq, sb.dma, sb.hdma);
    DOS_put_env (env_var);
  }

  noirq = MUP_get_int ("audio.no.irq");

  for (i = 0x30; i < 0x48; i++) {
    mixer.regs[i] = 0x20;
  }

  for (i = 0; i < LENOFA (dsp_write_ports); i++) {
    EIO_register_write_handler (sb.port + dsp_write_ports[i],
                                EIO_MASK_B, DSP_write);
  }

  for (i = 0; i < LENOFA (dsp_read_ports); i++) {
    EIO_register_read_handler (sb.port + dsp_read_ports[i],
                               EIO_MASK_B, DSP_read);
  }

  EIO_register_write_handler (sb.port + 0x4, EIO_MASK_BW, MIXER_write_index);
  EIO_register_read_handler (sb.port + 0x5, EIO_MASK_B, MIXER_read);
  EIO_register_write_handler (sb.port + 0x5, EIO_MASK_B, MIXER_write);

  EIO_register_read_handler (0x300, EIO_MASK_B, nada);

  INT_set (sb.irq + 8, IRQ);

  PIC_register (sb.irq, pic_handler);
  DMA_register_channel (sb.hdma, SB_read_DMA, dma_misc_handler);
  DMA_register_channel (sb.dma, SB_read_DMA, dma_misc_handler);
}
