#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/mman.h>

#include <inc/tid.h>
#include <inc/pie.h>
#include <inc/dos.h>
#include <inc/ldt.h>
#include <inc/asm.h>
#include <inc/sel.h>
#include <inc/sig.h>
#include <inc/int.h>
#include <inc/pit.h>
#include <inc/pic.h>
#include <inc/rtc.h>
#include <inc/vid.h>
#include <inc/evt.h>
#include <inc/x86.h>
#include <inc/red.h>
#include <inc/dpmi.h>
#include <inc/sou.h>
#include <inc/xms.h>
#include <inc/log.h>
#include <inc/mup.h>

static int log_level;
#define LOG_PREFIX "SIG"

#define SIGAX(r) GETLOWORD(SIGEAX(r))
#define SIGBX(r) GETLOWORD(SIGEBX(r))
#define SIGCX(r) GETLOWORD(SIGECX(r))
#define SIGDX(r) GETLOWORD(SIGEDX(r))
#define SIGSI(r) GETLOWORD(SIGESI(r))
#define SIGDI(r) GETLOWORD(SIGEDI(r))
#define SIGBP(r) GETLOWORD(SIGEBP(r))

static int r_lock;
#define r_lock() do{if(r_lock)abort();else r_lock=1;}while(0)
#define r_unlock() do{if(!r_lock)abort();else r_lock=0;}while(0)

struct unicontext {
  int signum;
  int trapno;
  int err;
  int cr2;
  sigset_t set;
  regs86 *regs;
};

union sigunion { regs86 regs; ub data[CONTEXT_SZ]; };

static struct {
  int x86, dpmi, vm86, sig, pic_loop;
} in;

static struct {
  us locked_stack_sel;
  ui locked_stack_base;
  ui locked_stack_top;
  us rm_stack_alias_sel;
  ui last_esp;
  ui last_rmc_ss_alias_base;
  int use_locked_stack;
  int flat_locked_stack;
} sinfo;

static struct {
  ul sampling_period;
  int nsamples;
  int pie_threshold;
  int vga_threshold;
  int nfaults;
  ul last_fault_tid;
} finfo = {100, 400, 1500, 1500, 0, 0};

static int sig_stack_size = 16384;
static struct vm86plus_struct *vm86p;
static int debugger_present = -1;
static char *debugger_command;
static union sigunion ctx;
static int need_confirmation;
static struct SIG_mouse_info minfo;

static void copy_context (void *dst, void *src);

typedef void (* int_trampoline) (ui nint, intregs r);

void SIG_call_int_trampoline (int_trampoline tramp, ui nint, ui esp, us ss);
void SIG_call_int_trampoline (int_trampoline tramp, ui nint, ui esp, us ss)
{
  ui *r;
  ui frame_addr;

  if (0 == (ss & 4)) {
    frame_addr = esp;
  }
  else {
    if (sinfo.locked_stack_sel == ss) {
      frame_addr = esp + sinfo.locked_stack_base;
    }
    else {
      frame_addr = SELLIN (ss, esp);
    }
  }

  linfo ("nint %#x ss:esp %04x %08x frame_addr %#x\n",
          nint, ss, esp, frame_addr);

  r = (ui *) frame_addr;
  r[INT_REG_CS] &= 0xffff;
  r[INT_REG_ES] &= 0xffff;
  r[INT_REG_DS] &= 0xffff;
  r[INT_REG_FS] &= 0xffff;
  r[INT_REG_GS] &= 0xffff;

  tramp (nint, r);
}

static int is_debugger_present ()
{
  FILE *f;
  char buf[256];
  pid_t ppid;

  ppid = getppid ();
  snprintf (buf, sizeof (buf), "/proc/%d/cmdline", ppid);

  f = fopen (buf, "r");
  if (NULL == f)
    return 0;

  fgets (buf, 256, f);
  fclose (f);

  return NULL != strstr (buf, "gdb");
}

static int attach_debugger ()
{
  int err;
  pid_t pid, cpid;
  char buf[8];
  char *argv[3];

  if (debugger_command) {
    pid = getpid ();
    snprintf (buf, sizeof (buf), "%#x", pid);

    cpid = fork ();
    if (-1 == cpid) {
      err = errno;
      log ("fork failed\n"
           "syserr: `%s'\n", strerror (err));
      return -1;
    }

    argv[0] = debugger_command;
    argv[1] = buf;
    argv[2] = NULL;

    if (0 == cpid) {
      if (-1 == execve (debugger_command, argv, environ)) {
        err = errno;
        log ("execl failed\nsyserr: `%s'\n", strerror (err));
      }
      exit (EXIT_FAILURE);
    }

    return cpid;
  }
  else {
    return -1;
  }
}

static void release_stack_chunk (regs86 *r)
{
  DEREF (r);

  if (in.dpmi) {
    linfo ("adjusting stack old %#x new %#x\n",
           ASM_stack_top, ASM_stack_top + ASM_stack_chunk_size);

    ASM_stack_top += ASM_stack_chunk_size;
  }
}

static void switch_stacks (regs86 *r)
{
  if (in.dpmi) {
    linfo ("switching stacks old %#x new %#x\n",
           ASM_stack_top, ASM_stack_top - ASM_stack_chunk_size);

    SIGSS (r) = ASM_ss;
    SIGESP (r) = ASM_stack_top;

    ASM_stack_top -= ASM_stack_chunk_size;
  }
}

void SIG_log_regs (regs86 *r)
{
  char m[X86_EFLAGS_BUF_SIZE];

  log ("sigregs:\n"
       "eax=%#010x ebx=%#010x ecx=%#010x edx=%#010x\n"
       "esi=%#010x edi=%#010x ebp=%#010x esp=%#010x\n"
       "eip=%#010x efl=%#010x %s\n"
       "cs=%#06x ds=%#06x es=%#06x ss=%#06x "
       "fs=%#06x gs=%#06x\n",
       (int) SIGEAX(r), (int) SIGEBX(r), (int) SIGECX(r), (int) SIGEDX(r),
       (int) SIGESI(r), (int) SIGEDI(r), (int) SIGEBP(r), (int) SIGESP(r),
       (int) SIGEIP(r), (int) SIGEFL(r), X86_efltoa (SIGEFL (r), m),
       SIGCS(r), SIGDS(r), SIGES(r), SIGSS(r), SIGFS(r), SIGGS(r)
    );

  if (in.vm86) {
    struct vm86_regs *vr;

    vr = &vm86p->regs;
    log ("vm86 regs:\n"
         "eax=%#010x ebx=%#010x ecx=%#010x edx=%#010x\n"
         "esi=%#010x edi=%#010x ebp=%#010x esp=%#010x\n"
         "eip=%#010x efl=%#010x %s\n"
         "cs=%#06x ds=%#06x es=%#06x ss=%#06x "
         "fs=%#06x gs=%#06x\n",
         (int) vr->eax, (int) vr->ebx, (int) vr->ecx, (int) vr->edx,
         (int) vr->esi, (int) vr->edi, (int) vr->ebp, (int) vr->esp,
         (int) vr->eip, (int) vr->eflags, X86_efltoa (vr->eflags , m),
         vr->cs, vr->ds, vr->es, vr->ss, vr->fs, vr->gs);
  }

  if (in.x86) {
    X86_dump_current ();
  }
}

int SIG_is_in_pic_loop ()
{
  return in.pic_loop;
}

static void pic_loop ()
{
  int nint;
  us irq_ss;
  ui irq_esp;

  if (in.dpmi) {
    int use_locked_stack;

    use_locked_stack =
      sinfo.use_locked_stack
      || (ASM_ss == SIGSS (&ctx.regs));

    if (use_locked_stack) {
      irq_ss  = sinfo.locked_stack_sel;
      irq_esp = sinfo.locked_stack_top;
    }
    else {
      irq_ss = SIGSS (&ctx.regs);
      irq_esp = SIGESP (&ctx.regs);
    }
  }
  else {
    irq_ss = 0;
    irq_esp = 0;
  }

  ldebug ("serving on %#06x %#010x\n", irq_ss, irq_esp);

  for (;;) {
    nint = PIC_run ();
    if (-1 == nint) {
      leverbose ("exiting\n");

      if (minfo.pending) {
        struct vm86_regs r;

        minfo.pending = 0;
        r.eax = minfo.mask;
        r.ebx = minfo.state;
        r.ecx = minfo.xpos;
        r.edx = minfo.ypos;
        r.esi = minfo.hor_mickeys;
        r.edi = minfo.ver_mickeys;
        r.cs  = minfo.seg;
        r.eip = minfo.off;
        r.eflags = 0;
        SIG_vm86 (&r, 1);
      }

      release_stack_chunk (&ctx.regs);
      raise (SIGUSR1);
    }
    else {
      us sel;
      ui off;

      leverbose ("nint %d\n", nint);
      INT_get_pm (nint, &sel, &off);

      ASM_simulate_irq (off, sel,
                        irq_esp,
                        irq_ss,
                        SIGFS (&ctx.regs),
                        SIGGS (&ctx.regs));

      leverbose ("completed for nint %d\n", nint);
    }
  }
}

static void keeping ()
{
  SOU_run ();

  if (0 == in.pic_loop) {
    PIT_run ();
    RTC_run ();
  }

  if (0 == r_lock) {
    r_lock = 1;
    VID_run ();
    EVT_run ();
    r_lock = 0;
  }
}

void SIG_break (int fatal)
{
  int was_blocked;

  TID_pause ();
  was_blocked = SIG_block ();
  VID_windowed ();

  if (-1 == debugger_present)
    exit (EXIT_FAILURE);

  if (debugger_present) {
    BPT;
  }
  else {
    int status;
    pid_t pid;

    pid = attach_debugger ();
    if (-1 == pid) {
      exit (EXIT_FAILURE);
    }
    if (fatal) waitpid (pid, &status, 0);
  }

  if (fatal) {
    exit (EXIT_FAILURE);
  }

  sleep (1);
  SIG_unblock (was_blocked);
  TID_resume ();
}

void SIG_wait ()
{
  sigset_t set;

  sigprocmask (SIG_BLOCK, NULL, &set);
  Fail_if (sigismember (&set, SIGALRM));
  sigsuspend (&set);
}

int SIG_block ()
{
  sigset_t set;
  sigset_t old_set;

  sigemptyset (&set);
  sigaddset (&set, SIGALRM);
  sigprocmask (SIG_BLOCK, &set, &old_set);
  return sigismember (&old_set, SIGALRM);
}

void SIG_unblock (int was_blocked)
{
  sigset_t set;

  if (0 == was_blocked) {
    sigemptyset (&set);
    sigaddset (&set, SIGALRM);
    sigprocmask (SIG_UNBLOCK, &set, 0);
  }
}

void SIG_mouse_callback (struct SIG_mouse_info *info)
{
  memcpy (&minfo, info, sizeof (minfo));
  minfo.pending = 1;
}

static void jump (regs86 *r)
{
  copy_context (r, ctx.data);
  lverbose ("jumping to pc %#06x %#010x sp %#06x %#010x\n",
            (int) SIGCS (&ctx.regs), (int) SIGEIP (&ctx.regs),
            (int) SIGSS (&ctx.regs), (int) SIGESP (&ctx.regs));
}

#define INSIDE_VM86_SYSCALL(r) \
  ((ASM_cs == SIGCS (r)) && (SIGEIP (r) == (reg86) ASM_vm86_ret))

static void maybe_service_irqs (regs86 *r)
{
  int pending, enabled, vm_enabled;

  enabled = INT_is_enabled ();
  pending = PIC_irq_pending ();

  if (in.vm86) {
    vm_enabled = vm86p->regs.eflags & X86_EFLAGS_VIF;
  }
  else {
    vm_enabled = -1;
  }

  ldebug ("in.pic_loop %d, enabled %d, vm_enabled %d, pending %d\n",
          in.pic_loop, enabled, vm_enabled, pending);

  if ((0 == in.pic_loop) && enabled) {
    if (pending) {
      in.pic_loop = 1;
      INT_disable ();

      copy_context (ctx.data, r);
      switch_stacks (r);

      SIGES (r) = ASM_es;
      SIGCS (r) = ASM_cs;
      SIGDS (r) = ASM_ds;
      SIGFS (r) = ASM_fs;
      SIGGS (r) = ASM_gs;

      SIGEIP (r) = (ui) pic_loop;
      SIGEFL (r) &= ~(X86_EFLAGS_DF | X86_EFLAGS_TF);
    }
  }
}

static int is_fault_freq_high ()
{
  int high;
  ul delta;

  high = 0;
  delta = TID_get_abs_nsec () - finfo.last_fault_tid;
  finfo.nfaults++;

  if (delta > finfo.sampling_period) {
    high = finfo.nfaults > finfo.nsamples;
    finfo.nfaults = 0;
    finfo.last_fault_tid += delta;
  }

  return high;
}

static int runinterp_vm86 (int threshold, int enable_alarm, struct vm86_regs *r)
{
  x86exit_e exit;
  sigset_t set;

  in.x86 = 1;

  if (enable_alarm) {
    sigemptyset (&set);
    sigaddset (&set, SIGALRM);
    sigprocmask (SIG_UNBLOCK, &set, NULL);
  }

  exit = X86_run16 (r, threshold);

  if (enable_alarm) {
    sigprocmask (SIG_BLOCK, &set, NULL);
  }

  in.x86 = 0;

  switch (exit) {
    case X86_EXIT_TEMPORARY:
      break;

    case X86_EXIT_REQUESTED:
      break;

    case X86_EXIT_BAD_INSN:
    case X86_EXIT_THRESHOLD:
      break;
  }
  return exit;
}

static int maybe_dpmi_callback (ui red, struct vm86_regs *r)
{
  struct DPMI_callback *cb;
  us ds, es;
  ui esi, edi, rm_stack_base;
  ub *data;

  cb = DPMI_get_callback (red - RED_CALLBACK_0);
  if (NULL == cb)
    return -1;

  Fail_if (0 == in.dpmi);

  rm_stack_base = MEM_REAL (r->ss << 4);

  if (sinfo.last_rmc_ss_alias_base != rm_stack_base) {
    LDT_set_selector_base (sinfo.rm_stack_alias_sel, rm_stack_base);
    sinfo.last_rmc_ss_alias_base = rm_stack_base;
  }

  ds = sinfo.rm_stack_alias_sel;
  esi = GETLOWORD (r->esp);
  es = cb->data_sel;
  edi = cb->data_off;
  data = (ub *) SELLIN (es, edi);

  DPMI_EAX (data) = r->eax;
  DPMI_EBX (data) = r->ebx;
  DPMI_ECX (data) = r->ecx;
  DPMI_EDX (data) = r->edx;
  DPMI_ESI (data) = r->esi;
  DPMI_EDI (data) = r->edi;
  DPMI_EBP (data) = r->ebp;
  DPMI_SP (data) = r->esp;
  DPMI_CFL (data) = r->eflags;
  DPMI_IP (data) = r->eip;
  DPMI_ES (data) = r->es;
  DPMI_CS (data) = r->cs;
  DPMI_SS (data) = r->ss;
  DPMI_DS (data) = r->ds;
  DPMI_FS (data) = r->fs;
  DPMI_GS (data) = r->gs;

  ASM_invoke_callback (esi, ds,
                       &edi, &es,
                       sinfo.locked_stack_top,
                       sinfo.locked_stack_sel,
                       cb->proc_off, cb->proc_sel);

  data = (ub *) SELLIN (es, edi);

  r->eax = DPMI_EAX (data);
  r->ebx = DPMI_EBX (data);
  r->ecx = DPMI_ECX (data);
  r->edx = DPMI_EDX (data);
  r->esi = DPMI_ESI (data);
  r->edi = DPMI_EDI (data);
  r->ebp = DPMI_EBP (data);
  r->esp = DPMI_SP (data);
  r->eflags = DPMI_CFL (data);
  r->eip = DPMI_IP (data);
  r->es = DPMI_ES (data);
  r->cs = DPMI_CS (data);
  r->ss = DPMI_SS (data);
  r->ds = DPMI_DS (data);
  r->fs = DPMI_FS (data);
  r->gs = DPMI_GS (data);
  return 0;
}

static int vm86_maybe_handle_int (struct vm86_regs *r)
{
  us *sp;
  us lsp;
  ui addr;
  ui nint;

  addr = MEM_REAL (SEGLIN (r->cs, GETLOWORD (r->eip)));
  nint = addr - RED_FLAT (RED_INTERRUPT_00);

  if ((nint < 0x100) || (RED_XMS_FUNC == nint)) {
    intregs ir;

    EAX (ir) = r->eax;
    EBX (ir) = r->ebx;
    ECX (ir) = r->ecx;
    EDX (ir) = r->edx;
    ESI (ir) = r->esi;
    EDI (ir) = r->edi;
    EBP (ir) = r->ebp;
    ESP (ir) = r->esp;
    CS (ir) = r->cs;
    DS (ir) = r->ds;
    ES (ir) = r->es;
    /* SS (ir) = r->ss; */
    FS (ir) = r->fs;
    GS (ir) = r->gs;
    EFL (ir) = r->eflags;

    if (RED_XMS_FUNC == nint)
      XMS_func (ir);
    else
      INT_simulate_real (nint, ir);

    r->eax = EAX (ir);
    r->ebx = EBX (ir);
    r->ecx = ECX (ir);
    r->edx = EDX (ir);
    r->esi = ESI (ir);
    r->edi = EDI (ir);
    r->ebp = EBP (ir);
    r->esp = ESP (ir);
    r->cs = CS (ir);
    r->ds = DS (ir);
    r->es = ES (ir);
    /* r->ss = SS (ir); */
    r->fs = FS (ir);
    r->gs = GS (ir);
    r->eflags = EFL (ir);

    lsp = GETLOWORD (r->esp);

    sp = (us *) MEM_REAL (SEGLIN (r->ss, lsp));
    r->eip = sp[0];
    r->cs = sp[1];
    /* r->eflags = sp[2]; */
    r->esp = (r->esp & ~0xffff) | (lsp + 6);

    return nint;
  }
  else {
    switch (nint) {
      case RED_SWITCH_TO_PM:
      case RED_RAW_SWITCH_TO_PM:
        return 0x100;
    }
    return maybe_dpmi_callback (nint, r);
  }
}

void SIG_vm86 (struct vm86_regs *regs, int no_stack)
{
  int is_gpf, save_in_vm86, save_in_pic_loop;
  struct vm86plus_struct *save_vm86p;
  struct vm86plus_struct ctx;
  struct vm86_regs *r;

  linfo ("enter\n");

  /* was_blocked = SIG_block (); */
  save_in_pic_loop = in.pic_loop;
  in.pic_loop = 1;

#if MEM_BASE != 0
  if (0 == in.vm86) {
    void *map;

    map = mmap (NULL, MEM_ram_size (),
                PROT_READ | PROT_WRITE,
                MAP_SHARED | MAP_FIXED,
                MEM_ram_fd (), 0);

    if (MAP_FAILED == map) {
      ERRFail ("zero mapping failed");
    }
  }
#endif

  save_vm86p = vm86p;
  save_in_vm86 = in.vm86;
  memset (&ctx, 0, sizeof (ctx));
  memcpy (&ctx.regs, regs, sizeof (struct vm86_regs));
  vm86p = &ctx;
  r = &ctx.regs;

  /* SIG_unblock (was_blocked); */

  if (no_stack) {
    us ss, sp;
    us *xsp;

    if (save_in_vm86) {
      ss = save_vm86p->regs.ss;
      sp = save_vm86p->regs.esp;
      linfo ("async %04x:%04x\n", ss, sp);
    }
    else {
      DPMI_get_rm_stack (&ss, &sp);
    }
    sp -= 6;
    r->ss = ss;
    r->esp = sp;

    xsp = (us *) MEM_REAL (SEGLIN (ss, GETLOWORD (sp)));
    xsp[0] = RED_SWITCH_TO_PM;
    xsp[1] = RED_SEGMENT;
    xsp[2] = r->eflags;
  }

  if ((0 == r->ss) || (0xffff0000 & r->esp)) {
    SIG_break (1);
  }

  in.vm86 = 1;
  is_gpf = 0;
  in.pic_loop = save_in_pic_loop;

  while (1) {
    int res;

    if ((0 == r->ss) || (0xffff0000 & r->esp)) {
      SIG_break (1);
    }

    res = vm86_maybe_handle_int (r);

    if (res >= 0) {
      is_gpf = 0;
      if (res == 0x100)
        break;
      else
        continue;
    }

    if (is_gpf) {
      PIE_emu16 (r);

      if (is_fault_freq_high ()) {
        runinterp_vm86 (finfo.pie_threshold, 0 , &ctx.regs);
      }
      is_gpf = 0;
      continue;
    }

    /* log ("%#06x %#010lx\n", r->cs, r->eip); */
    res = ASM_vm86 (VM86_ENTER, &ctx);
    /* log ("res = %#x\n", res); */

    switch (VM86_TYPE (res)) {
      case VM86_SIGNAL:
        break;

      case VM86_STI:
        lwarn ("sti\n");
        break;

      case VM86_INTx:
        Aldri ();

      case VM86_TRAP:
      {
        us seg, off;
        us *sp;

        if ((r->eflags & X86_EFLAGS_TF) && (debugger_present))
          break;

        r->esp = (r->esp & ~0xffff) + (GETLOWORD (r->esp) - 6);
        sp = (us *) MEM_REAL (SEGLIN (r->ss, GETLOWORD (r->esp)));
        sp[0] = r->eip;
        sp[1] = r->cs;
        sp[2] = r->eflags;
        r->eflags &= ~(X86_EFLAGS_IF | X86_EFLAGS_TF);
        INT_get_rm (VM86_ARG (res), &seg, &off);
        r->cs = seg;
        r->eip = off;
        /* log ("trap at %#x:%#x\n", r->cs, (int) r->eip); */
        break;
      }

      case VM86_UNKNOWN:
        is_gpf = 1;
        ldebug ("gpf at %#06x %#010lx\n", r->cs, r->eip);
        break;

      default:
        log ("res = %#x, type=%#x, arg=%#x\n",
             res, VM86_TYPE (res), VM86_ARG (res));
        SIG_break (1);
    }
  }

  /* was_blocked = SIG_block (); */

  save_in_pic_loop = in.pic_loop;
  in.pic_loop = 1;

  memcpy (regs, &ctx.regs, sizeof (struct vm86_regs));
  vm86p = save_vm86p;
  in.vm86 = save_in_vm86;

#if MEM_BASE != 0
  if (0 == in.vm86) {
    int res;

    res = munmap (NULL, MEM_ram_size ());
    if (-1 == res) {
      ERRFail ("zero munmapping failed");
    }
  }
#endif

  in.pic_loop = save_in_pic_loop;
  /* SIG_unblock (was_blocked); */

  linfo ("leave\n");
}

static void set_stack (int size)
{
  void *ptr;
  int res;
  struct sigaltstack ss;

#ifdef SIGNALS_PEDAL_TO_METAL
  if (size <= 0) {
    Fail ("bad sig.stack.size %d, should be positive integer",
	  sig_stack_size);
  }
#else
  if (size <= 0)
    return;
#endif

  ptr = malloc (size);
  if (NULL == ptr) {
    Fail ("failed to allocate %d bytes for signal stack", size);
  }
  ss.ss_sp = ptr;
  ss.ss_size = size;
  ss.ss_flags = 0;

  res = sigaltstack (&ss, 0);
  if (-1 == res) {
    ERRFail ("can not set signal stack");
  }
}

static void raw_real_mode (struct vm86_regs *r)
{
  regs86 *pr;

  SIG_vm86 (r, 0);

  in.pic_loop = 1;

  pr = &ctx.regs;
  SIGCS (pr) = GETLOWORD (r->esi);
  SIGDS (pr) = GETLOWORD (r->eax);
  SIGES (pr) = GETLOWORD (r->ecx);
  SIGSS (pr) = GETLOWORD (r->edx);
  SIGFS (pr) = 0;
  SIGGS (pr) = 0;

  SIGESP (pr) = r->ebx;
  SIGEIP (pr) = r->edi;
  SIGEBP (pr) = r->ebp;
  SIGEFL (pr) = r->eflags & ~(X86_EFLAGS_VM | X86_EFLAGS_TF | X86_EFLAGS_DF);
  raise (SIGUSR1);
}

static int maybe_handle_raw_switch (regs86 *r)
{
  ui *sp;
  struct vm86_regs *rr;

  if (! ((SEL_get_user_code () == SIGCS (r))
         && (RED_FLAT (RED_RAW_SWITCH_TO_RM) == SIGEIP (r))))
    return -1;

  linfo ("switching %d %#06x %#010x\n",
         in.pic_loop, SIGSS (r), (int) SIGESP (r));

  SIGSS (r) = ASM_ss;
  SIGESP (r) = sinfo.last_esp;
  SIGESP (r) -= sizeof (struct vm86_regs);
  ASM_stack_top = sinfo.last_esp - 16384;

  sp = (ui *) SIGESP (r);
  rr = (struct vm86_regs *) sp;

  memset (rr, 0, sizeof (*rr));
  rr->ds = SIGAX (r);
  rr->es = SIGCX (r);
  rr->ss = SIGDX (r);
  rr->cs = SIGSI (r);
  rr->fs = 0;
  rr->gs = 0;
  rr->esp = SIGBX (r);
  rr->eip = SIGDI (r);
  rr->ebp = SIGEBP (r);
  rr->eflags = (SIGEFL (r) | X86_EFLAGS_VM) & ~(X86_EFLAGS_TF);

  *--sp = SIGESP (r);
  *--sp = 0xdeadbeef;

  SIGESP (r) -= 8;

  SIGES (r) = ASM_es;
  SIGCS (r) = ASM_cs;
  SIGDS (r) = ASM_ds;
  SIGFS (r) = ASM_fs;
  SIGGS (r) = ASM_gs;

  SIGEFL (r) &= ~(X86_EFLAGS_TF | X86_EFLAGS_DF);
  SIGEIP (r) = (ui) &raw_real_mode;
  return 0;
}

static void runinterp (int threshold, regs86 *r, sigset_t *oldset)
{
  int confirm;
  sigset_t set;
  x86exit_e exit;
  int alarm_blocked;

  in.x86 = 1;
  confirm = need_confirmation;
  need_confirmation = 0;

  alarm_blocked = sigismember (oldset, SIGALRM);
  if (0 == alarm_blocked) {
    sigemptyset (&set);
    sigaddset (&set, SIGALRM);
    sigprocmask (SIG_UNBLOCK, &set, NULL);
  }

  if (confirm) {
    exit = X86_confirm32 (r, threshold);
  }
  else {
    exit = X86_run32 (r, threshold);
  }

  if (0 == alarm_blocked) {
    sigprocmask (SIG_BLOCK, &set, NULL);
  }

  in.x86 = 0;

  switch (exit) {
    case X86_EXIT_TEMPORARY:
      need_confirmation = 1;
      break;

    case X86_EXIT_REQUESTED:
      if ((0 == alarm_blocked) && (0 == need_confirmation)) {
        maybe_service_irqs (r);
      }
      break;

    case X86_EXIT_BAD_INSN:
    case X86_EXIT_THRESHOLD:
      break;
  }
}

static void signal_failed (struct unicontext *ctx)
{
  log ("Failed to handle signal %d `%s'\n"
       "trap: %#x, error: %#x, cr2: %#x\n",
       ctx->signum, strsignal (ctx->signum),
       ctx->trapno, ctx->err, ctx->cr2);
  SIG_log_regs (ctx->regs);
  SIG_break (1);
}

static void unihandler (struct unicontext *ctx)
{
  regs86 *r;

  r = ctx->regs;
  switch (ctx->signum) {
    case SIGBUS:
    case SIGSEGV:
    case SIGTRAP:
    {
      switch (ctx->trapno) {
        case 0x01:
          if (need_confirmation) {
            runinterp (finfo.pie_threshold, ctx->regs, &ctx->set);
          }
          else {
            if (-1 == PIE_trap (r)) {
              ui base;

              if (IS_GDT_SELECTOR (SIGCS (r)))
                base = 0;
              else
                base = SEL_get_selector_base (SIGCS (r));

              signal (SIGTRAP, SIG_DFL);
              log ("Unexpected trace at %#06x %010x %#x\n",
                   SIGCS (r), (int) SIGEIP (r), base + (int) SIGEIP (r));
              goto failed;
            }
            maybe_service_irqs (r);
          }
          break;

        case 0x03:
          if (ASM_cs == SIGCS (r))
            SIG_break (1);
          break;

        case 0x0c:
        {
          ui *sp;
          us _sp;

          if (-1 == PIE_stack_hack (r)) {
            log ("Stack fault\n");
            goto failed;
          }

          _sp = SIGESP (r);
          SIGESP (r) = _sp - 2 * 4;
          sp = (ui *) SELLIN (SIGSS (r), SIGESP (r));
          sp[0] = SIGEIP (r);
          sp[1] = SIGCS (r);
          SIGCS (r) = ASM_cs;
          SIGEIP (r) = (reg86) ASM_reload_stack;
          break;
        }

        case 0x0d:
          if (-1 == maybe_handle_raw_switch (r)) {
            PIE_emu (r);
            maybe_service_irqs (r);
            if (is_fault_freq_high ())
              runinterp (finfo.pie_threshold, r, &ctx->set);
          }
          break;

        case 0x0e:
        {
          ui tmp;

          tmp = ctx->cr2 - MEM_REAL (0xa0000);
          if (tmp < 0x20000) {
            if (INSIDE_VM86_SYSCALL (r)) {
              int alarm_blocked;

              alarm_blocked = sigismember (&ctx->set, SIGALRM);
              runinterp_vm86 (finfo.vga_threshold,
                              0 == alarm_blocked,
                              &vm86p->regs);
            }
            else {
              runinterp (finfo.vga_threshold, ctx->regs, &ctx->set);
            }
          }
          else {
            log ("Page fault\n");
            goto failed;
          }
          break;
        }

        default:
          goto failed;
      }
      break;
    }

    case SIGUSR1:
      in.pic_loop = 0;
      INT_enable ();
      jump (r);
      break;

    case SIGALRM:
      keeping ();
      if (in.x86) {
        X86_stop ();
      }
      else {
        maybe_service_irqs (r);
      }
      break;

    default:
  failed:
      signal_failed (ctx);
      break;
  }
}

#ifdef SIGNALS_PEDAL_TO_METAL

#include <unistd.h>
#include <sys/syscall.h>

typedef void (*sigaction_proto) (int, struct sigcontext_struct);

typedef unsigned long old_sigset_t;
struct y_sigaction {
  __sighandler_t my_sa_handler;
  old_sigset_t sa_mask;
  unsigned long sa_flags;
  void (*sa_restorer)(void);
};

static void copy_context (void *dst, void *src)
{
  struct sigcontext_struct *src_ctx;
  struct _fpstate *fpstate;

  src_ctx = (struct sigcontext_struct *) src;
  memcpy (dst, src, REGS86_SZ);

  fpstate = (struct _fpstate *) ((ui) dst + REGS86_SZ);
  Fail_if (NULL == fpstate);
  ((struct sigcontext_struct *) dst)->fpstate = fpstate;
  if (src_ctx->fpstate)
    memcpy (fpstate, src_ctx->fpstate, sizeof (*fpstate));

  leverbose ("copying context pc %#06x %#010lx sp %#06x %#010lx\n",
             SIGCS (((struct sigcontext_struct *) src)),
             SIGEIP (((struct sigcontext_struct *) src)),
             SIGSS (((struct sigcontext_struct *) src)),
             SIGESP (((struct sigcontext_struct *) src)));
}

static void sig_fatal (int signum, struct sigcontext_struct context)
{
  struct unicontext ctx;
  ASM_restore_fs_gs ();

  signal (SIGABRT, SIG_DFL);
  sigemptyset (&ctx.set);
  ctx.set.__val[0] = context.oldmask;
  ctx.signum = signum;
  ctx.err = context.err;
  ctx.trapno = context.trapno;
  ctx.cr2 = context.cr2;
  ctx.regs = &context;

  signal_failed (&ctx);
}

static void sig_handler (int signum, struct sigcontext_struct context)
{
  struct unicontext ctx;
  ASM_restore_fs_gs ();

  sigemptyset (&ctx.set);
  ctx.set.__val[0] = context.oldmask;
  ctx.signum = signum;
  ctx.err = context.err;
  ctx.trapno = context.trapno;
  ctx.cr2 = context.cr2;
  ctx.regs = &context;

  if (SIGFPE == signum) {
    if (NULL != context.fpstate) {
      context.fpstate = NULL;
    }
    else {
      log ("Division by zero\n");
      signal_failed (&ctx);
    }
  }
  else unihandler (&ctx);
}

static void set_handler (int signo, sigaction_proto handler)
{
  sigset_t set;
  struct y_sigaction sa;

  sigemptyset (&set);
  sigaddset (&set, SIGALRM);

  sa.sa_mask = set.__val[0];
  sa.my_sa_handler = (__sighandler_t) handler;
  sa.sa_flags = SA_RESTART | SA_ONSTACK;
  sa.sa_restorer = 0;
  syscall (SYS_sigaction, signo, &sa, NULL);
}

#else  /* SIGNALS_PEDAL_TO_METAL */

typedef void (*sigaction_proto) (int, siginfo_t *, void *);

static void copy_context (void *dst, void *src)
{
  memcpy (dst, src, CONTEXT_SZ);
}

static void sig_fatal (int signum, siginfo_t *si, void *a3)
{
  ucontext_t *uc;
  struct unicontext ctx;

  ASM_restore_fs_gs ();

  DEREF (si);
  uc = (ucontext_t *) a3;
  sigemptyset (&ctx.set);
  ctx.set.__val[0] = uc->uc_mcontext.oldmask;
  ctx.signum = signum;
  ctx.err = uc->uc_mcontext.gregs[REG_ERR];
  ctx.trapno = uc->uc_mcontext.gregs[REG_TRAPNO];
  ctx.cr2 = uc->uc_mcontext.cr2;
  ctx.regs = &uc->uc_mcontext.gregs;

  signal (SIGABRT, SIG_DFL);
  signal_failed (&ctx);
}

static void sig_handler (int signum, siginfo_t *si, void *a3)
{
  ucontext_t *uc;
  struct unicontext ctx;

  ASM_restore_fs_gs ();

  DEREF (si);
  uc = (ucontext_t *) a3;
  sigemptyset (&ctx.set);
  ctx.set.__val[0] = uc->uc_mcontext.oldmask;
  ctx.signum = signum;
  ctx.err = uc->uc_mcontext.gregs[REG_ERR];
  ctx.trapno = uc->uc_mcontext.gregs[REG_TRAPNO];
  ctx.cr2 = uc->uc_mcontext.cr2;
  ctx.regs = &uc->uc_mcontext.gregs;

  if (SIGFPE == signum) {
    log ("Division by zero\n");
    signal_failed (&ctx);
  }
  else unihandler (&ctx);
}

static void set_handler (int signum, sigaction_proto handler)
{
  struct sigaction sa;

  sigemptyset (&sa.sa_mask);
  sigaddset (&sa.sa_mask, SIGALRM);

  sa.sa_sigaction = handler;
  sa.sa_flags = SA_SIGINFO | SA_RESTART | (sig_stack_size > 0 ? SA_ONSTACK : 0);

  if (-1 == sigaction (signum, &sa, 0)) {
    ERRFail ("sigaction for signal %d failed", signum);
  }
}
#endif

static void switch_to_pmode (struct vm86_regs *r, regs86 *s)
{
  us *sp;
  us cs, ip;
  us cs_sel;
  us ss_sel;
  us ds_sel;
  us psp_sel;
  us env_sel;
  us locked_stack_sel;
  us rm_stack_alias_sel;

  ui cs_base;
  ui ss_base;
  ui ds_base;
  ui psp_base;
  ui env_base;
  ui env_size;
  ui priv_base;
  int has_ss;

  sp = (us *) MEM_REAL (SEGLIN (r->ss, GETLOWORD (r->esp)));
  cs = sp[1];
  ip = sp[0];

  linfo ("stack %p (%04x %04lx) at %04x:%04x\n",
         (void *) sp, r->ss, r->esp, cs, ip);

  cs_base = SEGLIN (cs, 0);
  ss_base = SEGLIN (r->ss, 0);
  ds_base = SEGLIN (r->ds, 0);

  cs_base = MEM_REAL (cs_base);
  ss_base = MEM_REAL (ss_base);
  ds_base = MEM_REAL (ds_base);

  psp_base = DOS_get_PSP ();
  env_base = MEM_REAL (SEGLIN ((*(us *) (psp_base + 0x2c)), 0));
  env_size = (*(us *) (env_base - 0x10 + 3)) << 4;

  priv_base = SEGLIN (r->es, 0);

  has_ss = r->ss != r->ds;

  cs_sel = LDT_alloc_selectors (6 + has_ss);
  ss_sel = cs_sel + 8;
  ds_sel = ss_sel + (has_ss ? 8 : 0);
  psp_sel = ds_sel + 8;
  env_sel = psp_sel + 8;
  locked_stack_sel = env_sel + 8;
  rm_stack_alias_sel = locked_stack_sel + 8;

  sinfo.locked_stack_top = DPMI_PM_ASYNC_STACK_PARAS << 4;

  if (sinfo.flat_locked_stack) {
    sinfo.locked_stack_base = MEM_REAL (0);
    sinfo.locked_stack_top += MEM_REAL (priv_base);

    LDT_set_selector_base (locked_stack_sel, sinfo.locked_stack_base);
    LDT_set_selector_limit (locked_stack_sel, ~0U);
    LDT_set_selector_access_rights (locked_stack_sel, 0xc0f2);
  }
  else {
    sinfo.locked_stack_base = MEM_REAL (priv_base);

    LDT_set_selector_base (locked_stack_sel, sinfo.locked_stack_base);
    LDT_set_selector_limit (locked_stack_sel, sinfo.locked_stack_top - 1);
    LDT_set_selector_access_rights (locked_stack_sel, 0x40f2);
  }

  LDT_set_selector_limit (rm_stack_alias_sel, 0xffff);
  LDT_set_selector_access_rights (rm_stack_alias_sel, 0x40f2);

  LDT_set_selector_base (cs_sel, cs_base);
  LDT_set_selector_limit (cs_sel, 0xffff);
  LDT_set_selector_access_rights (cs_sel, 0xfa);

  LDT_set_selector_base (ds_sel, ds_base);
  LDT_set_selector_limit (ds_sel, 0xffff);
  LDT_set_selector_access_rights (ds_sel, 0x40f2);

  if (has_ss) {
    LDT_set_selector_base (ss_sel, ss_base);
    LDT_set_selector_limit (ss_sel, 0xffff);
    LDT_set_selector_access_rights (ss_sel, 0x40f2);
  }

  LDT_set_selector_base (psp_sel, psp_base);
  LDT_set_selector_limit (psp_sel, 0xff);
  LDT_set_selector_access_rights (psp_sel, 0x40f2);

  LDT_set_selector_base (env_sel, env_base);
  LDT_set_selector_limit (env_sel, env_size - 1);
  LDT_set_selector_access_rights (env_sel, 0x00f2);

  LDT_make_selector_permanent (psp_sel);
  LDT_make_selector_permanent (env_sel);
  LDT_make_selector_permanent (locked_stack_sel);

  sinfo.locked_stack_sel = locked_stack_sel;
  sinfo.rm_stack_alias_sel = rm_stack_alias_sel;

  *(us *) (psp_base + 0x2c) = env_sel;

  SIGEAX (s) = r->eax;
  SIGEBX (s) = r->ebx;
  SIGECX (s) = r->ecx;
  SIGEDX (s) = r->edx;
  SIGESI (s) = r->esi;
  SIGEDI (s) = r->edi;
  SIGEBP (s) = r->ebp;
  SIGESP (s) = GETLOWORD (r->esp) + 4;
  SIGEIP (s) = ip;
  SIGCS (s) = cs_sel;
  SIGDS (s) = ds_sel;
  SIGES (s) = psp_sel;
  SIGSS (s) = ss_sel;
  SIGFS (s) = 0;
  SIGGS (s) = 0;
  in.dpmi = 1;
}

void SIG_start (regs86 *r)
{
  struct vm86_regs regs;
  struct itimerval iv;

  linfo ("starting\n");

  regs.eax = SIGEAX (r);
  regs.ebx = SIGEBX (r);
  regs.ecx = SIGECX (r);
  regs.edx = SIGEDX (r);
  regs.esi = SIGESI (r);
  regs.edi = SIGEDI (r);
  regs.ebp = SIGEBP (r);
  regs.esp = SIGESP (r);
  regs.eip = SIGEIP (r);
  regs.cs = SIGCS (r);
  regs.ds = SIGDS (r);
  regs.es = SIGES (r);
  regs.ss = SIGSS (r);
  regs.fs = SIGFS (r);
  regs.gs = SIGGS (r);
  regs.eflags |= X86_EFLAGS_IF;
  regs.eflags &= ~X86_EFLAGS_TF;

  iv.it_interval.tv_sec = 0;
  iv.it_interval.tv_usec = 1000000 / 100;
  iv.it_value.tv_sec = 0;
  iv.it_value.tv_usec = iv.it_interval.tv_usec;

  setitimer (ITIMER_REAL, &iv, 0);
  SIG_vm86 (&regs, 0);

  in.pic_loop = 1;
  switch_to_pmode (&regs, &ctx.regs);

  sinfo.last_esp = ASM_stack_top;
  raise (SIGUSR1);
}

char *SIG_strdup (const char *s, const char *file,
                  const char *func, int line)
{
  char *ptr;

  r_lock ();
  ptr = strdup (s);
  r_unlock ();

  if (NULL == ptr) {
    ERRFail ("(s=%p, file=`%s', func=`%s', line=%d)",
             s, file, func, line);
  }

  ldebug ("(s=%p, file=`%s', func=`%s', line=%d) = %p\n",
          s, file, func, line, ptr);
  return ptr;
}

void *SIG_malloc (unsigned int size, char *file, const char *func, int line)
{
  void *ptr;

  r_lock ();
  ptr = malloc (size);
  r_unlock ();

  if (NULL == ptr) {
    ERRFail ("(size=%u, file=`%s', func=`%s', line=%d)",
             size, file, func, line);
  }

  ldebug ("(size=%u, file=`%s', func=`%s', line=%d) = %p\n",
          size, file, func, line, ptr);

  return ptr;
}

void *SIG_realloc (void *ptr, unsigned int size,
                   char *file, const char *func, int line)
{
  DEREF (file);
  DEREF (func);
  DEREF (line);

  r_lock ();
  ptr = realloc (ptr, size);
  r_unlock ();

  if (NULL == ptr) {
    ERRFail ("(ptr=%p, size=%u, file=`%s', func=`%s', line=%d)",
             ptr, size, file, func, line);
  }

  ldebug ("(ptr=%p, size=%u, file=`%s', func=`%s', line=%d) = %p\n",
          ptr, size, file, func, line, ptr);

  return ptr;
}

void SIG_free (void *ptr, char *file, const char *func, int line)
{
  r_lock ();
  free (ptr);
  r_unlock ();

  DEREF (file);
  DEREF (func);
  DEREF (line);

  ldebug ("(ptr=%p, file=`%s', func=`%s', line=%d) = %p\n",
          ptr, file, func, line, ptr);
}

void SIG_pre_init ()
{
  static struct MUP_key keys[] = {
    {"stack.size"        , TAG_INT    , {.ip = &sig_stack_size}},
    {"use.locked.stack"  , TAG_INT    , {.ip = &sinfo.use_locked_stack}},
    {"flat.locked.stack" , TAG_INT    , {.ip = &sinfo.flat_locked_stack}},
    {"sampling.period"   , TAG_IN_INT , {.ii = 100}},
    {"interp.samples"    , TAG_INT    , {.ip = &finfo.nsamples}},
    {"pie.threshold"     , TAG_INT    , {.ip = &finfo.pie_threshold}},
    {"vga.threshold"     , TAG_INT    , {.ip = &finfo.vga_threshold}},
    {"debugger.command"  , TAG_STRING , {.sp = &debugger_command}},
  };

  static struct MUP_group group =
    {"sig.", LENOFA (keys), keys, NULL};

  MUP_add_group (&group);
}

void SIG_init ()
{
  int period;

  log_level = MUP_get_log_level ("sig");

  set_stack (sig_stack_size);
  debugger_present = is_debugger_present ();

  if (debugger_present)
    finfo.nsamples /= 2;

  period = MUP_get_int ("sig.sampling.period");
  if (period && (finfo.nsamples > 0)) {
    finfo.sampling_period = TID_NSEC / period;
  }
  else {
    finfo.sampling_period = ULLONG_MAX;
  }

  set_handler (SIGBUS, sig_handler);
  set_handler (SIGSEGV, sig_handler);
  set_handler (SIGALRM, sig_handler);
  set_handler (SIGUSR1, sig_handler);
  set_handler (SIGTRAP, sig_handler);
  set_handler (SIGFPE, sig_handler);

  set_handler (SIGILL, sig_fatal);
  set_handler (SIGABRT, sig_fatal);
}
