#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>

#define MUL 4
#include <inc/sig.h>
#include <inc/red.h>
#include <inc/int.h>
#include <inc/bit.h>
#include <inc/vga.h>
#include <inc/mem.h>
#include <inc/eio.h>
#include <inc/tid.h>
#include <inc/vid.h>
#include <inc/x86.h>
#include <inc/asm.h>
#include <inc/mup.h>
#include <inc/log.h>

#include <vgaemu/misc.h>
#include <vgaemu/vgaemu.h>
#include <vgabios/vgabios.h>

#define LOG_PREFIX "VGA"
static int log_level;

#define ADDR_A000 ((void *) MEM_A0000)
#define ADDR_B000 ((void *) MEM_B0000)
#define ADDR_B800 ((void *) MEM_B8000)
#define ADDR_C000 ((void *) MEM_C0000)

static DisplayState ds;
static void *vga_ram_ptr;

static int using_bios;
static int full_vga_in_vbe;
static ul retrace_period;
static ul retrace_frequency;
static int vertical_resolution;
static ub pal[0x300];
static int pal_index, pal_dirty;
static enum {Plain, Tiny, Full} emu_type = Tiny;
static char *bios_path;

static struct {
  int is_mode13;
  int is_chained;
  int ar_flipflop;
  ui msr;
  ui ri;
  ui last_layer;
  ui last_w, last_h, last_cr9;
  ub cr_index;
  ub cr[256];
  ub sr_index;
  ub sr[8];
  ub gr_index;
  ub gr[16];
} vga;

void VGA_set_vertical_resolution (int res)
{
  vertical_resolution = res;
}

static void set_mode13 ()
{
  int w = 320;
  int h = 200;

  vga.is_chained = 1;
  vga.is_mode13 = 1;
  vga.last_w = w;
  vga.last_h = h;
  vga.cr[0x1] = (w / 4) - 1;
  vga.cr[0x7] = 2;
  vga.cr[0x12] = 143;
  vga.cr[0x9] = 0x81;
  vga.sr[0x4] = 10;
  vga.cr[0xd] = 0;
  vga.cr[0xc] = 0;
}

static void full_vid_create (int nmode)
{
  struct VID_mode_info info;

  info.w = 640;
  info.h = 400;
  info.bpp = 16;
  info.typ = VID_TYPE_VGA;
  info.pitch = info.w * 2;
  info.rmask = 0x0000f800;
  info.gmask = 0x000007e0;
  info.bmask = 0x0000001f;
  info.amask = 0;
  ds.linesize = 1280;

  MEM_unmap_mappings_to_video ();
  MEM_protect_memory (MEM_A0000, 0x20000, MEM_MAP_NONE);

  if (-1 == VID_create (&info))
    Fail ("failed to set VGA mode %#x", nmode);
}

static void full_set_mode (int nmode)
{
  linfo ("(nmode=%#x)\n", nmode);

  if (0 == nmode)
    return;

  full_vid_create (nmode);

  linfo ("enter biosfn mode set\n");
  biosfn_set_video_mode (nmode);
  linfo ("leave biosfn mode set\n");

  memset (vga_ram_ptr, 0, VGA_RAM_SIZE);
  if (3 == nmode) {
    int i, j;
    ub *b;
    const ub *font;

    font = VID_vga_def8x16_font;
    b = ((ub *) vga_ram_ptr) + 2;
    for (i = 0; i < 256; i++) {
      for (j = 0; j < 16; j++) {
        b[j << 2] = *font++;
      }
      b += 32 << 2;
    }

    for (i = 0; i < 80 * 24; i++) {
      ((ub *) MEM_B8000)[i * 2 + 1] = 7;
    }
  }
}

static void tiny_set_mode (int nmode)
{
  struct VID_mode_info info;
  int i;
  int w, h;
  int bpp;

  linfo ("(nmode=%#x)\n", nmode);

  MEM_unmap_mappings_to_video ();

  switch (nmode) {
    case 2:
    case 3:
    {
      char *p;

      p = ADDR_B800;
      bpp = 8;

      MEM_VID_MODE = 3;
      MEM_VID_CURSOR_POSITIONS[0] = 0;
      MEM_VID_CURSOR_POSITIONS[1] = 0;
      MEM_VID_COLUMNS = 80;
      MEM_VID_ROWS = 24;

      for (i = 0; i < (MEM_VID_ROWS + 1) * MEM_VID_COLUMNS; i++) {
        p[i * 2] = 0;
        p[i * 2 + 1] = 7;
      }

      w = 8 * MEM_VID_COLUMNS;
      if (vertical_resolution) {
        h = vertical_resolution;
        vertical_resolution = 0;
      }
      else {
        h = 16 * (MEM_VID_ROWS + 1);
      }

      for (i = 0; i < 64 * 3; i++) {
        pal[i] = VID_vga_def64_palette[i] << 2;
      }
      break;
    }

    case 0x0d:
    case 0x12:
      log ("mode %#x is not really supported\n", nmode);

    case 0x13:
      w = 320;
      h = 200;
      bpp = 8;

      MEM_VID_MODE = 0x13;
      MEM_VID_CURSOR_POSITIONS[0] = 0;
      MEM_VID_CURSOR_POSITIONS[1] = 0;
      MEM_VID_COLUMNS = w;
      MEM_VID_ROWS = h - 1;


      if (0x13 == nmode) {
        set_mode13 ();
      }

      memcpy (pal, VID_vga_def256_palette, 256 * 3);
      MEM_map_address_to_video (MEM_A0000, 0, 0x10000);
      break;


    default:
      Fail ("unsupported mode %#x", nmode);
      break;
  }

  info.typ = VID_TYPE_VGA;
  info.w = w;
  info.h = h;
  info.pitch = w;
  info.bpp = bpp;
  info.rmask = info.gmask = info.bmask = info.amask = 0;

  if (-1 == VID_create (&info)) {
    Fail ("failed to set VGA mode %#x", nmode);
  }

  VID_set_palette (pal, 0, 256);
  VID_set_font (VID_vga_def8x16_font, 0, 256, 8, 16);

  pal_index = 0;
  pal_dirty = 0;
}

void VGA_set_mode (int nmode)
{
  if (Full == emu_type)
    full_set_mode (nmode);
  else
    tiny_set_mode (nmode);
}

void VGA_set_cursor_position (ub page, ub row, ub col)
{
  DEREF (page);

  MEM_VID_CURSOR_POSITIONS[0] = row;
  MEM_VID_CURSOR_POSITIONS[1] = col;
}

static void scroll_window (char *video, int pitch,
                           int ax, int ay, int bx, int by,
                           char filler)
{
  int i, j, k;
  char *dst, *src;

  k = (3 == MEM_VID_MODE) ? 2 : 1;

  dst = video + (ax * k) + ay * (pitch * k);
  src = dst + (pitch * k);


  for (i = 0; i < (by - ay) - 1; i++) {
    for (j = 0; j < (bx - ax) * k; j++)
      *dst++ = *src++;

    dst += (pitch * k) - j;
    src += (pitch * k) - j;
  }

  if (3 == MEM_VID_MODE) {
    for (j = 0; j < bx - ax; j++) {
      *dst++ = ' ';
      *dst++ = filler;
    }
  }
  else {
    for (j = 0; j < (bx - ax); j++)
      *dst++ = filler;
  }
}

static void print_string_TTY (const char *string, int size)
{
  char *video;
  const char *c;
  us x, y;
  us xres, yres;

  video = (char *) MEM_B8000;

  xres = MEM_VID_COLUMNS;
  yres = MEM_VID_ROWS + 1;

  c = string;

  x = MEM_VID_CURSOR_POSITIONS[0];
  y = MEM_VID_CURSOR_POSITIONS[1];

  while (size--) {

    if (y == yres) {
      scroll_window (video, xres, 0, 0, xres, yres, 7);
      y--;
    }

    switch (*c) {
      case '\b': (0 == x) ? xres - 1 : x--; break;
      case '\t': x += 8; break;
      case '\n': y++; break;
      case '\r': x = 0; break;
      default:
        video[x * 2 + y * xres * 2] = *c;
        x++;
    }
    if (x >= xres) {
      x %= xres;
      y++;
    }
    c++;
  }

  MEM_VID_CURSOR_POSITIONS[0] = x;
  MEM_VID_CURSOR_POSITIONS[1] = y;
}

void VGA_print_string_TTY (const char *string, int size)
{
  print_string_TTY (string, size);
}

static ui read_3da (int nport, int size)
{
#if 1
  static int hretr, vretr;
  static ul past;
  ul delta, now;

  DEREF (nport);
  DEREF (size);

  now = TID_get_abs_nsec ();
  delta = now - past;

  if (vretr) {
    if (delta > retrace_period) {
      vretr = 0;
    }
  }
  else {
    if (delta > retrace_frequency) {
      vretr = 9;
      past = now;
    }
    else {
      hretr = (delta % 49) > 35; /* DOSEMU magic */
    }
  }

  return 0xc4 | vretr | hretr;
#else
  int retval;
  static int ff;
  static ul moo;
  ul now, delta;


  DEREF (nport);
  DEREF (size);

  if (retrace_frequency) {
    if (0 == moo)
      moo = TID_get_abs_nsec ();

    now = TID_get_abs_nsec ();
    delta = (now - moo);

    if (ff && retrace_period)
      ff = delta < retrace_period;
    else
      ff = delta > retrace_frequency;
  }
  else {
    ff = !ff;
    now = 0;
  }

  if (ff) {
    ldebug ("retrace\n");
    moo = now;
    retval = 9;
  }
  else {
    ldebug ("no retrace\n");
    retval = 0;
  }

  return 0xc4 | retval;
#endif
}

static void write_pal (int nport, int size, ui val)
{
  DEREF (size);

  switch (nport) {
    case 0x3c7:
      pal_index = val;
      break;

    case 0x3c8:
      pal_index = val * 3;
      break;

    case 0x3c9:
      pal[pal_index++] = val;
      pal_dirty = 1;
      if (0x300 == pal_index) {
        pal_index = 0;
      }
  }
}

static ui read_3c8 (int nport, int size)
{
  DEREF (nport);
  DEREF (size);

  return pal_index;
}

static ui read_3c9 (int nport, int size)
{
  DEREF (nport);
  DEREF (size);

  return pal[pal_index++];
}

static ui read_3c2 (int nport, int size)
{
  DEREF (nport);
  DEREF (size);

  return vga.msr;
}

static void copy_to_layer (int nlayer)
{
  int i;
  /* ui start_addr, line_compare, line_offset; */
  ub *src, *dst;

  dst = MEM_video_buffer ();
  src = ADDR_A000;

  if (vga.gr[5] & 0x10) {
    for (i = 0; i < 0x10000 / 2;) {
      dst[nlayer + i * 4] = src[i];
      i += 2;
    }
  }
  else {
    for (i = 0; i < 0x10000; i++) {
      dst[nlayer + i * MUL] = src[i];
    }
  }
}

static void write_data (int nport, int size, ui data)
{
  DEREF (size);

  ldebug ("port[%#6x] = %#6x, size %d\n", nport, data, size);

  switch (nport) {
    case 0x3c5:
      if (vga.sr_index != 2) {
        linfo ("0x3c4 [%#x] = %#x\n", vga.sr_index, data);
      }
      vga.sr[vga.sr_index] = data;

      if (Tiny == emu_type)
        if (2 == vga.sr_index) {
          copy_to_layer (lsbindex (vga.last_layer));
          vga.last_layer = data;
        }
      break;

    case 0x3d5:
      if ((0xc != vga.cr_index) && (0xd != vga.cr_index)) {
        linfo ("0x3d4 [%#x] = %#x\n", vga.cr_index, data);
      }
      vga.cr[vga.cr_index] = data;
      break;

    case 0x3cf:
      if (vga.gr_index != 5) {
        linfo ("0x3ce [%#x] = %#x\n", vga.gr_index, data);
      }
      vga.gr[vga.gr_index] = data;
      break;

    case 0x3c2:
      vga.msr = data;
      break;

    default:
      Aldri ();
  }
}

static void write_index (int nport, int size, ui data)
{
  ui index;

  index = data & 0xff;

  ldebug ("nport[%#6x] = %#6x, size %d\n", nport, data, size);

  if (4 == size)
    Fail ("double word access to vga port %#x, data %#x", nport, data);

  switch (nport) {
    case 0x3c4:
      vga.sr_index = index;
      break;

    case 0x3d4:
      vga.cr_index = index;
      break;

    case 0x3ce:
      vga.gr_index = index;
      break;

    default:
      Aldri ();
  }

  if (2 == size)
    write_data (nport + 1, 1, data >> 8);
}

static void modex ()
{
  ui w, h, slpcr;

  slpcr = vga.cr[9] & 0x1f;

  w = (vga.cr[1] + 1) * MUL;
  h = vga.cr[0x12] | ((vga.cr[7] & 2) << 7) | ((vga.cr[7] & 0x40) << 3);
  h += 1;
  h /= slpcr + 1;

  if ((vga.last_w != w)
      || (vga.last_h != h)
      || (vga.last_cr9 != vga.cr[9])) {
    vga.last_w = w;
    vga.last_h = h;
    vga.last_cr9 = vga.cr[9];

    VID_resize_vga (w, h);
  }
}

void VGA_run (void)
{
  switch (emu_type) {
    case Full:
      if (full_vga_in_vbe) {
        if (pal_dirty) {
          VID_set_palette (pal, 0, 256);
          pal_dirty = 0;
        }
      }
      else
        vga_update_display ();
      break;

    case Tiny:
      if (vga.is_mode13) {
        int is_chained;

        is_chained = 0 != (vga.sr[0x4] & 0x8);

        switch ((vga.is_chained << 1) | is_chained) {
          case 1:
          {
            struct VID_mode_info info;

            set_mode13 ();

            info.typ = VID_TYPE_VGA;
            info.w = 320;
            info.h = 200;
            info.pitch = 320;
            info.bpp = 8;
            info.rmask = info.gmask = info.bmask = info.amask = 0;

            MEM_map_address_to_video (MEM_A0000, 0, 0x1000);
            if (-1 == VID_create (&info)) {
              Fail ("chain VGA");
            }
            break;
          }

          case 2:
          {
            struct VID_mode_info info;

            vga.last_layer = 0;

            info.typ = VID_TYPE_VGA;
            info.w = vga.last_w;
            info.h = vga.last_h;
            info.pitch = vga.last_w;
            info.rmask = info.gmask = info.bmask = info.amask = 0;
            info.bpp = 8;

            MEM_unmap_mappings_to_video ();
            if (-1 == VID_create (&info)) {
              Fail ("unchain VGA");
            }
          }

          case 0:
            modex ();
            break;

          case 3:
            break;

          default:
            Aldri ();
        }
        vga.is_chained = is_chained;
      }

    case Plain:
      if (pal_dirty) {
        VID_set_palette (pal, 0, 256);
        pal_dirty = 0;
      }

      if (3 == MEM_VID_MODE) {
        VID_text_mode_draw ((ub *) MEM_B8000);
      }
      break;
  }
  VID_update ();
}

static void ds_resize (DisplayState *ds, int w, int h)
{
  VID_resize_vga (w, h);
  ds->linesize = w * 2;
}

static void ds_update (DisplayState *ds, int x, int y, int w, int h)
{
  DEREF (ds);
  DEREF (x);
  DEREF (y);
  DEREF (w);
  DEREF (h);
  /* log ("x=%d y=%d w=%d h=%d\n", x, y, w, h); */
}

/* HERE BE DRAGONS */
static IOPortReadFunc *vga_ioport_read;
static IOPortWriteFunc *vga_ioport_write;

static ui vga_read_helper (int nport, int size)
{
  ui ret;

  ret = vga_ioport_read (NULL, nport);
  if (2 == size)
    ret |= (vga_ioport_read (NULL, nport + 1) << 8);

  switch (nport) {
    case 0x3da: ret = read_3da (nport, size); break;
    case 0x3c8: ret = read_3c8 (nport, size); break;
    case 0x3c9: ret = read_3c9 (nport, size); break;
    default:
      lverbose ("nport=%#x, size=%d, ret=%#x\n", nport, size, ret);
      break;
  }
  return ret;
}

static void vga_write_helper (int nport, int size, ui val)
{
  vga_ioport_write (NULL, nport, val & 0xff);
  if (2 == size)
    vga_ioport_write (NULL, nport + 1, val >> 8);

  write_pal (nport, size, val);
  lverbose ("nport=%#x, size=%d, val=%#x\n", nport, size, val);
}

int register_ioport_read(int start, int length, IOPortReadFunc *func, int size)
{
  int i, mask;

  DEREF (size);

  Fail_if (vga_ioport_read && func != vga_ioport_read);

  vga_ioport_read = func;
  for (i = start; i < start + length; i++) {
    if ((i == 0x3ce) || (i == 0x3d4))
      mask = EIO_MASK_BW;
    else
      mask = EIO_MASK_B;

    EIO_register_read_handler (i, mask, vga_read_helper);
  }
  return 0;
}

int register_ioport_write(int start, int length, IOPortWriteFunc *func, int size)
{
  int i;

  DEREF (size);

  Fail_if (vga_ioport_write && func != vga_ioport_write);

  vga_ioport_write = func;
  for (i = start; i < start + length; i++) {
    EIO_register_write_handler (i, EIO_MASK_BW, vga_write_helper);
  }
  return 0;
}

static void loc_I10 (INT_ARGS)
{
  switch (AH (r)) {
    case 0x00:
    {
      if (full_vga_in_vbe) {
        X86_catch_VGA (1);
        full_vid_create (AL (r));
      }
      linfo ("bios set mode %#x\n", AL (r));
      full_vga_in_vbe = 0;
      if (0 == using_bios)
        I10 (r);
      break;
    }

    case 0x4f:
    {
      int al;

      al = AL (r);
      I10 (r);
      if ((2 == al) && (0x4f == AX (r))) {
        full_vga_in_vbe = 1;
        X86_catch_VGA (0);
        MEM_protect_memory (MEM_A0000, 0x20000, MEM_MAP_RDWR);
      }
      break;
    }

    default:
      linfo ("%#x\n", AX (r));
      if (0 == using_bios) I10 (r);
  }
}

static void insert_int10_redirector (us old_seg, us old_off)
{
  us redir_seg, redir_off;
  int redirector_size;
  int redirector_addr;

  INT_set (0x10, loc_I10);

  ASM_bios_int10_handler[0] = old_off;
  ASM_bios_int10_handler[1] = old_seg;

  redirector_size = ASM_int10_redirector_end - ASM_int10_redirector;
  redirector_addr = MEM_REAL (0xcffff) - redirector_size;
  memcpy ((void *) redirector_addr, ASM_int10_redirector, redirector_size);

  redir_seg = redirector_addr >> 4;
  redir_off = redirector_addr - (redir_seg << 4);
  INT_set_rm (0x10, redir_seg, redir_off);
}

static int maybe_init_BIOS ()
{
  us seg, off;
  struct vm86_regs regs;
  int fd;
  int nread;

  if (NULL == bios_path) {
    INT_set (0x10, loc_I10);
    biosfn_set_video_mode (0);
    VGA_set_mode (3);
    return 0;
  }

  fd = open (bios_path, O_RDONLY);
  if (-1 == fd) {
    ERRFail ("failed to open `%s'", bios_path);
  }

  nread = read (fd, ADDR_C000, 0x10000 - 1);
  if (-1 == nread) {
    ERRFail ("failed to read VGA bios");
  }

  if (-1 == close (fd)) {
    ERRFail ("failed to close VGA bios fd=%d", fd);
  }

  full_vid_create (3);

  using_bios = 1;
  regs.eip = 3;
  regs.cs = 0xc000;
  regs.eflags = 0;

  SIG_vm86 (&regs, 1);

  INT_get_rm (0x10, &seg, &off);
  insert_int10_redirector (seg, off);

  regs.cs = seg;
  regs.eip = off;
  regs.eax = 0;
  SIG_vm86 (&regs, 1);

  regs.cs = seg;
  regs.eip = off;
  regs.eax = 3;
  SIG_vm86 (&regs, 1);

  return 1;
}

static int vga_type_setter (char *desc)
{
  if ((0 == strcmp (desc, "plain")))
    emu_type = Plain;
  else if (0 == strcmp (desc, "tiny"))
    emu_type = Tiny;
  else if (0 == strcmp (desc, "full"))
    emu_type = Full;
  else {
    log ("bogus vga emulation type `%s'\n"
         "must be one of: `plain', `tiny', `full'\n", desc);
    return -1;
  }
  return 0;
}

void VGA_pre_init ()
{
  static struct MUP_key keys[] = {
    {"emu"               , TAG_SETTER , {.setter = vga_type_setter}},
    {"retrace.freq"      , TAG_IN_INT , {.ii = 70}},
    {"retrace.period"    , TAG_IN_INT , {.ii = 0}},
    {"bios"              , TAG_STRING , {.sp = &bios_path}},
  };

  static struct MUP_group group =
    {"vga.", LENOFA (keys), keys, NULL};

  MUP_add_group (&group);
}

void VGA_init (void)
{
  int rfreq, rperiod;

  log_level = MUP_get_log_level ("vga");

  rfreq = MUP_get_int ("vga.retrace.freq");
  rperiod = MUP_get_int ("vga.retrace.period");

  if (rfreq)
    retrace_frequency = TID_NSEC / rfreq;

  if (rperiod)
    retrace_period = rperiod * (TID_NSEC / 1000000ULL);

  if (Full == emu_type) {
    X86_catch_VGA (1);
    vga_ram_ptr = MEM_map (VGA_RAM_SIZE, MEM_MAP_RDWR);

    if (NULL == vga_ram_ptr) {
      Fail ("failed to allocate %d bytes for VGA RAM", VGA_RAM_SIZE);
    }

    ds.depth = 16;
    ds.data = MEM_video_buffer ();
    ds.dpy_update = ds_update;
    ds.dpy_resize = ds_resize;
    ds.dpy_refresh = NULL;
    ds.linesize = 1280;

    vga_init (&ds, vga_ram_ptr, 0, VGA_RAM_SIZE);
    maybe_init_BIOS ();
  }
  else {
    MEM_CRT_ADDR = 0x3d4;

    EIO_register_read_handler (0x3da, EIO_MASK_B, read_3da);
    EIO_register_read_handler (0x3c2, EIO_MASK_B, read_3c2);
    EIO_register_write_handler (0x3c2, EIO_MASK_B, write_data);

    EIO_register_write_handler (0x3d4, EIO_MASK_BW, write_index);
    EIO_register_write_handler (0x3d5, EIO_MASK_B, write_data);

    EIO_register_write_handler (0x3c4, EIO_MASK_BW, write_index);
    EIO_register_write_handler (0x3c5, EIO_MASK_B, write_data);

    EIO_register_write_handler (0x3ce, EIO_MASK_BW, write_index);
    EIO_register_write_handler (0x3cf, EIO_MASK_B, write_data);

    EIO_register_write_handler (0x3c7, EIO_MASK_B, write_pal);
    EIO_register_write_handler (0x3c8, EIO_MASK_B, write_pal);
    EIO_register_write_handler (0x3c9, EIO_MASK_B, write_pal);

    EIO_register_read_handler (0x3c8, EIO_MASK_B, read_3c8);
    EIO_register_read_handler (0x3c9, EIO_MASK_B, read_3c9);

    VGA_set_mode (3);
  }
}
