#ifdef USE_HERMES
#include <Hermes/Hermes.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <SDL/SDL.h>

#include <inc/sig.h>            /* xmalloc, ... */
#include <inc/vbe.h>
#include <inc/vga.h>
#include <inc/vid.h>
#include <inc/mem.h>
#include <inc/tid.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "VID"
static int log_level;

static SDL_Surface *screen, *surface;

static struct vid {
  int w, h, pitch, bpp, rmask, gmask, bmask, amask;
  int screen_w, screen_h;
  int font_w, font_h;
  ub *font;
  int flags;
  vidtype_e typ;
  SDL_PixelFormat bestvfmt;
  SDL_Color palette[256];
  ul period;
} vid;

#ifdef USE_HERMES
int use_hermes;
HermesFormat hermes_src_fmt;
HermesFormat hermes_dst_fmt;
HermesHandle hermes_cvt_handle;
HermesHandle hermes_pal_handle;
HermesHandle hermes_clr_handle;
#endif

static ub *vid_buf;
static int new_vid_flags;
static int pal_dirty;
static ul last_update;
static int must_resize;
static int vid_lock;
static const char *last_func;

#define MUST_LOCK() do {                                \
  if (vid_lock) {                                       \
    Fail ("attempt to re-enter video from %s in %s",    \
          __func__, last_func);                         \
  }                                                     \
  vid_lock = 1;                                         \
  last_func = __func__;                                 \
} while (0)

#define IS_LOCKED() (1 == vid_lock)
#define LOCK() do { vid_lock = 1; last_func = __func__; } while (0)
#define UNLOCK() do {                           \
  Fail_if (vid_lock == 0);                      \
  vid_lock = 0;                                 \
  last_func = NULL;                             \
} while (0)

void VID_resize_vga (int w, int h)
{
  vid.w = vid.screen_w = w;
  vid.h = vid.screen_h = h;
  switch (vid.bpp) {
    case 8: vid.pitch = w; break;
    case 16: vid.pitch = w * 2; break;
    case 32: vid.pitch = w * 4; break;
  }
  must_resize = 1;
}

#ifdef USE_HERMES
void VID_resize (int w, int h)
{
  vid.screen_w = w;
  vid.screen_h = h;
  must_resize = 1;
}

static void log_hermes (const char *cap)
{
  DEREF (cap);

  if (LOG_INFO <= log_level) {
    log ("%s\n"
         "src_fmt\n"
         "indexed %d\n"
         "bits %d\n"
         "r %#010x\n"
         "g %#010x\n"
         "b %#010x\n"
         "a %#010x\n"
         "\n"
         "dst_fmt\n"
         "indexed %d\n"
         "bits %d\n"
         "r %#010x\n"
         "g %#010x\n"
         "b %#010x\n"
         "a %#010x\n",

         cap,
         hermes_src_fmt.indexed,
         hermes_src_fmt.bits,
         hermes_src_fmt.r,
         hermes_src_fmt.g,
         hermes_src_fmt.b,
         hermes_src_fmt.a,

         hermes_dst_fmt.indexed,
         hermes_dst_fmt.bits,
         hermes_dst_fmt.r,
         hermes_dst_fmt.g,
         hermes_dst_fmt.b,
         hermes_dst_fmt.a);
  }

}
#endif

static void log_video (const char *cap)
{
  struct SDL_PixelFormat *fmt;

  DEREF (cap);
  fmt = screen->format;

  if (LOG_INFO <= log_level) {
    log ("%s: screen\n"
         "BitsPerPixel %d\n"
         "BytesPerPixel %d\n"
         "pitch %d\n"
         "Rmask %#010x\n"
         "Gmask %#010x\n"
         "Bmask %#010x\n"
         "Amask %#010x\n"
         "\n"
         "vid_bpp %d\n"
         "vid_xres %d\n"
         "vid_yres %d\n"
         "vid_pitch %d\n",
         cap,

         fmt->BitsPerPixel,
         fmt->BytesPerPixel,
         screen->pitch,
         fmt->Rmask,
         fmt->Gmask,
         fmt->Bmask,
         fmt->Amask,

         vid.bpp,
         vid.w,
         vid.h,
         vid.pitch
      );
  }
#ifdef USE_HERMES
  if (use_hermes)
    log_hermes (cap);
#endif
}

int VID_create (struct VID_mode_info *info)
{
  int was_blocked;
  char *msg;

  msg = (char *) -1;
  was_blocked = SIG_block ();
  MUST_LOCK ();

  linfo ("mode %s\n"
         "w %d h %d bpp %d pitch %d\n"
         "rmask %#x gmask %#x bmask %#x amask %#x\n",
         VID_TYPE_VGA == info->typ ? "vga" : "vbe",
         info->w, info->h,
         info->bpp,
         info->pitch,
         info->rmask, info->gmask, info->bmask, info->amask);

  screen = SDL_SetVideoMode (info->w, info->h,
                             vid.bestvfmt.BitsPerPixel, vid.flags);

  if (NULL == screen) {
    msg = "SDL_SetVideoMode";
    goto SDL_error;
  }

  vid.w = vid.screen_w = info->w;
  vid.h = vid.screen_h = info->h;
  vid.bpp = info->bpp;
  vid.pitch = info->pitch;
  vid.rmask = info->rmask;
  vid.gmask = info->gmask;
  vid.bmask = info->bmask;
  vid.amask = info->amask;

#ifdef USE_HERMES
  if (use_hermes) {
    struct SDL_PixelFormat *fmt;

    hermes_src_fmt.indexed = 8 == vid.bpp;
    hermes_src_fmt.bits = (15 == vid.bpp) ? 16 : vid.bpp;
    hermes_src_fmt.r = vid.rmask;
    hermes_src_fmt.g = vid.gmask;
    hermes_src_fmt.b = vid.bmask;
    hermes_src_fmt.a = vid.amask;

    fmt = screen->format;
    hermes_dst_fmt.indexed = 8 == fmt->BitsPerPixel;
    hermes_dst_fmt.bits = fmt->BitsPerPixel;
    hermes_dst_fmt.r = fmt->Rmask;
    hermes_dst_fmt.g = fmt->Gmask;
    hermes_dst_fmt.b = fmt->Bmask;
    hermes_dst_fmt.a = fmt->Amask;

    if (0 == Hermes_ConverterRequest (hermes_cvt_handle,
                                      &hermes_src_fmt,
                                      &hermes_dst_fmt)) {
      Fail ("Hermes_ConverterRequest failed\n");
    }

    Hermes_ClearerRequest (hermes_clr_handle, &hermes_src_fmt);
    Hermes_ClearerClear (hermes_clr_handle,
                         vid_buf, 0, 0, vid.w, vid.h, vid.pitch,
                         0, 0, 0, 0);
  }
#endif

  log_video (__func__);
  memset (vid.palette, 0, sizeof (vid.palette));
  pal_dirty = 1;
  vid.typ = info->typ;
  UNLOCK ();
  must_resize = 1;
  SIG_unblock (was_blocked);
  return 0;

 SDL_error:
  log ("%s failed with `%s'\n", msg, SDL_GetError ());
  UNLOCK ();
  SIG_unblock (was_blocked);
  return -1;
}

int VID_is_locked ()
{
  return vid_lock;
}

int VID_lock ()
{
  if (vid_lock) {
    lwarn ("video already locked\n");
    return 1;
  }
  else {
    vid_lock = 1;
    return 0;
  }
}

void VID_maybe_unlock (int was_locked)
{
  vid_lock = was_locked;
}

void VID_unlock ()
{
  if (0 == vid_lock) {
    lwarn ("attempt to unlock not locked video\n");
    return;
  }
  vid_lock = 0;
}

void VID_set_palette (const ub *pal, int first, int ncolors)
{
  int i;

  for (i = first; i < ncolors; i++) {
    vid.palette[i].r = pal[i * 3 + 0] << 2;
    vid.palette[i].g = pal[i * 3 + 1] << 2;
    vid.palette[i].b = pal[i * 3 + 2] << 2;
  }
  pal_dirty = 1;
}

void VID_set_font (const ub *font, int from,
                   int nelems, int xres, int yres)
{
  ub *new_font;

  MUST_LOCK ();
  if (8 != xres) {
    Fail ("unsupported font width %d", xres);
  }

  ldebug ("(font=%p, from=%d, nelems=%d, xres=%d, yres=%d)\n",
          font, from, nelems, xres, yres);

  if (NULL == vid.font) {
    new_font = xmalloc (256 * yres);
    memset (new_font, 0, 256 * yres);
  }
  else {
    if (yres != vid.font_h) {
      xfree (vid.font);
      new_font = xmalloc (256 * yres);
      memset (new_font, 0, 256 * yres);
    }
    else {
      new_font = vid.font;
    }
  }

  memcpy (new_font + from * yres, font, nelems * yres);

  vid.font = new_font;
  vid.font_w = xres;
  vid.font_h = yres;
  UNLOCK ();
}

void VID_text_mode_draw (const ub *src)
{
  ub *dst, *font;
  const ub *row;
  int x, y;
  int xres, yres;
  int cell_xres, cell_yres;

  if (IS_LOCKED ())
    return;

  LOCK ();
  Fail_if (NULL == vid.font);
  Fail_if (NULL == vid_buf);

  cell_xres = vid.font_w;
  cell_yres = vid.font_h;
  xres = vid.w / cell_xres;
  yres = vid.h / cell_yres;
  dst = vid_buf;
  font = vid.font;

  row = src;
  for (y = 0; y < yres; y++, row += xres * 2) {
    int yy;

    for (yy = 0; yy < cell_yres; yy++) {
      ub *fd;
      const ub *crow;

      fd = font + yy;
      crow = row;

      for (x = 0; x < xres; x++) {
        int xx;
        ub code, attr, bits;
        ub fg, bg;
        int fi, bi;

        code = *crow++;
        attr = *crow++;

#define BM(x) (!(x) - 1)
        fi = BM (attr & 0x08) & 0x38;
        bi = BM (attr & 0x80) & 0x38;
#undef BM

        fg = (attr & 7) + fi;
        bg = ((attr >> 4) & 7) - fg + bi;

        bits = fd[code * cell_yres];

        for (xx = 0; xx < cell_xres; xx++) {
          int mask;

          mask = ((bits >> (7 - xx)) & 1) - 1;
          *dst++ = (mask & bg) + fg;
        }
      }
    }
  }

  UNLOCK ();
}

static void refresh ()
{
  switch (vid.typ) {
    case VID_TYPE_VBE:
      ldebug ("vbe run\n");
      VBE_run ();

    case VID_TYPE_VGA:
      ldebug ("vga run\n");
      VGA_run ();
      break;
  }
}

#ifdef USE_HERMES
static void hermes_set_pal ()
{
  int i;
  int32 *hermes_pal;

  hermes_pal = Hermes_PaletteGet (hermes_pal_handle);
  for (i = 0; i < 256; i++) {
    int32 entry;

    entry = (vid.palette[i].r << 16)
      | (vid.palette[i].g << 8)
      | (vid.palette[i].b);
    hermes_pal[i] = entry;
  }
  Hermes_PaletteInvalidateCache (hermes_pal_handle);
  Hermes_ConverterPalette (hermes_cvt_handle, hermes_pal_handle, 0);
}

static void hermes_blit ()
{
  if (SDL_MUSTLOCK (screen))
    SDL_LockSurface (screen);

  if (0 == Hermes_ConverterCopy (hermes_cvt_handle,
                                 vid_buf,
                                 0, 0, vid.w, vid.h, vid.pitch,
                                 screen->pixels,
                                 0, 0, screen->w, screen->h, screen->pitch)) {
    Fail ("Hermes_ConverterCopy failed\n");
  }

  if (SDL_MUSTLOCK (screen))
    SDL_UnlockSurface (screen);
}
#endif

void VID_run ()
{
  if (IS_LOCKED ())
    return;

  LOCK ();

  if (must_resize) {
    must_resize = 0;
    screen = SDL_SetVideoMode (vid.screen_w, vid.screen_h,
                               screen->format->BitsPerPixel, vid.flags);
    if (NULL == screen) {
      Fail ("SDL_SetVideoMode failed with `%s'", SDL_GetError ());
    }

#ifdef USE_HERMES
    if (0 == use_hermes) {
#endif
      if (surface) {
        SDL_FreeSurface (surface);
      }

      surface =
        SDL_CreateRGBSurfaceFrom (vid_buf,
                                  vid.w, vid.h,
                                  vid.bpp,
                                  vid.pitch,
                                  vid.rmask,
                                  vid.gmask,
                                  vid.bmask,
                                  vid.amask);

      if (NULL == surface) {
        Fail ("SDL_CreateRGBSurfaceFrom failed with `%s'",
              SDL_GetError ());
      }

      /* SDL_FillRect (surface, NULL, 0); */
#ifdef USE_HERMES
    }
#endif
  }

  if (pal_dirty) {
    ldebug ("set palette\n");
    pal_dirty = 0;
#ifdef USE_HERMES
    if (use_hermes) {
      hermes_set_pal ();
    }
    /* else */
#else
    SDL_SetColors (surface, vid.palette, 0, 256);
#endif
  }

  if (new_vid_flags != vid.flags) {
    ldebug ("flip fullscreen\n");
    vid.flags = new_vid_flags;
    SDL_WM_GrabInput (SDL_GRAB_OFF);
    SDL_ShowCursor (vid.flags & SDL_FULLSCREEN ? SDL_DISABLE : SDL_ENABLE);
    screen = SDL_SetVideoMode (screen->w, screen->h,
                               screen->format->BitsPerPixel, vid.flags);
    if (NULL == screen) {
      Fail ("SDL_SetVideoMode failed with `%s'",
            SDL_GetError ());
    }
  }

  UNLOCK ();

  if (TID_get_nsec_diff (last_update) >= vid.period) {
    refresh ();
    last_update = TID_get_abs_nsec ();
  }
}

void VID_flip_fullscreen ()
{
  new_vid_flags = vid.flags ^ SDL_FULLSCREEN;
}

void VID_windowed ()
{
  if (screen) {
    new_vid_flags = vid.flags & ~SDL_FULLSCREEN;
    VID_run ();
  }
}

void VID_update ()
{
  if (IS_LOCKED ())
    return;

  LOCK ();
  ldebug ("update\n");

#ifdef USE_HERMES
  if (use_hermes) {
    hermes_blit ();
  }
  else
#endif
    SDL_BlitSurface (surface, NULL, screen, NULL);

  SDL_Flip (screen);
  UNLOCK ();
}

void VID_set_title (const char *title)
{
  static char buf[1024];

  snprintf (buf, 1024, "Minde %s", title);
  MUST_LOCK ();
  SDL_WM_SetCaption (buf, 0);
  UNLOCK ();
}

static void exitfn ()
{
  SIG_block ();
  SDL_QuitSubSystem (SDL_INIT_VIDEO);
  SDL_WM_GrabInput (SDL_GRAB_ON);
  SDL_WM_GrabInput (SDL_GRAB_OFF);
}

void VID_pre_init ()
{
  static struct MUP_key keys[] =
    {{"video.refresh.freq"   , TAG_IN_INT , {.ii = 30}}};

  MUP_add_keys (keys, LENOFA (keys));
}

void VID_init ()
{
  int rfreq;
  const SDL_VideoInfo *vinfo;

  log_level = MUP_get_log_level ("vid");

  if (-1 == SDL_InitSubSystem (SDL_INIT_VIDEO)) {
    Fail ("SDL_InitSubSystem(SDL_INIT_VIDEO) failed with %s",
          SDL_GetError ());
  }

  atexit (exitfn);

  vinfo = SDL_GetVideoInfo ();

  if (NULL == vinfo) {
    Fail ("SDL_GetVideoInfo error %s", SDL_GetError ());
  }

  memcpy (&vid.bestvfmt, vinfo->vfmt, sizeof (vid.bestvfmt));

  rfreq = MUP_get_int ("video.refresh.freq");
  rfreq = rfreq ? rfreq : 30;
  vid.period = TID_NSEC / rfreq;

#ifdef USE_HERMES
  use_hermes = Hermes_Init ();

  if (use_hermes) {
    hermes_cvt_handle = Hermes_ConverterInstance (HERMES_CONVERT_NORMAL);
    if (0 == hermes_cvt_handle) {
      lwarn ("Hermes_ConverterInstance failed giving up on Hermes\n");
      use_hermes = 0;
    }
    else {
      hermes_pal_handle = Hermes_PaletteInstance ();
      if (0 == hermes_pal_handle) {
        lwarn ("Hermes_PaletteInstance failed giving up on Hermes\n");
        use_hermes = 0;
      }
      else {
        hermes_clr_handle = Hermes_ClearerInstance ();
        if (0 == hermes_clr_handle)
          lwarn ("Hermes_ClearerInstance failed\n");
      }
    }
  }
#endif

  vid.flags = SDL_HWSURFACE | SDL_DOUBLEBUF;
#ifdef USE_HERMES
  if (use_hermes)
    vid.flags |= SDL_RESIZABLE;
#endif

  vid_buf = MEM_video_buffer ();
  new_vid_flags = vid.flags;

  VGA_init ();
  VBE_init ();
}
