#include <stdlib.h>
#include <string.h>

#include <inc/mem.h>
#include <inc/red.h>
#include <inc/int.h>
#include <inc/xms.h>
#include <inc/log.h>
#include <inc/mup.h>
#include <inc/dmm.h>

#define LOG_PREFIX "XMS"
static int log_level;

static int active;

#define NUM_HANDLES 20

struct desc {
  ub handle;
  ub lock_count;
  ui size;
  ui addr;
};

#define EMM_COUNT(e)    (*(ui*)((e)+0))
#define EMM_SRC(e)      (*(us*)((e)+4))
#define EMM_SRC_OFF(e)  (*(ui*)((e)+6))
#define EMM_DST(e)      (*(us*)((e)+10))
#define EMM_DST_OFF(e)  (*(us*)((e)+12))

static int free_handles = NUM_HANDLES;
static struct desc xms[NUM_HANDLES];

enum {
  SUCCESS             = 0x00,
  NOTIMPL             = 0x80,
  VDISK               = 0x81,
  A20ERR              = 0x82,
  GENERR              = 0x8e,
  FATERR              = 0x8f,
  NOHMA               = 0x90,
  HMAINUSE            = 0x91,
  DXSMALL             = 0x92,
  HMANOTALLOC         = 0x93,
  A20STILLENABLED     = 0x94,
  XMSEXHAUSTED        = 0xa0,
  OUTOFHANDLES        = 0xa1,
  INVHANDLE           = 0xa2,
  SRCINVHANDLE        = 0xa3,
  SRCINVOFFSET        = 0xa4,
  DSTINVHANDLE        = 0xa5,
  DSTINVOFFSET        = 0xa6,
  INVLENGTH           = 0xa7,
  OVERLAP             = 0xa8,
  PARITYERR           = 0xa9,
  BLOCKNOTLOCKED      = 0xaa,
  BLOCKLOCKED         = 0xab,
  BLOCKLOCKOVERFLOW   = 0xac,
  BLOCKFAILED         = 0xad,
  ONLYSMALLERUMBAVAIL = 0xb0,
  NOUMB               = 0xb1,
  INVUMBSEG           = 0xb2,
};

static int alloc (int kbs)
{
  int i;
  ui addr;

  for (i = 0; i < NUM_HANDLES; i++) {
    if (0 == xms[i].handle)
      break;
  }

  free_handles--;

  addr = (ui) DMM_alloc (kbs * 1024);
  Fail_if (0 == addr);

  xms[i].handle = i + 1;
  xms[i].lock_count = 0;
  xms[i].size = kbs;
  xms[i].addr = addr;

  return xms[i].handle;
}

static int is_handle_invalid (us handle)
{
  if (handle >= NUM_HANDLES)
    return 1;

  if (xms[handle].handle != handle + 1)
    return 2;

  return 0;
}

void XMS_func (INT_ARGS)
{
  int err;

  if (log_level < LOG_DEBUG)
    linfo ("ah = %#04x\n", AH (r));
  else {
    ldebug ("ah = %#04x ", AH (r));
    INT_log_regs (r);
  }

  switch (AH (r)) {
    case 0x00:
      SET_AX (r, 0x0200);
      SET_DX (r, 0);
      return;

    case 0x01:
    case 0x02:
      err = NOHMA;
      goto error;

    case 0x03:
    case 0x05:
      break;

    case 0x04:
    case 0x06:
      err = A20STILLENABLED;
      goto error;

    case 0x07:
      SET_AX (r, 1);
      return;

    case 0x08: {
      int largest, total;

      largest = DMM_largest () / 1024;
      total = (MEM_ram_size () - 0x100000) / 1024;

      SET_AX (r, largest > MEM_MEM_SIZE ? MEM_MEM_SIZE : largest);
      SET_DX (r, total > MEM_MEM_SIZE ? MEM_MEM_SIZE : total);
      return;
    }

    case 0x09: {
      ui kbs;

      if (0 == free_handles) {
        err = OUTOFHANDLES;
        lwarn ("XMS out of handles\n");
        goto error;
      }

      kbs = DX (r);

      if (DMM_largest () < (kbs * 1024)) {
        err = XMSEXHAUSTED;
        lwarn ("memory exhausted");
        goto error;
      }

      SET_DX (r, alloc (kbs));
      break;
    }

    case 0x0a: {
      us h;

      h = DX (r) - 1;

      if (is_handle_invalid (h)) {
        err = INVHANDLE;
        goto error;
      }

      if (xms[h].lock_count) {
        err = BLOCKLOCKED;
        goto error;
      }

      xms[h].handle = 0;
      DMM_free ((void *) xms[h].addr, xms[h].size * 1024);
      break;
    }

    case 0x0b: {
      ub *emm;
      us src_h, dst_h;
      ui src_o, dst_o;
      ui src_len, dst_len;
      ui src_ea, dst_ea;
      ui size;
      ui count;

      emm = (ub *) (DS (r) * 16 + SI (r));
      src_h = EMM_SRC (emm);
      dst_h = EMM_DST (emm);
      src_o = EMM_SRC_OFF (emm);
      dst_o = EMM_DST_OFF (emm);
      count = EMM_COUNT (emm);

      if (count & 1) {
        err = INVLENGTH;
        BPT;
        goto error;
      }

      if (src_h) {
        us h;

        h = src_h - 1;

        if (is_handle_invalid (h)) {
          err = SRCINVHANDLE;
          goto error;
        }

        size = xms[h].size * 1024;
        if (src_o > size) {
          err = SRCINVOFFSET;
          goto error;
        }

        src_ea = xms[h].addr + src_o;
        src_len = size - src_o;
      }
      else {
        src_ea = ((src_o >> 16) << 4) + (src_o & 0xffff);
        src_len = 0xa0000 - src_o;
      }

      if (dst_h) {
        us h;

        h = dst_h - 1;

        if (is_handle_invalid (h)) {
          err = DSTINVHANDLE;
          goto error;
        }

        size = xms[h].size * 1024;
        if (dst_o > size) {
          err = DSTINVOFFSET;
          goto error;
        }

        dst_ea = xms[h].addr + dst_o;
        dst_len = size - dst_o;
      }
      else {
        dst_ea = ((dst_o >> 16) << 4) + (dst_o & 0xffff);
        dst_len = 0xa0000 - dst_o;
      }

      if ((count > src_len) || (count > dst_len)) {
        err = INVLENGTH;
        BPT;
        goto error;
      }

      if ((src_ea < dst_ea) && ((src_ea + count) > dst_ea)) {
        err = OVERLAP;
        goto error;
      }

      if ((dst_ea < src_ea) && ((dst_ea + count) > src_ea)) {
        err = OVERLAP;
        goto error;
      }

      memcpy ((void *) dst_ea, (void *) src_ea, count);
      break;
    }

    case 0x0c: {
      us h;

      h = DX (r) - 1;

      if (is_handle_invalid (h)) {
        err = INVHANDLE;
        goto error;
      }

      if (0xffff == xms[h].lock_count) {
        err = BLOCKLOCKOVERFLOW;
        goto error;
      }

      xms[h].lock_count++;
      SET_DX (r, xms[h].addr >> 16);
      SET_BX (r, xms[h].addr & 0xffff);
      break;
    }

    case 0x0d: {
      us h;

      h = DX (r) - 1;

      if (is_handle_invalid (h)) {
        err = INVHANDLE;
        goto error;
      }

      if (0 == xms[h].lock_count) {
        err = BLOCKLOCKOVERFLOW;
        goto error;
      }

      xms[h].lock_count--;
      break;
    }

    case 0x0e: {
      us h;

      h = DX (r) - 1;
      if (is_handle_invalid (h)) {
        err = INVHANDLE;
        goto error;
      }

      SET_BH (r, xms[h].lock_count);
      SET_BL (r, free_handles);
      SET_DX (r, xms[h].size);
      break;
    }

    case 0x0f: {
      us h;
      ui new_addr, old_addr;
      ui new_size, old_size;

      h = DX (r) - 1;
      if (is_handle_invalid (h)) {
        err = INVHANDLE;
        goto error;
      }

      old_addr = xms[h].addr;
      old_size = xms[h].size * 1024;
      new_size = BX (r) * 1024;

      new_addr = (ui) DMM_realloc ((void *) old_addr, old_size, new_size);
      if (0 == new_addr) {
        err = XMSEXHAUSTED;
        goto error;
      }

      xms[h].addr = new_addr;
      xms[h].size = BX (r);
      break;
    }

    case 0x10:
    case 0x11:
    case 0x12:
      err = NOUMB;
      goto error;

    default:
      UNHANDLED_INT (r);
      break;
  }

  SET_AX (r, 1);
  return;

 error:
  lwarn ("XMS error for function %#x = %#x\n", AH (r), err);
  SET_AX (r, 0);
  SET_BL (r, err);
}

int XMS_active ()
{
  return active;
}

void XMS_pre_init ()
{
  static struct MUP_key keys[] = {
    {"xms.active", TAG_INT, {.ip = &active}},
  };

  MUP_add_keys (keys, LENOFA (keys));
}

void XMS_init ()
{
  log_level = MUP_get_log_level ("xms");

  if (active) {
    *(ub *) RED_FLAT (RED_XMS_JUMP) = 0xea;
    *(us *) RED_FLAT (RED_XMS_JUMP + 1) = RED_XMS_FUNC;
    *(us *) RED_FLAT (RED_XMS_JUMP + 3) = RED_SEGMENT;
    *(ub *) RED_FLAT (RED_XMS_FUNC) = RED_ICE;
  }
}
